!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


program sample
  use amgs_module
  implicit none
  integer(kind=4)   :: row, i, width, rowsize, maxiter
  real(kind=8)   :: v, threshold
  ! crs matrix data
  integer(kind=4), allocatable :: crs_index(:),crs_column(:) 
  real(kind=8), allocatable    ::  X(:), B(:), crs_val(:)
  logical :: SETUP_AMG=.false.

  ! -------------------------------------------------------------
  ! *** 2D poisson problem with 5-point stencil in CRS format ***
  ! *** Problem domain size is width x width                  ***
  ! *** CRS matrix: crs_index(0:), crs_column(:), crs_val(:)  ***
  ! -------------------------------------------------------------
  read(*,*) width
  rowsize = width*width
  allocate(crs_index(0:rowsize),crs_column(rowsize*5), crs_val(rowsize*5))
  crs_index(0) = 0

  i = 0
  DO row = 1, rowsize
     if(row-width > 0) then
        i=i+1;crs_val(i) = -1.0;     crs_column(i) = row-width;  
     end if
     if(mod((row-1),width) /= 0) then
        i=i+1; crs_val(i) = -1.0;     crs_column(i) = row-1;      
     end if
     i=i+1;crs_val(i) = 4.0;     crs_column(i) = row;        
     if(mod((row+1),width) /= 1) then
        i=i+1; crs_val(i) = -1.0;     crs_column(i) = row+1;      
     end if
     if(row+width <= rowsize) then
        i=i+1; crs_val(i) = -1.0;     crs_column(i) = row+width;  
     end if
     crs_index(row) = i
  END DO
  allocate(B(rowsize), X(rowsize))

  ! right hand side vector is set
  do row = 1, rowsize
     B(row) = 0
     do i = crs_index(row-1)+1, crs_index(row)
        B(row) = B(row)+crs_val(i)
     end do
  end do

  X = 0
  ! -----------------------------------------------
  ! *** AMG parameters is set from AMG.dat file ***
  ! -----------------------------------------------
!C  CALL amgs_parameter_set(10)
  CALL amgs_parameter_option_file(10, "AMGS_OPTION.dat")
  CALL amgs_option_info(1)

  ! ----------------------------------------------------------
  ! *** crs problem matrix is stored in amgs libary module ***
  ! ----------------------------------------------------------
  call amgs_store_problem_matrix(crs_index, crs_column, crs_val, rowsize)

  ! ----------------------------------
  ! *** terminal condition is set ***
  ! ----------------------------------
  maxiter = 20
  threshold = 1.0E-7
  call amgs_terminal_val_set(threshold)

  ! ----------------------------------
  ! *** AMG_solver is called ***
  ! ----------------------------------
  call amgs_solver(X, B, maxiter, SETUP_AMG)

  ! -------------------------------------------------------------------
  ! *** clear problem matrix and multi-level data in library module ***
  ! -------------------------------------------------------------------
  call amgs_clear_matrices()

  deallocate(B, X, crs_index, crs_column, crs_val)
end program sample
