!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif

MODULE AMG_cycle
CONTAINS

  SUBROUTINE amg_cycles(N, problem_B, problem_X, LEVEL_NUM, Temp)
    USE amg_parameter, only: cycle_kind
    USE data_structure_for_AMG
    IMPLICIT NONE

    INTEGER(kind=kint),  INTENT(in)    :: LEVEL_NUM
    INTEGER(kind=kint), INTENT(in)     :: N

    REAL   (kind=kreal), DIMENSION(:), target ::  problem_B
    !C DIMENSION(N)
    REAL   (kind=kreal), DIMENSION(:), target ::  problem_X
    !C DIMENSION(N)
!!$    REAL   (kind=kreal), DIMENSION(N) :: Temp
    REAL   (kind=kreal), DIMENSION(:) :: Temp
    logical, save :: setinitial = .true.

    HIERARCHICAL_DATA(1) % B => problem_B
    HIERARCHICAL_DATA(1) % X => problem_X


    if(cycle_kind==0) then
       call V_cycle(LEVEL_NUM, Temp)
    else if(cycle_kind==1) then
       call F_cycle(LEVEL_NUM, Temp)
    else if(cycle_kind==2) then
       call W_cycle(LEVEL_NUM, Temp)
    end if

  END SUBROUTINE amg_cycles
  
  SUBROUTINE F_cycle(LEVEL_NUM, Temp)
    USE data_structure_for_AMG
    USE sgs_relaxation
    USE ilu_relaxation
    USE amg_parameter, only: f_first_iteration_number,gs_kind
    IMPLICIT NONE

    INTEGER(kind=kint),  INTENT(in)    :: LEVEL_NUM
    INTEGER(kind=kint)                 :: NP, N, gs_k

    REAL   (kind=kreal), DIMENSION(:), pointer :: D, ILU_D
    !C DIMENSION(N)
!!$    REAL   (kind=kreal), DIMENSION(N), target ::  problem_B
!!$    REAL   (kind=kreal), DIMENSION(:) ::  problem_B
    !C DIMENSION(N)
!!$    REAL   (kind=kreal), DIMENSION(N), target::  problem_X
!!$    REAL   (kind=kreal), DIMENSION(:) ::  problem_X
    !C DIMENSION(N)

    REAL   (kind=kreal), DIMENSION(: ), pointer::  B
    !C DIMENSION(N)
    REAL   (kind=kreal), DIMENSION(: ), pointer::  X   
    !C DIMENSION(N)

    REAL   (kind=kreal), DIMENSION(:), pointer::  AU, ILU_AU
    REAL   (kind=kreal), DIMENSION(:), pointer::  AL, ILU_AL

    REAL   (kind=kreal), DIMENSION(:), pointer :: coarser_B
    REAL   (kind=kreal), DIMENSION(:), pointer :: coarser_X
!!$    REAL   (kind=kreal), DIMENSION(N) :: Temp
    REAL   (kind=kreal), DIMENSION(:) :: Temp
    REAL   (kind=kreal) :: R_val,B_val,w,X_val,R_norm,GR_norm

    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INU
    !C DIMENSION(0:N)
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAU
    !C DIMENSION(NPU)
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INL
    !C DIMENSION(0:N)
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAL

    INTEGER(kind=kint ) :: nth_lev,i,inod,j,ieL,isL,is,ie,coarser_N,isU,ieU,k,coarser_NP
    INTEGER(kind=kint ) :: coarser_NEIBPETOT

    type(INTER_LEVEL_OPERATOR),pointer :: R, P

    
    DO nth_lev=1,LEVEL_NUM-1
       N =  HIERARCHICAL_DATA(nth_lev) % N
       NP=  HIERARCHICAL_DATA(nth_lev) % NP
       INU=>HIERARCHICAL_DATA(nth_lev) % INU 
       INL=>HIERARCHICAL_DATA(nth_lev) % INL 
       X => HIERARCHICAL_DATA(nth_lev) % X
       B => HIERARCHICAL_DATA(nth_lev) % B
       D => HIERARCHICAL_DATA(nth_lev) % D
       IAU=>HIERARCHICAL_DATA(nth_lev) % IAU
       IAL=>HIERARCHICAL_DATA(nth_lev) % IAL
       AU =>HIERARCHICAL_DATA(nth_lev) % AU
       AL =>HIERARCHICAL_DATA(nth_lev) % AL
       
       coarser_B => HIERARCHICAL_DATA(nth_lev+1) % B
       coarser_X => HIERARCHICAL_DATA(nth_lev+1) % X
       if(ILU0_SMOOTH_FL) then
          ILU_D => HIERARCHICAL_DATA(nth_lev) % ILU0_D
          ILU_AL => HIERARCHICAL_DATA(nth_lev) % ILU0_AL
          ILU_AU => HIERARCHICAL_DATA(nth_lev) % ILU0_AU
          CALL ilu(N, D, ILU_D, AL, ILU_AL, INL, IAL, AU, ILU_AU, INU, IAU, B, X, Temp)
       else
          if(gs_kind  == 0)  then
             gs_k = 0
          else if(gs_kind==1) then
             gs_k = 1
          else
             write(*,*) "Error gs_kind"
             stop
          end if
          CALL sgs(N,NP,D,AL,INL,IAL,AU,INU,IAU,B,X,f_first_iteration_number,nth_lev,gs_k)
       end if

       DO j = 1, N
          R_val= B(j) - D(j)*X(j)
          isU= INU(j-1)+1
          ieU= INU(j)
          DO i= isU, ieU
             inod=IAU(i)+ZERO_ORIGIN
             R_val= R_val - AU(i) * X(inod)
          END DO
          isL= INL(j-1)+1
          ieL= INL(j)
          DO i= isL, ieL
             inod= IAL(i)+ZERO_ORIGIN
             R_val= R_val - AL(i) * X(inod)
          END DO
          Temp(j)=R_val
       END DO
       
       !C restrict the residual vector
       R => HIERARCHICAL_DATA(nth_lev+1) % R
       coarser_N= HIERARCHICAL_DATA(nth_lev+1) % N
       coarser_NP=HIERARCHICAL_DATA(nth_lev+1) % NP

       DO j= 1, coarser_NP
          is= R % IN(j-1)+1
          ie= R % IN(j)
          B_val=0.0
          DO i=is, ie
             inod = R % CN(i)
             B_val= B_val + R % V(i) * Temp(inod)
          END DO
          coarser_B(j)= B_val
       END DO

       
       DO j=1,coarser_NP
          coarser_X(j)= 0.0
       END DO
    END DO

    DO nth_lev = LEVEL_NUM-1, 1, -1
       call v_cycle_mod(LEVEL_NUM, Temp, nth_lev+1, nth_lev)
    END DO

  END SUBROUTINE F_cycle

  SUBROUTINE W_cycle(LEVEL_NUM, Temp)
    IMPLICIT NONE
    include 'precision.inc'

    INTEGER(kind=kint),  INTENT(in)    :: LEVEL_NUM
    REAL(kind=kreal) :: Temp(:)
    INTEGER(kind=kint) :: w_lev
    
    w_lev = LEVEL_NUM/2+mod(LEVEL_NUM, 2)
    
    call v_cycle_mod(LEVEL_NUM, Temp, 1, w_lev)
    call v_cycle_mod(LEVEL_NUM, Temp, w_lev, 1)
    
  END SUBROUTINE W_cycle



  SUBROUTINE v_cycle(LEVEL_NUM, Temp)
    IMPLICIT NONE
    include 'precision.inc'

    INTEGER(kind=kint),  INTENT(in)    :: LEVEL_NUM
    REAL(kind=kreal) :: Temp(:)
    
    call v_cycle_mod(LEVEL_NUM, Temp, 1, 1)
    
    
  END SUBROUTINE v_cycle

  SUBROUTINE v_cycle_mod(LEVEL_NUM, Temp, PRE_LEVEL, POST_LEVEL)
    USE data_structure_for_AMG
    USE sgs_relaxation
    USE ilu_relaxation
    USE amg_parameter, only: iteration_number_up, iteration_number_down, iteration_number_bottom, gs_kind
    IMPLICIT NONE

    INTEGER(kind=kint),  INTENT(in)    :: LEVEL_NUM, PRE_LEVEL, POST_LEVEL
    INTEGER(kind=kint)                 :: N, NP, gs_k

    REAL   (kind=kreal), DIMENSION(:), pointer :: D, ILU_D

    REAL   (kind=kreal), DIMENSION(:), pointer::  B
    !C DIMENSION(N)
    REAL   (kind=kreal), DIMENSION(:), pointer::  X   
    !C DIMENSION(N)

    REAL   (kind=kreal), DIMENSION(:), pointer::  AU, ILU_AU
    REAL   (kind=kreal), DIMENSION(:), pointer::  AL, ILU_AL

    REAL   (kind=kreal), DIMENSION(:), pointer :: coarser_B
    REAL   (kind=kreal), DIMENSION(:), pointer :: coarser_X
!!$    REAL   (kind=kreal), DIMENSION(N) :: Temp
    REAL   (kind=kreal), DIMENSION(:) :: Temp
    REAL   (kind=kreal) :: R_val,B_val,w,X_val,R_norm,GR_norm

    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INU
    !C DIMENSION(0:N)
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAU
    !C DIMENSION(NPU)
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INL
    !C DIMENSION(0:N)
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAL

    INTEGER(kind=kint ) :: nth_lev,i,inod,j,ieL,isL,is,ie,coarser_N,isU,ieU,k,coarser_NP
    INTEGER(kind=kint ) :: coarser_NEIBPETOT

    type(INTER_LEVEL_OPERATOR), pointer :: R, P

!C    changed..
!!$    do i = 1, N
!!$       problem_X(i) = 0.0D0
!!$    END do

    DO nth_lev=PRE_LEVEL, LEVEL_NUM-1
       N= HIERARCHICAL_DATA(nth_lev) % N
       NP=HIERARCHICAL_DATA(nth_lev) % NP
       INU=>HIERARCHICAL_DATA(nth_lev) % INU 
       INL=>HIERARCHICAL_DATA(nth_lev) % INL 
       X => HIERARCHICAL_DATA(nth_lev) % X
       B => HIERARCHICAL_DATA(nth_lev) % B
       D => HIERARCHICAL_DATA(nth_lev) % D
       IAU=>HIERARCHICAL_DATA(nth_lev) % IAU
       IAL=>HIERARCHICAL_DATA(nth_lev) % IAL
       AU =>HIERARCHICAL_DATA(nth_lev) % AU
       AL =>HIERARCHICAL_DATA(nth_lev) % AL
       
       coarser_B => HIERARCHICAL_DATA(nth_lev+1) % B
       coarser_X => HIERARCHICAL_DATA(nth_lev+1) % X


!!$       !C- debug
!!$       R_val = 0
!!$       DO j = 1, N
!!$          R_val = R_val+X(j)**2
!!$       END DO
!!$       write(*,*) nth_lev,"X: before smoothing", dsqrt(R_val)
!!$       !C- debug
       
       

       if(ILU0_SMOOTH_FL) then
          ILU_D => HIERARCHICAL_DATA(nth_lev) % ILU0_D
          ILU_AL => HIERARCHICAL_DATA(nth_lev) % ILU0_AL
          ILU_AU => HIERARCHICAL_DATA(nth_lev) % ILU0_AU
          CALL ilu(N, D, ILU_D, AL, ILU_AL, INL, IAL, AU, ILU_AU, INU, IAU, B, X, Temp)
       else
!!$          if(nth_lev == 1)then

!!$          CALL jacobi(N,NP,D,AL,INL,IAL,AU,INU,IAU,B,X,iteration_number_down,nth_lev)
!!$          else
          if(gs_kind  == 0)  then
             gs_k = 0
          else if(gs_kind==1) then
             gs_k = 1
          else
             write(*,*) "Error gs_kind"
             stop
          end if


          CALL sgs(N,NP,D,AL,INL,IAL,AU,INU,IAU,B,X,iteration_number_down,nth_lev,gs_k)
!!$          end if

       end if
!!$       !C- debug
!!$       R_val = 0
!!$       DO j = 1, N
!!$          R_val = R_val+X(j)**2
!!$       END DO
!!$       write(*,*) nth_lev,"X: after smoothing", dsqrt(R_val)
!!$       !C- debug

!!$       !C- debug
!!$       R_val = 0
!!$       DO j = 1, N
!!$          R_val = R_val+B(j)**2
!!$       END DO
!!$       write(*,*) nth_lev,"B: after smoothing", dsqrt(R_val)
!!$       !C- debug

       DO j = 1, N
          R_val= B(j) - D(j)*X(j)
          isU= INU(j-1)+1
          ieU= INU(j)
          DO i= isU, ieU
             inod=IAU(i)+ZERO_ORIGIN
             R_val= R_val - AU(i) * X(inod)
          END DO
          isL= INL(j-1)+1
          ieL= INL(j)
          DO i= isL, ieL
             inod= IAL(i)+ZERO_ORIGIN
             R_val= R_val - AL(i) * X(inod)
          END DO

          Temp(j)=R_val

       END DO

       !C restrict the residual vector
       R => HIERARCHICAL_DATA(nth_lev+1) % R
       coarser_N= HIERARCHICAL_DATA(nth_lev+1) % N
       coarser_NP=HIERARCHICAL_DATA(nth_lev+1) % NP

       DO j= 1, coarser_NP
          is= R % IN(j-1)+1
          ie= R % IN(j)
          B_val=0.0
          DO i=is, ie
             inod = R % CN(i)
             B_val= B_val + R % V(i) * Temp(inod)
          END DO
          coarser_B(j)= B_val
       END DO

       DO j=1,coarser_NP
          coarser_X(j)= 0.0
       END DO
    END DO
    
    N  = HIERARCHICAL_DATA(LEVEL_NUM) % N
    NP = HIERARCHICAL_DATA(LEVEL_NUM) % NP

    X => HIERARCHICAL_DATA(LEVEL_NUM) % X
    B => HIERARCHICAL_DATA(LEVEL_NUM) % B
    AL=> HIERARCHICAL_DATA(LEVEL_NUM) % AL
    INL=>HIERARCHICAL_DATA(LEVEL_NUM) % INL 

    INU => HIERARCHICAL_DATA(LEVEL_NUM) % INU 
    D   => HIERARCHICAL_DATA(LEVEL_NUM) % D
    IAU => HIERARCHICAL_DATA(LEVEL_NUM) % IAU
    IAL => HIERARCHICAL_DATA(LEVEL_NUM) % IAL
    AU  => HIERARCHICAL_DATA(LEVEL_NUM) % AU

!!$    !C- debug
!!$    R_val = 0
!!$    DO j = 1, N
!!$       R_val = R_val+X(j)**2
!!$    END DO
!!$    write(*,*) LEVEL_NUM,"X: before smoothing", dsqrt(R_val)
!!$    !C- debug

    
    if(ILU0_SMOOTH_FL) then
       ILU_D => HIERARCHICAL_DATA(LEVEL_NUM) % ILU0_D
       ILU_AL => HIERARCHICAL_DATA(LEVEL_NUM) % ILU0_AL
       ILU_AU => HIERARCHICAL_DATA(LEVEL_NUM) % ILU0_AU
       CALL ilu(N, D, ILU_D, AL, ILU_AL, INL, IAL, AU, ILU_AU, INU, IAU, B, X, Temp)
    else
       CALL sgs(N, NP, D, AL, INL, IAL, AU, INU, IAU, B, X, iteration_number_bottom, LEVEL_NUM, 0)
!!$       CALL jacobi(N, NP, D, AL, INL, IAL, AU, INU, IAU, B, X, iteration_number_bottom, LEVEL_NUM)
    end if

!!$    !C- debug
!!$    R_val = 0
!!$    DO j = 1, N
!!$       R_val = R_val+B(j)**2
!!$    END DO
!!$    write(*,*) LEVEL_NUM,"B: after smoothing:", dsqrt(R_val)
!!$
!!$    R_val = 0
!!$    DO j = 1, N
!!$       R_val = R_val+X(j)**2
!!$    END DO
!!$    write(*,*) nth_lev,"X: after smoothing:", dsqrt(R_val)
!!$    !C- debug


    DO nth_lev=LEVEL_NUM-1,POST_LEVEL,-1
       N= HIERARCHICAL_DATA(nth_lev) % N
       NP=HIERARCHICAL_DATA(nth_lev) % NP

       INU          => HIERARCHICAL_DATA(nth_lev) % INU
       INL          => HIERARCHICAL_DATA(nth_lev) % INL
       X            => HIERARCHICAL_DATA(nth_lev) % X
       B            => HIERARCHICAL_DATA(nth_lev) % B
       D            => HIERARCHICAL_DATA(nth_lev) % D
       IAU          => HIERARCHICAL_DATA(nth_lev) % IAU
       IAL          => HIERARCHICAL_DATA(nth_lev) % IAL
       AU           => HIERARCHICAL_DATA(nth_lev) % AU
       AL           => HIERARCHICAL_DATA(nth_lev) % AL
       coarser_X    => HIERARCHICAL_DATA(nth_lev+1) % X
       P => HIERARCHICAL_DATA(nth_lev+1) % P

       DO j= 1, N
          is= P % IN(j-1)+1
          ie= P % IN(j)
          X_val=0.0

          DO i=is, ie
             inod = P % CN(i)
             X_val= X_val+P % V(i) * coarser_X(inod)
          END DO
          X(j)=X(j)+X_val
       END DO

!!$       !C- debug
!!$       R_val = 0
!!$       DO j = 1, N
!!$          R_val = R_val+X(j)**2
!!$       END DO
!!$       write(*,*) nth_lev,"X:Before smoothing", dsqrt(R_val)
!!$       !C- debug
    


       if(ILU0_SMOOTH_FL) then
          ILU_D => HIERARCHICAL_DATA(nth_lev) % ILU0_D
          ILU_AL => HIERARCHICAL_DATA(nth_lev) % ILU0_AL
          ILU_AU => HIERARCHICAL_DATA(nth_lev) % ILU0_AU
          CALL ilu(N, D, ILU_D, AL, ILU_AL, INL, IAL, AU, ILU_AU, INU, IAU, B, X, Temp)       
       else
!!$          if(nth_lev == 1)then
!!$          CALL jacobi(N,NP,D,AL,INL,IAL,AU,INU,IAU,B,X,iteration_number_up,nth_lev)
!!$          else 
          if(gs_kind  == 0)  then
             gs_k = 0
          else if(gs_kind==1) then
             gs_k = 2
          else
             write(*,*) "Error gs_kind"
             stop
          end if

          CALL sgs(N,NP,D,AL,INL,IAL,AU,INU,IAU,B,X,iteration_number_up,nth_lev, gs_k)
          
!!$          !C- debug
!!$          R_val = 0
!!$          DO j = 1, N
!!$             R_val = R_val+X(j)**2
!!$          END DO
!!$          write(*,*) nth_lev,"X: after smoothing:", dsqrt(R_val)
!!$          !C- debug

!!$          end if
       end if
    END DO
  END SUBROUTINE v_cycle_mod


  SUBROUTINE v_check(count, LEVEL_NUM, mu)
    USE data_structure_for_AMG
    USE sgs_relaxation
    USE ilu_relaxation
    IMPLICIT NONE

    INTEGER(kind=kint),  INTENT(in)    :: LEVEL_NUM, count
    REAL   (kind=kreal), INTENT(inout) :: mu

    INTEGER(kind=kint)                 :: N, NP

    REAL   (kind=kreal), DIMENSION(:), pointer :: D, ILU_D
    !C DIMENSION(N)
    REAL   (kind=kreal), target, allocatable, DIMENSION(:) ::  problem_B
!!$    REAL   (kind=kreal), DIMENSION(:), target ::  problem_B
    !C DIMENSION(N)
    REAL   (kind=kreal), target, allocatable, DIMENSION(:)::  problem_X
!!$    REAL   (kind=kreal), DIMENSION(:), target::  problem_X
    !C DIMENSION(N)

    REAL   (kind=kreal), DIMENSION(: ), pointer::  B
    !C DIMENSION(N)
    REAL   (kind=kreal), DIMENSION(: ), pointer::  X   
    !C DIMENSION(N)

    REAL   (kind=kreal), DIMENSION(:), pointer::  AU, ILU_AU
    REAL   (kind=kreal), DIMENSION(:), pointer::  AL, ILU_AL

    REAL   (kind=kreal), DIMENSION(:), pointer :: coarser_B
    REAL   (kind=kreal), DIMENSION(:), pointer :: coarser_X
!!$    REAL   (kind=kreal), DIMENSION(N) :: Temp
    REAL   (kind=kreal), allocatable, DIMENSION(:) :: Temp
    REAL   (kind=kreal) :: R_val,B_val,w,X_val,R_norm,GR_norm, e_nrm, e_nrm2

    
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INU
    !C DIMENSION(0:N)
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAU
    !C DIMENSION(NPU)
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INL
    !C DIMENSION(0:N)
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAL

    INTEGER(kind=kint ) :: nth_lev,i,inod,j,ieL,isL,is,ie,coarser_N,isU,ieU,k,coarser_NP
    INTEGER(kind=kint ) :: coarser_NEIBPETOT, cnt
    
    type(INTER_LEVEL_OPERATOR),pointer :: R, P

    REAL  :: rv


    N = HIERARCHICAL_DATA(1) % N

    allocate(problem_B(N), problem_X(N), Temp(N))

    !C random value is set to problem_X(:)
    call random_seed()
    DO i=1, N
       call RANDOM_NUMBER(rv)
       problem_X(i) = 0.5+1.0/2.0*rv
    END DO
    problem_B(:) = 0


    HIERARCHICAL_DATA(1) % B =>problem_B
    HIERARCHICAL_DATA(1) % X =>problem_X

    !C    changed..
!!$    do i = 1, N
!!$       problem_X(i) = 0.0D0
!!$    END do

    DO cnt = 1, count
       DO nth_lev=1,LEVEL_NUM-1
          N= HIERARCHICAL_DATA(nth_lev) % N
          NP=HIERARCHICAL_DATA(nth_lev) % NP
          INU=>HIERARCHICAL_DATA(nth_lev) % INU 
          INL=>HIERARCHICAL_DATA(nth_lev) % INL 
          X => HIERARCHICAL_DATA(nth_lev) % X
          B => HIERARCHICAL_DATA(nth_lev) % B
          D => HIERARCHICAL_DATA(nth_lev) % D
          IAU=>HIERARCHICAL_DATA(nth_lev) % IAU
          IAL=>HIERARCHICAL_DATA(nth_lev) % IAL
          AU =>HIERARCHICAL_DATA(nth_lev) % AU
          AL =>HIERARCHICAL_DATA(nth_lev) % AL

          coarser_B => HIERARCHICAL_DATA(nth_lev+1) % B
          coarser_X => HIERARCHICAL_DATA(nth_lev+1) % X

          if(ILU0_SMOOTH_FL) then
             ILU_D => HIERARCHICAL_DATA(nth_lev) % ILU0_D
             ILU_AL => HIERARCHICAL_DATA(nth_lev) % ILU0_AL
             ILU_AU => HIERARCHICAL_DATA(nth_lev) % ILU0_AU
             CALL ilu(N, D, ILU_D, AL, ILU_AL, INL, IAL, AU, ILU_AU, INU, IAU, B, X, Temp)
          else
             CALL sgs(N,NP,D,AL,INL,IAL,AU,INU,IAU,B,X,2,nth_lev, 0)
          end if


          DO j = 1, N
             R_val= B(j) - D(j)*X(j)
             isU= INU(j-1)+1
             ieU= INU(j)
             DO i= isU, ieU
                inod=IAU(i)+ZERO_ORIGIN
                R_val= R_val - AU(i) * X(inod)
             END DO
             isL= INL(j-1)+1
             ieL= INL(j)
             DO i= isL, ieL
                inod= IAL(i)+ZERO_ORIGIN
                R_val= R_val - AL(i) * X(inod)
             END DO

             Temp(j)=R_val

          END DO

          !C restrict the residual vector
          R => HIERARCHICAL_DATA(nth_lev+1) % R
          coarser_N= HIERARCHICAL_DATA(nth_lev+1) % N
          coarser_NP=HIERARCHICAL_DATA(nth_lev+1) % NP

          DO j= 1, coarser_NP
             is= R % IN(j-1)+1
             ie= R % IN(j)
             B_val=0.0
             DO i=is, ie
                inod = R % CN(i)
                B_val= B_val + R % V(i) * Temp(inod)
             END DO
             coarser_B(j)= B_val
          END DO

          DO j=1,coarser_NP
             coarser_X(j)= 0.0
          END DO
       END DO


       N=HIERARCHICAL_DATA(LEVEL_NUM) % N
       NP=HIERARCHICAL_DATA(LEVEL_NUM) % NP

       X => HIERARCHICAL_DATA(LEVEL_NUM) % X
       B => HIERARCHICAL_DATA(LEVEL_NUM) % B
       AL=> HIERARCHICAL_DATA(LEVEL_NUM) % AL
       INL=>HIERARCHICAL_DATA(LEVEL_NUM) % INL 

       INU => HIERARCHICAL_DATA(LEVEL_NUM) % INU 
       D   => HIERARCHICAL_DATA(LEVEL_NUM) % D
       IAU => HIERARCHICAL_DATA(LEVEL_NUM) % IAU
       IAL => HIERARCHICAL_DATA(LEVEL_NUM) % IAL
       AU  => HIERARCHICAL_DATA(LEVEL_NUM) % AU

       if(ILU0_SMOOTH_FL) then
          ILU_D => HIERARCHICAL_DATA(LEVEL_NUM) % ILU0_D
          ILU_AL => HIERARCHICAL_DATA(LEVEL_NUM) % ILU0_AL
          ILU_AU => HIERARCHICAL_DATA(LEVEL_NUM) % ILU0_AU
          CALL ilu(N, D, ILU_D, AL, ILU_AL, INL, IAL, AU, ILU_AU, INU, IAU, B, X, Temp)
       else
          CALL sgs(N, NP, D, AL, INL, IAL, AU, INU, IAU, B, X, 2, LEVEL_NUM,0)
       end if



       DO nth_lev=LEVEL_NUM-1,1,-1
          N= HIERARCHICAL_DATA(nth_lev) % N
          NP=HIERARCHICAL_DATA(nth_lev) % NP

          INU          => HIERARCHICAL_DATA(nth_lev) % INU
          INL          => HIERARCHICAL_DATA(nth_lev) % INL
          X            => HIERARCHICAL_DATA(nth_lev) % X
          B            => HIERARCHICAL_DATA(nth_lev) % B
          D            => HIERARCHICAL_DATA(nth_lev) % D
          IAU          => HIERARCHICAL_DATA(nth_lev) % IAU
          IAL          => HIERARCHICAL_DATA(nth_lev) % IAL
          AU           => HIERARCHICAL_DATA(nth_lev) % AU
          AL           => HIERARCHICAL_DATA(nth_lev) % AL
          coarser_X    => HIERARCHICAL_DATA(nth_lev+1) % X
          P => HIERARCHICAL_DATA(nth_lev+1) % P


          DO j= 1, N
             is= P % IN(j-1)+1
             ie= P % IN(j)
             X_val=0.0
             DO i=is, ie
                inod = P % CN(i)
                X_val= X_val+P % V(i) * coarser_X(inod)
             END DO
             X(j)=X(j)+X_val
          END DO


          if(ILU0_SMOOTH_FL) then
             ILU_D => HIERARCHICAL_DATA(nth_lev) % ILU0_D
             ILU_AL => HIERARCHICAL_DATA(nth_lev) % ILU0_AL
             ILU_AU => HIERARCHICAL_DATA(nth_lev) % ILU0_AU
             CALL ilu(N, D, ILU_D, AL, ILU_AL, INL, IAL, AU, ILU_AU, INU, IAU, B, X, Temp)       
          else
             CALL sgs(N, NP, D, AL, INL, IAL, AU, INU, IAU, B, X, 2, nth_lev,0)
          end if
       END DO
       
!!$       if(cnt >= count-1) then
          e_nrm = 0.0
          DO i = 1, N
             e_nrm = e_nrm + X(i)*X(i)
          END DO
          e_nrm = dsqrt(e_nrm)
          write(*,*) e_nrm
!!$       end if

       if(cnt == count-1) e_nrm2=e_nrm
       
    END DO

    if(e_nrm2 == 0.0) then
       mu = 0.1
    else
       mu = e_nrm/e_nrm2
       if(mu<0.1) mu = 0.1
    end if
    
    deallocate(problem_B, problem_X, Temp)
  END SUBROUTINE v_check


  SUBROUTINE matrix_counting(LEVEL_NUM,SOLVER_COMM,my_rank)
    USE data_structure_for_AMG
    IMPLICIT NONE

    integer(kind=kint) ,intent(in) :: LEVEL_NUM,SOLVER_COMM,my_rank
    integer(kind=kint) :: i,j,k, nonzero_l,nonzero_g,level,n,np
    REAL,DIMENSION(:),allocatable::comm_rate
    REAL:: rate,crate_l,crate_g

    allocate(comm_rate(LEVEL_NUM))
    DO level=1,LEVEL_NUM
       n= HIERARCHICAL_DATA(level) % N
       np=HIERARCHICAL_DATA(level) % NP
       !C in the case of n == 0
       if(np == 0) then
          crate_l=1
       else
          crate_l=REAL(n)/np
       end if
       crate_g=0

       crate_g=crate_l
       comm_rate(level) = crate_g
    END DO

    nonzero_l=0
    k = 0
    DO level=1,LEVEL_NUM
       n= HIERARCHICAL_DATA(level) % N
       nonzero_l=nonzero_l+ n + HIERARCHICAL_DATA(level)%INU(n)
       nonzero_l=nonzero_l+  HIERARCHICAL_DATA(level)%INL(n)
       k = k + n
    END DO

    nonzero_g=0

    nonzero_g=nonzero_l
    i=nonzero_g


    n=HIERARCHICAL_DATA(1)%N

    !C in the case of n == 0
    nonzero_l = 0 
    if(n>0) then
       nonzero_l=n+HIERARCHICAL_DATA(1)%INU(n)+HIERARCHICAL_DATA(1)%INL(n)
    end if

    nonzero_g = 0

    nonzero_g=nonzero_l
    j=nonzero_g

    if(my_rank==0)then
       rate=REAL(i)/j
       write(*,*) "nonzeros in all level:", i, "rate of level:",rate
       rate=REAL(k)/n
       write(*,*) "rows in all level:", k, "rate of level:", rate
!!$       write(*,*) comm_rate
    end if
  END SUBROUTINE matrix_counting


  SUBROUTINE single_matrix_counting(LEVEL_NUM)
    USE data_structure_for_AMG
    IMPLICIT NONE

    integer(kind=kint) ,intent(in) :: LEVEL_NUM
    integer(kind=kint) :: i, j ,k, nonzero_l, nonzero_g, level, n, np
    REAL,DIMENSION(:),allocatable::comm_rate
    REAL:: rate,crate_l,crate_g


    i = 0
    k = 0
    DO level = 1,LEVEL_NUM
       n= HIERARCHICAL_DATA(level) % N
       i=i+n + HIERARCHICAL_DATA(level)%INU(n)
       i=i + HIERARCHICAL_DATA(level)%INL(n)
       k = k + n
    END DO

    j = 0 
    n=HIERARCHICAL_DATA(1)%N
    if(n>0) then
       j=n+HIERARCHICAL_DATA(1)%INU(n)+HIERARCHICAL_DATA(1)%INL(n)
    end if


    rate=dble(i)/j
    write(*,*) "nonzeros in all level:", i, "rate of level:",rate
    rate=dble(k)/n
    write(*,*) "dof in all level:", k, "rate of level:",rate

  END SUBROUTINE single_matrix_counting

END MODULE  AMG_cycle
