#ifndef __AMG_define_struct_H
#define __AMG_define_struct_H

#include "mySparseMatrix.h"
#include "myCUDA.h"

enum SMOOTHER_TYPE{JC, GS, JCGS, IDR, LUD, CG, SMOOTHER_LENGTH};

class color_object{
public:
	int length;
	int colorNum;
	int *color;
	int *histogram;

	color_object()
	{
		color     = NULL;
		histogram = NULL;
	}

	~color_object()
	{
		if(color    ) CUDA_FREE_HOST(color    );
		if(histogram) CUDA_FREE_HOST(histogram);
	}
};


class level_object{
public:
	color_object cl_obj;
	int *ordering;

	lu_crs a;
	crs r;
	crs p;

	double *l;
	double *u;

	lu_icrss da;

	icrss dr;
	icrss dp;

	int n;
	int np;
	double *db;
	double *dx;
	double *de;

	int smootherType;
	int numberOfTrials;
	int colorMax;
	double threshold;
	double omega;

	void (*Smoother)(level_object*);

	level_object()
	{
		ordering = NULL;

		l = NULL;
		u = NULL;

		db = NULL;
		dx = NULL;
		de = NULL;
	}

	~level_object()
	{
		da.clear();
		dr.clear();
		dp.clear();


		if(ordering) CUDA_FREE_HOST(ordering);

		if(l) CUDA_FREE_HOST(l);
		if(u) CUDA_FREE_HOST(u);
	}
};
typedef std::vector<level_object> level_objects;


enum ww{R, RT, P, PT, S, ST, T, V, WW_LENGTH };

class object{
public:
	int levelNum;
	level_objects level;

	double *dww[WW_LENGTH];

	double first_residual;
	double residual;
	double *threshold;
	double *restol;
	int rowsize;

	bool (*Terminal)(object *obj);

	object()
	{
		int i;
		restol    = NULL;
		threshold = NULL;

		for(i=0; i<WW_LENGTH; i++){
			dww[i] = NULL;
		}
	}

	~object()
	{
		int i;
		level.clear();

		if(threshold && threshold != &this->residual) CUDA_FREE(threshold);

		for(i=0; i<WW_LENGTH; i++){
			if(dww[i]) CUDA_FREE(dww[i]);
		}
	}
};


#endif /* __AMG_define_struct_H */
