!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

module ATmodule
  use solver_parameter, only: parameter_type
  include 'precision.inc'

  type(parameter_type), save :: parameters(100)
  integer(kind=kint) :: size

contains
  
  subroutine ATChooseParam(tv, solver_setup_fl, expected_call_times, chosenparam)
    implicit none
    real(kind=kreal), intent(in) :: tv
    logical, intent(in) :: solver_setup_fl
    integer(kind=kint), intent(in) :: expected_call_times
    type(parameter_type), intent(out) :: chosenparam

    integer(kind=kint) :: i
    logical, save :: initfl = .true.

    chosenparam % cycle_kind = 1
    chosenparam % number_of_vectors = 5
    chosenparam % iteration_number_up = 1
    chosenparam % iteration_number_down = 1
    chosenparam % iteration_number_bottom = 30
    chosenparam % f_first_iteration_number = 1
    chosenparam % method = 22
    chosenparam % gs_kind = 0
    chosenparam % theta = 0.05
    chosenparam % dump_jacobi_weight = 0.666
    chosenparam % dump_gauss_seidel_weight = 1.0
    chosenparam % solver_string = "CG"
    
  end subroutine ATChooseParam

  subroutine ATReportCost(setup_time, iter_time, iter, tv, reach_conv, solver_setup_fl)
    implicit none
    integer(kind=kint) :: iter 
    real(kind=kreal) :: tv, setup_time, iter_time
    logical :: solver_setup_fl, reach_conv

    write(*,*) "time", setup_time, iter_time 

  end subroutine ATReportCost
end module ATmodule
