!C C bindings for AMGS subroutines
MODULE amgs_c_interface_module
  use amgs_module
  use, intrinsic :: iso_c_binding
  implicit none

contains
  subroutine amgs_c_parameter_set() bind(C)
    call amgs_parameter_set(10)
  end subroutine amgs_c_parameter_set

  subroutine amgs_c_parameter_option_file() bind(C)
    call amgs_parameter_option_file(10, "AMGS_OPTION.dat")
  end subroutine amgs_c_parameter_option_file
  
  subroutine amgs_c_option_info(op) bind(C)
    integer (c_int), intent(in), value :: op
    
    call amgs_option_info(op)
    
  end subroutine amgs_c_option_info


  subroutine amgs_c_store_problem_matrix(crs_index, crs_column, crs_val, rowsize, nnz) bind(C)
    type (c_ptr), intent(in), value :: crs_val, crs_column, crs_index
    integer (c_int), intent(in), value :: rowsize, nnz
    integer i

    real (c_double), pointer :: f_val(:)
    integer (c_int), pointer :: f_index(:), f_column(:)
    call c_f_pointer(crs_index, f_index, [rowsize+1])
    call c_f_pointer(crs_column, f_column, [nnz])
    call c_f_pointer(crs_val, f_val, [nnz])
    
    f_column = f_column + 1
    call amgs_store_problem_matrix(f_index, f_column, f_val, rowsize)
    f_column = f_column - 1
    
  end subroutine amgs_c_store_problem_matrix

  subroutine amgs_c_terminal_val_set(reltol) bind(C)
    real (c_double), intent(in), value :: reltol
    call amgs_terminal_val_set(reltol)
  end subroutine amgs_c_terminal_val_set

  subroutine amgs_c_solver(x, RHS, rowsize, iter, setup_flag) bind(C)
    real (c_double), dimension(*), intent(inout) :: x, RHS
    integer (c_int), intent(in), value :: rowsize
    integer (c_int), intent(inout) :: iter
    real (c_double), dimension(1:rowsize) :: b, y
    logical, intent(in), value :: setup_flag
    ! integer i
    ! in order to have diagonal values greater than 0.
    ! RHS(1:rowsize) = -RHS(1:rowsize)
    call amgs_solver(x(1:rowsize), RHS(1:rowsize), iter, setup_flag)
    ! do i = 1, rowsize
    !    b(i) = RHS(i)
    ! end do
    ! call amgs_solver(y, b, iter, .FALSE.)
    ! do i = 1, rowsize
    !    x(i) = y(i)
    ! end do
  end subroutine amgs_c_solver

  subroutine amgs_c_clear_matrices() bind(C)
    call amgs_clear_matrices()
  end subroutine amgs_c_clear_matrices
  


  
end module amgs_c_interface_module
