!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif

MODULE data_creation_stuben
CONTAINS

  SUBROUTINE setup_phase()
    use problem_matrix
    CALL data_creation(NP, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, LEVEL_NUM) 
  END SUBROUTINE setup_phase

  subroutine clear_matrices()
    use problem_matrix
    
    NULL_SPACE_CONSTANT = .false.
    
    call clear_multi_lev_matrix(LEVEL_NUM)
    
    if(associated(D)) then
       deallocate(INL,INU,IAU,IAL,AU,AL,D)
    end if
  end subroutine clear_matrices

  
  SUBROUTINE data_creation(NP, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, LEVEL_NUM)
    USE data_structure_for_AMG
    USE AMG_cycle
    USE sgs_relaxation
    USE ilu_relaxation
    USE amg_parameter, only: theta, method, dump_jacobi_weight, MAX_LEVEL_SIZE, &
         & MIN_NODE_SIZE, theta_reduction_rate
    
    IMPLICIT NONE
    INTEGER(kind=kint), INTENT(in)  :: NP, NPL, NPU
    INTEGER(kind=kint), INTENT(out) :: LEVEL_NUM
!!$    REAL   (kind=kreal),INTENT(inout),TARGET :: D(NP )
!!$    REAL   (kind=kreal),INTENT(inout),TARGET :: AU(NPU)
!!$    REAL   (kind=kreal),INTENT(inout),TARGET :: AL(NPL)
!!$    INTEGER(kind=kint ),INTENT(in),   TARGET :: IAU(NPU)
!!$    INTEGER(kind=kint ),INTENT(in),   TARGET :: IAL(NPL)
!!$    INTEGER(kind=kint ),INTENT(in),   TARGET :: INU(0:NP), INL(0:NP)

    INTEGER(kind=kint ),INTENT(in),   TARGET :: IAU(:)
    INTEGER(kind=kint ),INTENT(in),   TARGET :: IAL(:)
    INTEGER(kind=kint ),INTENT(in),   TARGET :: INU(0:), INL(0:)
    REAL   (kind=kreal),INTENT(inout),TARGET :: D(:)
    REAL   (kind=kreal),INTENT(inout),TARGET :: AU(:)
    REAL   (kind=kreal),INTENT(inout),TARGET :: AL(:)

    !C pointers for expanding the plus non-zero elements
    REAL   (kind=kreal), POINTER ::  BD(:)
    REAL   (kind=kreal), POINTER ::  BAU(:)
    REAL   (kind=kreal), POINTER ::  BAL(:)
    INTEGER(kind=kint ), POINTER ::  BINU(:)
    INTEGER(kind=kint ), POINTER ::  BIAU(:)
    INTEGER(kind=kint ), POINTER ::  BINL(:)
    INTEGER(kind=kint ), POINTER ::  BIAL(:)

    !C pointers for adaptive amg cycle
    REAL   (kind=kreal), POINTER ::  PX(:), coarser_PX(:)
    REAL   (kind=kreal), POINTER ::  PD(:)
    REAL   (kind=kreal), POINTER ::  PAU(:)
    REAL   (kind=kreal), POINTER ::  PAL(:)
    INTEGER(kind=kint ), POINTER ::  PINU(:)
    INTEGER(kind=kint ), POINTER ::  PIAU(:)
    INTEGER(kind=kint ), POINTER ::  PINL(:)
    INTEGER(kind=kint ), POINTER ::  PIAL(:)
    INTEGER(kind=kint ) :: nu1, nu2, adaptive_cycle, max_adaptive_cycle 
    REAL   (kind=kreal) :: e_nrm, e_nrm2


    INTEGER(kind=kint ), POINTER :: NI(:), PNI(:), NIT(:), PNIT(:)
    INTEGER(kind=kint ), ALLOCATABLE :: node_index(:)

    INTEGER(kind=kint)  :: coarser_level_size, level, finer_level_size, count
    INTEGER(kind=kint)  :: finer_level_NP, i, j, k, is, ie, col
    INTEGER(kind=kint)  :: finer_level_NPL, finer_level_NPU
    REAL(kind=kreal)    :: v
    
    LOGICAL            :: finish_flag, global_finish_flag
    
    TYPE(INTER_LEVEL_OPERATOR),POINTER :: P,R

    REAL   (kind=kreal) :: mu, c_rate

    !C SYM_FLAG == 1: symmetric mat, 0: unsymmetric mat
    INTEGER(kind=kint) :: SYM_FLAG = 0
    LOGICAL :: ac_flag
    
    LOGICAL :: AGGRESSIVE_COARSENING_FL      = .false.
    LOGICAL :: FULL_AGGRESSIVE_COARSENING_FL = .false.

    LOGICAL :: DIRECT_INTERPOLATION_FL  = .false.
    LOGICAL :: CLASSICAL_INTERPOLATION_FF_FL  = .false.
    LOGICAL :: STANDARD_INTERPOLATION_FF_FL  = .false.
    LOGICAL :: POSITIVE_COUPLING_FL = .false.
    LOGICAL :: POSITIVE_EXPAND_FL = .false.
    LOGICAL :: ADAPTIVE_AMG_FL = .false.

    LOGICAL :: JACOBI_RELAXATION_FL = .false.
    
    LOGICAL :: V_CHECK_FL = .false.
    LOGICAL :: CR_CHECK_FL = .false.
    LOGICAL :: THETA_TUNING_FL = .false.
    LOGICAL :: RESET_METHOD = .false.
    LOGICAL :: SWITCH_FL = .false.

    LOGICAL :: ADD_NEW_C_FLAG = .false.
    
    LOGICAL :: AD_EXIT = .false.
    LOGICAL :: SA_FL = .false.


    REAL(kind=kreal), pointer ::  ILU_D(:), ILU_AL(:), ILU_AU(:)

    REAL  :: rv, switch_v

    INTEGER :: SCORE = 2
    REAL   (kind=kreal) :: density
    REAL   (kind=kreal) :: epsilon= 0.20
    REAL   (kind=kreal) :: plus   = 0.50
    REAL   (kind=kreal) :: pepsilon = 0.50
    REAL   (kind=kreal) :: NONZERO_THRSLD = 1.0E-5

    REAL   (kind=kreal) :: theta_copy
    
    INTEGER(kind=kint ) :: nonzero_a, nonzero_1


!!$    theta  = 0.25

    !C-- theta means the critierion for strong connection.
!    method = 8
!!$    method = 0
    if(method == 0) then       !C amg SNSN
    else if(method == 1) then
       CLASSICAL_INTERPOLATION_FF_FL = .true. !C amg_CL_FF
    else if(method == -1) then
       STANDARD_INTERPOLATION_FF_FL = .true. !C amg_ST_FF
    else if(method == -2) then !C amg_AD_FF
       ADAPTIVE_AMG_FL = .true. 
    else if(method == 2) then  !C amg_fj
       JACOBI_RELAXATION_FL = .true.
    else if(method == 3) then  !C amg_posi 
       POSITIVE_COUPLING_FL = .true.
    else if(method == 4) then  !C amg_posi_fj
       POSITIVE_COUPLING_FL = .true.
       JACOBI_RELAXATION_FL = .true.
    else if(method == 5) then  !C amg_a1
       AGGRESSIVE_COARSENING_FL = .true.
       SCORE = 1
    else if(method == 6) then !C amg_a1_fj
       AGGRESSIVE_COARSENING_FL      = .true.
       JACOBI_RELAXATION_FL = .true.
       SCORE = 1
    else if(method == 7) then !C amg_a1_dc
       AGGRESSIVE_COARSENING_FL = .true.
       DIRECT_INTERPOLATION_FL  = .true.
       SCORE = 1
    else if(method == 8) then !C amg_a1_dc_fj
       AGGRESSIVE_COARSENING_FL      = .true.
       DIRECT_INTERPOLATION_FL  = .true.
       JACOBI_RELAXATION_FL = .true.
       SCORE = 1
    else if(method == 9) then !C amg_a2  A2NSN
       AGGRESSIVE_COARSENING_FL      = .true.
    else if(method == 10) then !C amg_a2_fj 
       AGGRESSIVE_COARSENING_FL      = .true.
       JACOBI_RELAXATION_FL = .true.       
    else if(method == 11) then !C amg_a2_dc 
       AGGRESSIVE_COARSENING_FL      = .true.
       DIRECT_INTERPOLATION_FL  = .true.
    else if(method == 12) then !C amg_a2_dc_fj A2NDJ
       AGGRESSIVE_COARSENING_FL = .true.
       DIRECT_INTERPOLATION_FL  = .true.
       JACOBI_RELAXATION_FL     = .true.
    else if(method == 13) then !C amg_dc
       DIRECT_INTERPOLATION_FL  = .true.
    else if(method == 14) then !C amg_dc_fj SNDJ
       DIRECT_INTERPOLATION_FL  = .true.
       JACOBI_RELAXATION_FL = .true.
    else if(method == 15) then !C amg_dc_posi
       DIRECT_INTERPOLATION_FL  = .true.
       POSITIVE_COUPLING_FL = .true.
    else if(method == 16) then !C amg_dc_posi_fj SYDJ
       DIRECT_INTERPOLATION_FL  = .true.
       POSITIVE_COUPLING_FL = .true.
       JACOBI_RELAXATION_FL = .true.
    else if(method == 17) then !C amg_fa1
       FULL_AGGRESSIVE_COARSENING_FL = .true.
       SCORE = 1
    else if(method == 18) then !C amg_fa2
       FULL_AGGRESSIVE_COARSENING_FL = .true.
       
    else if(method == 19) then !C amg_dc_posi2
       POSITIVE_EXPAND_FL = .true.
       DIRECT_INTERPOLATION_FL  = .true.

    else if(method == 20) then !C amg_posi2
       POSITIVE_EXPAND_FL = .true.

    else if(method == 21) then !C switch from A2NSN
       AGGRESSIVE_COARSENING_FL      = .true. 
       CR_CHECK_FL = .true.
       SWITCH_FL = .true.
       switch_v =0.4
    else if(method == 22) then !C SAAMG
       SA_FL      = .true. 
    end if

    HIERARCHICAL_DATA(1) % N   = NP
    HIERARCHICAL_DATA(1) % NP  = NP
    HIERARCHICAL_DATA(1) % NPL = NPL
    HIERARCHICAL_DATA(1) % NPU = NPU
    HIERARCHICAL_DATA(1) % D   => D
    HIERARCHICAL_DATA(1) % INL => INL
    HIERARCHICAL_DATA(1) % INU => INU 
    HIERARCHICAL_DATA(1) % IAL => IAL
    HIERARCHICAL_DATA(1) % IAU => IAU
    HIERARCHICAL_DATA(1) % AL  => AL
    HIERARCHICAL_DATA(1) % AU  => AU
    
    if(ADAPTIVE_AMG_FL) then
       e_nrm = 0
       !C allocate and initialize prototype vector in X(:)
       allocate(HIERARCHICAL_DATA(1) % X(NP))

       call random_seed()

       DO i = 1, NP
          call RANDOM_NUMBER(v)
!!$          v = 0.5
!!$          v=1-0.5*v
          HIERARCHICAL_DATA(1) % X(i) = v
          e_nrm = e_nrm +v*v
       END DO

       !C initial error's  2-norm
       e_nrm = log(e_nrm)/2
       
       nu1 = 3
       nu2 = 3
    end if

    max_adaptive_cycle = 2
    DO adaptive_cycle = 1, max_adaptive_cycle
       
       if(RESET_METHOD) then

          RESET_METHOD = .false.

          !C This structure of Do loop does'nt allow to change to Adaptive AMG. 
!!$          call clear_coarse_matrices(LEVEL_NUM) 
          
       end if
          
       
       if(ADAPTIVE_AMG_FL .and. adaptive_cycle > 1) then
          
          !C e_nrm2 stores 2-norm of error vector after adaptive cycle
          e_nrm2 = 0
          DO i = 1, HIERARCHICAL_DATA(1) % N
             e_nrm2 = e_nrm2 + HIERARCHICAL_DATA(1) % X(i)**2
          END DO
          write(*,*) e_nrm2
          e_nrm2 = log(e_nrm2)/2
          !C when error is reduced enough, then stop the adaptive cycle
          if(e_nrm2-e_nrm < -4) then
             write(*,*) "Stop adaptive cycle."
!             write(*,*) "Stop adaptive cycle. log( Error reduction of 2-norm ):", e_nrm2, e_nrm
             exit
          end if
          
          !C clear coarse matrices in HIERARCHICAL_DATA(:)
          call clear_coarse_matrices(LEVEL_NUM)          
       end if


       theta_copy = theta
       !C coarser levels are created..
       DO level = 2, MAX_LEVEL_SIZE

          if(SA_FL) then
             call create_interpolation_SA()
          else
             AD_EXIT = .false.
             call create_interpolation_STUBEN()
             if(AD_EXIT) exit
          end if
          
          theta_copy=theta_copy*theta_reduction_rate
          

          if(JACOBI_RELAXATION_FL .and. .not. ac_flag) then
             write(*,*) "jacobi relaxation"
             call jacobi_relaxation(level, coarser_level_size, node_index, epsilon)
          end if

          call transpose_P_in_R(level, coarser_level_size)


          if(POSITIVE_EXPAND_FL) then
             !C-----------
             call get_back_expand_plus(BD, BINL, BINU, BIAL, BIAU, BAL, BAU, &
                  & level)
          end if

          call RAP(level, 5)
          

          if(ADAPTIVE_AMG_FL) then
             coarser_PX => HIERARCHICAL_DATA(level) % X

             !C restrict only C-point values: PX -> coarser_PX
             DO i = 1, finer_level_size
                is = node_index(i)
                if(is >0 ) then
                   coarser_PX(is) = PX(i)
                end if
             END DO
!             write(*,*) level, ":", coarser_PX(1:10) 
          end if

          
          deallocate(node_index)
          
          global_finish_flag = (coarser_level_size < MIN_NODE_SIZE)
          IF(global_finish_flag) EXIT
          
       END DO

       if(V_CHECK_FL .and. adaptive_cycle == 1)then
          i=20
          call v_check(i, level, mu)
          write(*,*) "VC2 ## estimated CR convergence rate: ", mu
          write(*,*) "VC2 ## aggressive coarsening flag:    ", ac_flag
       end if



       
       LEVEL_NUM = level       
       if(level > MAX_LEVEL_SIZE) LEVEL_NUM = MAX_LEVEL_SIZE
       

       if(RESET_METHOD) cycle
       if(.not.ADAPTIVE_AMG_FL .or. adaptive_cycle == max_adaptive_cycle) EXIT

       !C ------- only ADAPTIVE AMG executes the following part

       !C relaxation nu1+nu2 times at the coarsest level
       PINU => HIERARCHICAL_DATA(LEVEL_NUM) % INU 
       PINL => HIERARCHICAL_DATA(LEVEL_NUM) % INL 
       PD   => HIERARCHICAL_DATA(LEVEL_NUM) % D
       PIAU => HIERARCHICAL_DATA(LEVEL_NUM) % IAU
       PIAL => HIERARCHICAL_DATA(LEVEL_NUM) % IAL
       PAU  => HIERARCHICAL_DATA(LEVEL_NUM) % AU
       PAL  => HIERARCHICAL_DATA(LEVEL_NUM) % AL
       PX   => HIERARCHICAL_DATA(LEVEL_NUM) % X
       finer_level_size = HIERARCHICAL_DATA(LEVEL_NUM) % NP
       
       call sgs_zero(finer_level_size, finer_level_size, PD, PAL, PINL, &
            & PIAL, PAU, PINU, PIAU, PX, nu1+nu2) 
       
       DO level = LEVEL_NUM-1, 1, -1

          !C interpolate level+1 -> level : coarser_PX -> PX 
          PINU       => HIERARCHICAL_DATA(level)   % INU 
          PINL       => HIERARCHICAL_DATA(level)   % INL 
          PD         => HIERARCHICAL_DATA(level)   % D
          PIAU       => HIERARCHICAL_DATA(level)   % IAU
          PIAL       => HIERARCHICAL_DATA(level)   % IAL
          PAU        => HIERARCHICAL_DATA(level)   % AU
          PAL        => HIERARCHICAL_DATA(level)   % AL
          PX         => HIERARCHICAL_DATA(level)   % X
          coarser_PX => HIERARCHICAL_DATA(level+1) % X
          P          => HIERARCHICAL_DATA(level+1) % P
          finer_level_size = HIERARCHICAL_DATA(level) % NP

          DO i = 1, finer_level_size
             PX(i) = 0
             is = P % IN(i-1)+1
             ie = P % IN(i)
             DO j = is, ie
                col = P % CN(j)
                PX(i) = PX(i) + P % V(j) * coarser_PX(col)
             END DO
          END DO

!          write(*,*) level, ":", coarser_PX(1:10) 

          !C relaxation nu2 times
          call sgs_zero(finer_level_size, finer_level_size, PD, PAL, PINL, PIAL, PAU, PINU, PIAU, PX, nu2)
          
       END DO
       
    END DO

    if(ILU0_SMOOTH_FL) then
       DO level = 1, LEVEL_NUM
          !C allocate and copy the ILU0
          finer_level_size = HIERARCHICAL_DATA(level)   % N
          finer_level_NPL = HIERARCHICAL_DATA(level)   % NPL
          finer_level_NPU = HIERARCHICAL_DATA(level)   % NPU
          finer_level_NPU = HIERARCHICAL_DATA(level)   % NPU

          allocate(HIERARCHICAL_DATA(level) % ILU0_D(finer_level_size), &
               &   HIERARCHICAL_DATA(level) % ILU0_AL(finer_level_NPL), &
               &   HIERARCHICAL_DATA(level) % ILU0_AU(finer_level_NPU))
          
          ILU_D  => HIERARCHICAL_DATA(level) % ILU0_D
          ILU_AL => HIERARCHICAL_DATA(level) % ILU0_AL
          ILU_AU => HIERARCHICAL_DATA(level) % ILU0_AU
          PINU => HIERARCHICAL_DATA(level) % INU 
          PINL => HIERARCHICAL_DATA(level) % INL 
          PIAU => HIERARCHICAL_DATA(level) % IAU 
          PIAL => HIERARCHICAL_DATA(level) % IAL 
          
          ILU_D(:)  = HIERARCHICAL_DATA(level) % D(:)
          ILU_AL(:) = HIERARCHICAL_DATA(level) % AL(:)
          ILU_AU(:) = HIERARCHICAL_DATA(level) % AU(:)
          write(*,*) "level", level
          call ilu0_setup(finer_level_size, ILU_D, ILU_AL, PINL, PIAL, ILU_AU, PINU, PIAU)
          
       END DO
    end if
    
#ifdef PRINT_REZ
    nonzero_1 = NPL+NPU+NP
    nonzero_a  = 0
    DO i = 1, LEVEL_NUM
       nonzero_a = nonzero_a + HIERARCHICAL_DATA(i) % NPL + &
            &      HIERARCHICAL_DATA(i) % NPU + &
            &      HIERARCHICAL_DATA(i) % NP
    END DO
    write(*,*) "nonzero element complexity: ",nonzero_a/dble(nonzero_1)
    
#endif    
    
    if(ADAPTIVE_AMG_FL) then
       deallocate(HIERARCHICAL_DATA(1) % X)
    end if

  contains 
    subroutine create_interpolation_STUBEN()
      
      if(POSITIVE_EXPAND_FL) then
         !C-----------       
         call expand_plus_elements(BD, BINL, BINU, BIAL, BIAU, BAL, BAU, &
              &    level, pepsilon)
      end if

      finer_level_size = HIERARCHICAL_DATA(level - 1) % N
      finer_level_NP   = HIERARCHICAL_DATA(level - 1) % NP
      finer_level_NPL  = HIERARCHICAL_DATA(level - 1) % NPL
      finer_level_NPU  = HIERARCHICAL_DATA(level - 1) % NPU       

      if(ADAPTIVE_AMG_FL) then
         PINU => HIERARCHICAL_DATA(level - 1) % INU 
         PINL => HIERARCHICAL_DATA(level - 1) % INL 
         PD   => HIERARCHICAL_DATA(level - 1) % D
         PIAU => HIERARCHICAL_DATA(level - 1) % IAU
         PIAL => HIERARCHICAL_DATA(level - 1) % IAL
         PAU  => HIERARCHICAL_DATA(level - 1) % AU
         PAL  => HIERARCHICAL_DATA(level - 1) % AL
         PX   => HIERARCHICAL_DATA(level - 1) % X
         !C prototype smoothing
         !C relaxation nu1 times at each level
         call sgs_zero(finer_level_size, finer_level_NP, PD, &
              & PAL, PINL, PIAL, PAU, PINU, PIAU, PX, nu1) 
      end if


      if(THETA_TUNING_FL) then
         density = DBLE(finer_level_NPU+finer_level_NPL+finer_level_size)/finer_level_size/finer_level_size
         theta_copy = (EXP(-(3*density-1)**2)+3*density)/(3+density)
         write(*,*) "theta:",theta_copy, "densitiy:",density , "at",level
      end if
      
      ALLOCATE(NI(finer_level_NPL + finer_level_NPU))
      ALLOCATE(PNI(0:2*finer_level_size))

      !C to be changed for positive couplings..
      ! write(*,*) "creating Ni"       
      CALL stu_neighbors(PNI, NI, theta_copy, level)

      ALLOCATE(PNIT(0:2*finer_level_size))
      j = PNI(2*finer_level_size)
      ALLOCATE(NIT(j))

      ! write(*,*) "transpose Ni"
      call transpose_ni(finer_level_size, PNI, NI, PNIT, NIT)

      !C step 1
      ALLOCATE(node_index(finer_level_size))
      node_index = 0
      ! write(*,*) "coloring"
      call coloring(finer_level_size, PNIT, NIT, PNI, NI, node_index, coarser_level_size)
      !C node_index has 3 values -2, -1, 0. they are C-point, F-point, other points. 
      !C Those points influences no points.

      if(CLASSICAL_INTERPOLATION_FF_FL .or. ADAPTIVE_AMG_FL) then
         call f_f_check(level, node_index, coarser_level_size, NONZERO_THRSLD)          
!!$          call strong_f_f_check(level, node_index, coarser_level_size, PNI, NI, PNIT, NIT)
      end if

      if(STANDARD_INTERPOLATION_FF_FL) then
         call strong_f_f_check(level, node_index, coarser_level_size, PNI, NI, PNIT, NIT)
      end if

      ac_flag = .false.       
      if(coarser_level_size > 1 .and. FULL_AGGRESSIVE_COARSENING_FL .or. &
           & coarser_level_size > 1 .and. level == 2 .and. AGGRESSIVE_COARSENING_FL) then
         !C when there is only 1 C-point, aggressive_coarsening leaves no C-points.
         !C step 3
         ! write(*,*) "aggressive coarsening"
         call aggressive_coasening(finer_level_size, PNIT, NIT, PNI, NI, node_index, &
              & coarser_level_size, SCORE)
         ac_flag = .true.
      end if

      !C step 2 
      if(POSITIVE_COUPLING_FL) then
         ! write(*,*) "posi_couple_check"
         call posi_couple_check(plus, PNI, NI, PNIT, NIT, node_index, &
              & coarser_level_size, level)
      end if


      !C node_index is changed to store c-points' number. F-point is checked as 0.
      j = 0
      DO i = 1, finer_level_size
         if(node_index(i) == -2) then
            j = j+1
            node_index(i) = j
         else if(node_index(i) /= -1) then
            node_index(i) = 0
          end if
      END DO
      
      if(level == 2 .and. CR_CHECK_FL .and. adaptive_cycle == 1) then

         i=10
         call sgs_cr(i, level-1, node_index, mu)
         c_rate = REAL(coarser_level_size)/finer_level_size
         write(*,*) "CR ## estimated CR convergence rate: ", mu
         write(*,*) "CR ## C-point rate:                  ", c_rate
         write(*,*) "CR ## aggressive coarsening flag:    ", ac_flag

         if(SWITCH_FL .and. mu > switch_v) then
            !write(*,*) "SWITCHING to ST_AT"
            RESET_METHOD = .true.

            CR_CHECK_FL = .false. 
            AGGRESSIVE_COARSENING_FL = .false.    !C from  A2NSN
            STANDARD_INTERPOLATION_FF_FL = .true. !C to amg_ST_FF_AT
            THETA_TUNING_FL = .false. !C without automatic tuning

            deallocate(node_index, NI, PNI, NIT, PNIT)
            AD_EXIT=.true.
            return
         end if
      end if

      if(ac_flag) then
         !write(*,*) "multi-path interpolation"
         !C to be changed for positive couplings..
         call multipath_interpolation(level, PNIT, NIT, PNI, NI, node_index, coarser_level_size)
         !write(*,*) "jacobi relaxation"
         call jacobi_relaxation(level, coarser_level_size, node_index, epsilon)
      else if(DIRECT_INTERPOLATION_FL) then
         !write(*,*) "direct interpolation"
         !C to be changed for positive couplings..
         call direct_interpolation(level, node_index)
      else if(CLASSICAL_INTERPOLATION_FF_FL) then
         !write(*,*) "classical interpolation FF"
         call classical_interpolation_FF(level, PNI, NI, node_index, coarser_level_size, NONZERO_THRSLD)
      else if(STANDARD_INTERPOLATION_FF_FL) then
         !write(*,*) "standard interpolation FF"
         ADD_NEW_C_FLAG = .false.
         call standard_interpolation_FF(level, PNI, NI, node_index, coarser_level_size, ADD_NEW_C_FLAG)

         !C add new C points which has a problematic F-F connnection.
         if(ADD_NEW_C_FLAG) then
            ADD_NEW_C_FLAG = .false.

            j = 0
            DO i = 1, finer_level_size
               if(node_index(i) == -2 .or. node_index(i) > 0) then
                  j = j+1
                  node_index(i) = j
               else if(node_index(i) /= -1) then
                  node_index(i) = 0
               end if
            END DO
            coarser_level_size = j

            !write(*,*) "2nd standard interpolation FF"
            call standard_interpolation_FF(level, PNI, NI, node_index, coarser_level_size, ADD_NEW_C_FLAG)
            if(ADD_NEW_C_FLAG) stop 'there is problematic F-F connection. in standard_interpolation_FF()'

         end if
         
      else if(ADAPTIVE_AMG_FL) then
         !write(*,*) "adaptive interpolation FF"
         ADD_NEW_C_FLAG = .false.
         call adaptive_interpolation_FF(level, PNI, NI, node_index, coarser_level_size, NONZERO_THRSLD, &
              & PX, ADD_NEW_C_FLAG)

         !C add new C points which has a problematic F-F connnection.
         if(ADD_NEW_C_FLAG) then
            ADD_NEW_C_FLAG = .false.

            j = 0
            DO i = 1, finer_level_size
               if(node_index(i) == -2 .or. node_index(i) > 0) then
                  j = j+1
                  node_index(i) = j
               else if(node_index(i) /= -1) then
                  node_index(i) = 0
               end if
            END DO
            coarser_level_size = j

            !write(*,*) "2nd adaptive interpolation FF"
            call adaptive_interpolation_FF(level, PNI, NI, node_index, coarser_level_size, NONZERO_THRSLD, &
                 & PX, ADD_NEW_C_FLAG)

            if(ADD_NEW_C_FLAG) stop 'there is problematic F-F connection. in adaptive_interpolation_FF()'
         end if

      else
         !write(*,*) "standard interpolation"
         !C to be changed for positive couplings..
         call standard_interpolation(level, PNI, NI, node_index, epsilon)
      end if
      
      DEALLOCATE(NI,PNI,PNIT,NIT)

    end subroutine create_interpolation_STUBEN


    !C output data : coarser_level_size, P % IN, P % CN, P % V, P % ROW_SIZE
    subroutine create_interpolation_SA()
      REAL(kind=kreal), ALLOCATABLE :: tilde_D(:)
      INTEGER(kind=kint), POINTER   :: in_aggregates_result(:)
      INTEGER(kind=kint), POINTER   :: aggregates_result(:)
      INTEGER(kind=kint)            :: in_aggregates_result_size

      TYPE(row_node),     POINTER   :: Temp_P(:,:)
      INTEGER(kind=kint), POINTER   :: Temp_P_row_size(:)
      INTEGER(kind=kint)            :: RATE_OF_SPACE
      
      finer_level_size = HIERARCHICAL_DATA(level - 1) % N
      finer_level_NP   = HIERARCHICAL_DATA(level - 1) % NP
      finer_level_NPL  = HIERARCHICAL_DATA(level - 1) % NPL
      finer_level_NPU  = HIERARCHICAL_DATA(level - 1) % NPU       


      ALLOCATE(NI(finer_level_NPL + finer_level_NPU))
      ALLOCATE(PNI(0:2*finer_level_NP))

      ALLOCATE(node_index(finer_level_NP))
      ALLOCATE(tilde_D   (finer_level_NP))

      if(SYM_FLAG == 1) then
         CALL neighbors(PNI, NI, theta_copy, level, node_index, tilde_D)
      else
         CALL neighbors_unsym(PNI, NI, theta_copy, level, node_index, tilde_D)
      end if

      CALL indpdt_agrgt(PNI, NI, level, node_index, in_aggregates_result,   &
           & aggregates_result, in_aggregates_result_size)


      coarser_level_size = in_aggregates_result_size
      global_finish_flag = (coarser_level_size < MIN_NODE_SIZE)


      !C temporary work area for inter-level matrix P is allocated
      !C (RATE_OF_SPACE+0.5)*finer_level_NP*8 byte
      RATE_OF_SPACE = 30 + 80 * (level - 2)

      IF(RATE_OF_SPACE > coarser_level_size) RATE_OF_SPACE = coarser_level_size


      ALLOCATE(Temp_P(RATE_OF_SPACE, finer_level_NP))
      ALLOCATE(Temp_P_row_size(finer_level_NP))

      CALL smooth_aggregate(PNI, NI, dump_jacobi_weight, level,      &
           & in_aggregates_result_size, in_aggregates_result,        &
           & aggregates_result, Temp_P, Temp_P_row_size, tilde_D,    &
           & RATE_OF_SPACE)

      CALL agrgt_dealloc (in_aggregates_result, aggregates_result)
      DEALLOCATE(NI)
      DEALLOCATE(PNI)
      DEALLOCATE(tilde_D)

      !C for single CPU case
      coarser_level_size = in_aggregates_result_size

      !--C check Temp_P
      DO i=1,finer_level_NP
         IF(Temp_P_row_size(i) > RATE_OF_SPACE) THEN
            WRITE(*,*) "RATE_OF_SPACE should be enlarged "
            STOP 'The allocation error : Temp_P in make_rap'
         END IF
      END DO

      !C-- Temp_P is recorded in HIERARCHICAL_DATA()..
      ALLOCATE(HIERARCHICAL_DATA(level) % P)
      P => HIERARCHICAL_DATA(level) % P
      P % ROW_SIZE = HIERARCHICAL_DATA(level - 1) % NP
      k = P % ROW_SIZE
      ALLOCATE(P % IN(0:k)) 
      P % IN(0) = 0
      DO i = 1, k
         P % IN(i) = P % IN(i - 1) + Temp_P_row_size(i)
      END DO
      i = P % IN(k)

      ALLOCATE(HIERARCHICAL_DATA(level) % P % CN(i))
      ALLOCATE(HIERARCHICAL_DATA(level) % P % V(i))

      k = 1
      DO i = 1, P % ROW_SIZE
         DO j = 1, Temp_P_row_size(i)
            P % CN(k) = Temp_P(j,i) % column
            P % V(k)  = Temp_P(j,i) % value
            k = k + 1
         END DO
      END DO

      DEALLOCATE(Temp_P)
      DEALLOCATE(Temp_P_row_size)
      !C-- end:Temp_P is recorded in HIERARCHICAL_DATA()..

    end subroutine create_interpolation_SA
    
  END SUBROUTINE data_creation




!---SAAMG subroutines
  SUBROUTINE neighbors(PNI, NI, theta, LEVEL_NO, node_index, tilde_D)
    USE data_structure_for_AMG
    IMPLICIT NONE
    
    INTEGER(kind=kint ) :: N, NP
    REAL   (kind=kreal), INTENT(in):: theta
    
    INTEGER(kind=kint ), INTENT(out) :: NI  (:)
    INTEGER(kind=kint ), INTENT(out) :: PNI(0:)
    INTEGER(kind=kint ), INTENT(out) :: node_index(:)
    
    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO
    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)
    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)

    REAL(kind=kreal)  :: tilde_D(:)

    INTEGER(kind=kint) :: i,j,isU,ieU,isL,ieL,inod,nni,nni_before,size,count
    REAL   (kind=kreal):: t

    NP = HIERARCHICAL_DATA(LEVEL_NO-1) % NP     
    N  = HIERARCHICAL_DATA(LEVEL_NO-1) % N 
    
    D   => HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO-1) % AU
    
    PNI(0) = 0
    nni    = 0

    DO i = 1,NP
       tilde_D(i) = D(i)
       node_index(i) = 0
    END DO

    DO j = 1,NP

       isL = INL(j-1)+1
       ieL = INL(j  )
       
       DO i= isL, ieL
          inod= IAL(i)+ZERO_ORIGIN
          t= AL(i)

          IF(t*t > dabs(D(j)*D(inod)*theta*theta) .AND. (D(j)*t<0 .or. D(inod)*t<0)) THEN
             nni=nni+1
             NI(nni)=i
          ELSE
!!$             if(t*D(j)<0) then 
             tilde_D(j) = tilde_D(j)+t
!!$             end if
          ENDIF
       ENDDO
       PNI(j)=nni
    END DO

    DO j = 1,NP
       isU = INU(j-1)+1
       ieU = INU(j  )
       
       DO i = isU, ieU
          inod= IAU(i)+ZERO_ORIGIN
          t=AU(i)
          IF(t*t > dabs(D(j)*D(inod)*theta*theta) .AND. (D(inod)*t<0 .or. D(j)*t<0)) THEN
             nni=nni+1
             NI(nni)=i
          ELSE
!!$             if(t*D(j)<0) then 
             tilde_D(j)=tilde_D(j)+t
!!$             end if
          END IF
       ENDDO
       PNI(NP+j)=nni
    END DO

    DO j=1,N
       size=PNI(j)-PNI(j-1)+PNI(NP+j)-PNI(NP+j-1)
       IF(size==0 ) THEN
          node_index(j)=-1
       END IF
    ENDDO

  END SUBROUTINE neighbors

  !C this subroutine considers the case of unsymmetric problem matrix.
  !C But symmetricity of nonzero structure of the matrix is assumed.
  SUBROUTINE neighbors_unsym(PNI, NI, theta, LEVEL_NO, node_index, tilde_D)
    USE data_structure_for_AMG
    IMPLICIT NONE

    INTEGER(kind=kint ) :: N,NP
    REAL   (kind=kreal), INTENT(in):: theta
    
    INTEGER(kind=kint ), INTENT(out) :: NI  (:)
    INTEGER(kind=kint ), INTENT(out) :: PNI(0:)
    INTEGER(kind=kint ), INTENT(out) :: node_index(:)
    
    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO
    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)
    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)

    REAL(kind=kreal)  :: tilde_D(:)

    INTEGER(kind=kint), ALLOCATABLE :: check_U(:)

    INTEGER(kind=kint) :: i,j,k,isU,ieU,isL,ieL,inod,nni,nni_before,size,count
    INTEGER(kind=kint) :: NPU,NPL
    REAL   (kind=kreal):: tl,tu

    NP   = HIERARCHICAL_DATA(LEVEL_NO-1) % NP     
    N    = HIERARCHICAL_DATA(LEVEL_NO-1) % N 
    NPU  = HIERARCHICAL_DATA(LEVEL_NO-1) % NPU 
    NPL  = HIERARCHICAL_DATA(LEVEL_NO-1) % NPL 
    
    D   => HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO-1) % AU
    
    PNI(0) = 0
    nni    = 0
    
    ALLOCATE(check_U(NPU))
    check_U=0
    
    DO i = 1,NP
       tilde_D(i) = D(i)
       node_index(i) = 0
    END DO
    
    DO j = 1,NP
       isL = INL(j-1)+1
       ieL = INL(j  )

       DO i = isL, ieL
          inod = IAL(i)+ZERO_ORIGIN
          tl = AL(i)

          
          !C determining the cordinate position of upper matrix
          isU = INU(inod - 1) + 1; ieU = INU(inod)
          DO k= isU, ieU
             IF(IAU(k)+ZERO_ORIGIN == j) EXIT
          END DO
          IF(k > ieU) STOP "error in neighbor()"
          tu = AU(k)


          
          IF((tl * tl > dabs(D(j) * D(inod) * theta * theta).and. D(j)*tl<0 ) .OR. &
               & (tu * tu > dabs(D(j)*D(inod)*theta*theta).and. D(inod)*tu<0) ) THEN
             check_U(k) = 1
             nni = nni + 1
             NI(nni) = i
          ELSE
             tilde_D(j) = tilde_D(j) + tl
          ENDIF
       ENDDO
       PNI(j) = nni
    END DO
    
    DO j = 1, NP
       isU = INU(j - 1) + 1; ieU = INU(j)
       DO i = isU, ieU
          IF(check_U(i) == 1) THEN
             nni = nni + 1
             NI(nni) = i
          ELSE
             tilde_D(j) = tilde_D(j) + AU(i)
          END IF
       ENDDO
       PNI(NP + j) = nni
    ENDDO
    
    DEALLOCATE(check_U)

    DO j = 1, N
       size = PNI(j) - PNI(j - 1) + PNI(NP + j) - PNI(NP + j - 1)
       IF(size == 0) node_index(j) = -1
    END DO

  END SUBROUTINE neighbors_unsym

SUBROUTINE indpdt_agrgt (PNI,NI,LEVEL_NO,node_index,in_aggregates_result, &
     & aggregates_result,in_aggregates_result_size)
  USE data_structure_for_AMG
    USE queue_mod
    IMPLICIT NONE

    INTEGER(kind=kint), DIMENSION(: ), INTENT(in) :: NI
    INTEGER(kind=kint), DIMENSION(0:), INTENT(in) :: PNI
    INTEGER(kind=kint), INTENT(in)                :: LEVEL_NO
    INTEGER(kind=kint), DIMENSION(:), INTENT(inout) :: node_index
    INTEGER(kind=kint), DIMENSION(:),pointer :: in_aggregates_result
    INTEGER(kind=kint), DIMENSION(:),pointer :: aggregates_result
    INTEGER(kind=kint), INTENT(out) :: in_aggregates_result_size


    !C    TYPE(INTER_LEVEL_OPERATOR),POINTER :: P
    INTEGER(kind=kint), DIMENSION(:,:), ALLOCATABLE :: Temp
    
    INTEGER(kind=kint), DIMENSION(:,:), ALLOCATABLE :: P_queue
    INTEGER(kind=kint), DIMENSION(:),   ALLOCATABLE :: aggregate_flags

    INTEGER(kind=kint) :: head, tail, queue_size

    INTEGER(kind=kint ) :: N,NP
    REAL   (kind=kreal),POINTER, DIMENSION(:)  ::  D
    REAL   (kind=kreal),POINTER, DIMENSION(:)  ::  AU
    REAL   (kind=kreal),POINTER, DIMENSION(:)  ::  AL

    INTEGER(kind=kint ),POINTER, DIMENSION(:) ::  INU
    INTEGER(kind=kint ),POINTER, DIMENSION(:) ::  IAU
    INTEGER(kind=kint ),POINTER, DIMENSION(:) ::  INL
    INTEGER(kind=kint ),POINTER, DIMENSION(:) ::  IAL
    
    INTEGER(kind=kint ),allocatable, DIMENSION(:) :: node_record

    INTEGER(kind=kint) :: i,s,s_u,k,f,aggr,i_CN,size_of_aggregates,count,l,j,m,front,back
    INTEGER(kind=kint) :: node_record_counter,node,score,insert_point,max_score

    REAL   (kind=kreal) :: r
    INTEGER(kind=kint),parameter :: aggregate_number=1
    INTEGER(kind=kint),parameter :: aggregate_index=2
    logical :: flag

    s  = HIERARCHICAL_DATA(LEVEL_NO-1) % N 
    NP = HIERARCHICAL_DATA(LEVEL_NO-1) % NP 

    ALLOCATE(Temp(0:s,2))
    ALLOCATE(P_queue(2,s))
    ALLOCATE(aggregate_flags(s))
    ALLOCATE(node_record(s))
    
    D  =>HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL=>HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU=>HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL=>HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU=>HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL =>HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU =>HIERARCHICAL_DATA(LEVEL_NO-1) % AU

    s_u=INU(s)

    !C aggregates is stored in P temporarily 
    !C    P=>HIERARCHICAL_DATA(LEVEL_NO) % P    
    
    !C Temp =0
    Temp(1:s,aggregate_number) = 0
    Temp(0:s,aggregate_index) = 0
    
    DO i=1,s
       if(node_index(i)==0) then 
          node_index(i)=1
       end if
    END DO
    
    !C initialize Priority queue
    !C P_queue has score and node index. node_index has the index of P_queue
    !C -1 in node_index : selected as part of aggregate
    !C  0 in node_index : not selected as part of aggregate, but it's neighbor is aggregate.

    P_queue(1:2,1:s)=0
    aggregate_flags(1:s)=0
    
    
    !C interface part is scored, and enqueued
    !C Temp(,aggregate_number) is used as buffer
    queue_size = 0
    
    !C enqueue nodes whose score is 1
    DO count=1,s
       if(Temp(count,aggregate_number)==1)then
          Temp(count,aggregate_number)=0
          !C this case is enqueued
       else if(node_index(count)==1)then
          queue_size = queue_size+1
          node_index(count) = queue_size
          P_queue(1,queue_size) = 1
          P_queue(2,queue_size) = count
       end if
    END DO

    size_of_aggregates=0

    DO while(P_queue(1,1) > 0)
       
       i=P_queue(2,1)
       size_of_aggregates = size_of_aggregates+1

       !C----layer1 start
       DO k=PNI(i-1)+1,PNI(i)
          j=IAL( NI(k) )+ZERO_ORIGIN
          
          Temp(j, aggregate_number) = size_of_aggregates
          
          if(node_index(j)>0) then
             !C dequeue node j
             call dequeue(j, P_queue, node_index,queue_size)
          end if
          node_index(j)=-1
          aggregate_flags(j)=i
          
       END DO
       DO k=PNI(i+NP-1)+1,PNI(i+NP)
          if(IAU(NI(k))+ZERO_ORIGIN>s) cycle
          j=IAU(NI(k))+ZERO_ORIGIN
          Temp(j, aggregate_number) = size_of_aggregates
          
          if(node_index(j)>0) then
             !C dequeue node j
             call dequeue(j, P_queue, node_index,queue_size)
          end if
          node_index(j)=-1
          aggregate_flags(j)=i
       END DO
       
       Temp(i , aggregate_number) = size_of_aggregates
       
       !C dequeue node i
       call dequeue(i, P_queue, node_index,queue_size)
       
       node_index(i) = -1
       aggregate_flags(i) = i
       !C----layer1 end

       !C----layer2 and 3 
       call dequeue_scoreup_Layer23(i, s, NP, queue_size, P_queue,  & 
            & node_index, aggregate_flags, node_record, NI, PNI, IAL, IAU)
    END DO


    deallocate( node_record, aggregate_flags )

    aggregate_check: DO count=1,s
       if(node_index(count)==0) then
          lower_part: DO k=PNI(count-1)+1,PNI(count)
             !C aggr is the aggregate number of next to count
             aggr = Temp(IAL( NI(k) )+ZERO_ORIGIN, aggregate_number)
             IF(aggr > 0) Then
                Temp(count, aggregate_number) = aggr
                exit lower_part
             END IF
          END DO lower_part

          IF(Temp(count, aggregate_number) > 0) Then
             cycle aggregate_check
          END IF
          
          upper_part: DO k=PNI(count+NP-1)+1,PNI(count+NP)
             !C aggr is the aggregate number of next to i
             if(IAU(NI(k))+ZERO_ORIGIN>s) cycle
             aggr = Temp(IAU(NI(k))+ZERO_ORIGIN, aggregate_number)
             
             IF(aggr > 0) Then
                Temp(count,aggregate_number) = aggr
                exit upper_part
             END IF
          END DO upper_part
          IF(Temp(count, aggregate_number) == 0) Then
             write(*,*) "independent node in aggregate()",node_index(count),count
          END IF
       end if
    END DO aggregate_check

    DEALLOCATE(P_queue)
  
    in_aggregates_result_size=size_of_aggregates
    ALLOCATE( in_aggregates_result(0:size_of_aggregates),aggregates_result(s) )

    in_aggregates_result(0)=0

    DO i=1,s
       aggr=Temp(i,aggregate_number)
       if(aggr>0) Temp(aggr,aggregate_index) = Temp(aggr,aggregate_index)+1
    END DO

    !C size_of_aggregates has the size of aggregates
    !C P % IN(aggr)+1 points the place for new data of aggr
    DO i=1,size_of_aggregates
       Temp(i,aggregate_index) = Temp(i,aggregate_index)+Temp(i-1,aggregate_index)
       in_aggregates_result(i) = Temp(i-1,aggregate_index)
    END DO

    !C unneeded nodes belong to aggregate NO 0.
    DO i=1,s
       aggr = Temp(i, aggregate_number)
       if(aggr>0) then
          k=in_aggregates_result(aggr)+1
          in_aggregates_result(aggr)=k
          aggregates_result(k)=i
       end if
    END DO
    
#ifdef PRINT_REZ
    write(*,*) size_of_aggregates, " aggregates are selected from ", in_aggregates_result(size_of_aggregates)
#endif
    
    DEALLOCATE(Temp)
    
  end subroutine indpdt_agrgt

  SUBROUTINE agrgt_dealloc ( in_aggregates_result,aggregates_result )
    implicit none
    include 'precision.inc'

    INTEGER(kind=kint), pointer :: in_aggregates_result(:)
    INTEGER(kind=kint), pointer :: aggregates_result(:)
    
    deallocate(in_aggregates_result, aggregates_result)
  END SUBROUTINE agrgt_dealloc

  SUBROUTINE smooth_aggregate (PNI, NI, omega, LEVEL_NO, in_aggregates_result_size, &
       & in_aggregates_result, aggregates_result, Temp_P, Temp_P_row_size, tilde_D, &
       & Temp_P_SIZE)
    USE data_structure_for_AMG
    IMPLICIT NONE

    INTEGER(kind=kint ), INTENT(in) :: PNI(0:)
    INTEGER(kind=kint ), INTENT(in) :: NI(:)
    REAL   (kind=kreal), INTENT(in) :: omega
    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO
    INTEGER(kind=kint ), INTENT(in) :: in_aggregates_result_size
    REAL   (kind=kreal), INTENT(in) :: tilde_D(:)
    INTEGER(kind=kint ), INTENT(in) :: Temp_P_SIZE
    
    INTEGER(kind=kint) :: in_aggregates_result(0:)
    INTEGER(kind=kint) :: aggregates_result(:)
    TYPE(row_node)     :: Temp_P(:,:)
    INTEGER(kind=kint) :: Temp_P_row_size(:)

    REAL   (kind=kreal),ALLOCATABLE :: Temp(:,:)

    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)
    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)
    
    REAL   (kind=kreal) :: omega2
    INTEGER(kind=kint ) :: i, j, k, l, row, index, isL, ieL, isU, ieU, is, ie, NP, N
    INTEGER(kind=kint ) :: connected_node, connected_node_start, connected_node_end

    INTEGER(kind=kint)  :: row_no, row_size, A_row, A_col, P_col
    INTEGER(kind=kint), ALLOCATABLE :: Temp_int(:)

    NP = HIERARCHICAL_DATA(LEVEL_NO - 1) % NP
    N  = HIERARCHICAL_DATA(LEVEL_NO - 1) % N 

    ALLOCATE(Temp_int(NP))

    D   => HIERARCHICAL_DATA(LEVEL_NO - 1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO - 1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO - 1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO - 1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO - 1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO - 1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO - 1) % AU


    DO i = 1, NP
       Temp_P_row_size(i) = 0
    END DO
    
    
    !C-- Temp_int() has aggregate NO for each node.
    Temp_int = 0
    DO k = 1, in_aggregates_result_size
       is = in_aggregates_result(k - 1) + 1
       ie = in_aggregates_result(k)
       DO i = is, ie
          row_no = aggregates_result(i)
          Temp_int(row_no) = k
       END DO
    END DO
    
    !C-- Temp_P's space is calculated for smoothed aggregations
    !C   and input column and value( initial values 0 )
    DO A_row = 1, NP
       !C  element's  value of the aggregate is 1 in array P      
!!$#ifdef SMOOTH_TILDE_A             
       isL = PNI(A_row-1) + 1
       ieL = PNI(A_row)
!!$#else
!!$       isL = INL(A_row-1) + 1
!!$       ieL = INL(A_row)
!!$#endif
       DO i = isL, ieL
!!$#ifdef SMOOTH_TILDE_A             
          l = NI(i)
!!$#else
!!$          l = i
!!$#endif
          A_col = IAL(l)+ZERO_ORIGIN
          P_col = Temp_int(A_col)
          IF(P_col == 0) CYCLE
          
          DO k = 1, Temp_P_row_size(A_row) 
             IF( Temp_P(k,A_row) % column == P_col ) EXIT
          END DO
          IF(k > Temp_P_row_size(A_row)) THEN
             IF(k <= Temp_P_SIZE) THEN
                Temp_P_row_size(A_row)    = k
                Temp_P(k, A_row) % column = P_col
                Temp_P(k, A_row) % value  = 0.0
             ELSE
                STOP 'Temp_P should be enlarged : smooth_aggregate: lower part'
             END IF
          END IF
       END DO
!!$#ifdef SMOOTH_TILDE_A             
       isU = PNI(A_row + NP -1) + 1
       ieU = PNI(A_row + NP)
!!$#else
!!$       isU = INU(A_row - 1) + 1
!!$       ieU = INU(A_row)
!!$#endif
       DO i = isU,ieU
!!$#ifdef SMOOTH_TILDE_A             
          l = NI(i)
!!$#else
!!$          l = i
!!$#endif
          A_col = IAU(l)+ZERO_ORIGIN
          P_col = Temp_int(A_col)
          IF(P_col == 0) CYCLE

          DO k=1, Temp_P_row_size(A_row)
             IF( Temp_P(k, A_row) % column == P_col) EXIT
          END DO
          IF(k > Temp_P_row_size(A_row)) THEN
             IF(k <= Temp_P_SIZE) THEN                
                Temp_P_row_size(A_row) = k
                Temp_P(k, A_row) % column = P_col
                Temp_P(k, A_row) % value  = 0.0
             ELSE
                STOP 'Temp_P should be enlarged : smooth_aggregate upper part'                                      
             END IF
          END IF
       END DO
       
       !C- diagonal part
       A_col = A_row
       P_col = Temp_int(A_col)
       IF(P_col == 0) CYCLE
       
       DO k = 1, Temp_P_row_size(A_row)
          IF(Temp_P(k,A_row) % column == P_col) EXIT
       END DO
       IF(k > Temp_P_row_size(A_row)) THEN
          IF(k <= Temp_P_SIZE) THEN                
             Temp_P_row_size(A_row) = k
             Temp_P(k, A_row) % column = P_col
             Temp_P(k, A_row) % value  = 0.0
          ELSE
             STOP 'Temp_P should be enlarged : smooth_aggregate diagonal'                                      
          END IF
       END IF
    END DO

    !C-- smooth aggregates
    DO A_row = 1, NP
       row_size = Temp_P_row_size(A_row)
       IF(row_size == 0) CYCLE

       !C  element's  value of the aggregate is 1 in array P      
!!$#ifdef SMOOTH_TILDE_A             
       isL = PNI(A_row - 1) + 1
       ieL = PNI(A_row)
!!$#else
!!$       isL = INL(A_row - 1) + 1
!!$       ieL = INL(A_row)
!!$#endif
       DO i = isL, ieL
!!$#ifdef SMOOTH_TILDE_A             
          l = NI(i)
!!$#else
!!$          l = i
!!$#endif
          A_col = IAL(l)+ZERO_ORIGIN
          P_col = Temp_int(A_col)
          IF(P_col == 0) CYCLE
          
          DO k = 1, Temp_P_row_size(A_row) 
             IF( Temp_P(k,A_row) % column == P_col ) EXIT
          END DO
          IF(k > Temp_P_row_size(A_row)) THEN
             STOP 'unexpected error : smooth_aggregate()'
          END IF
          Temp_P(k, A_row) % value = Temp_P(k, A_row) % value - AL(l)
       END DO
       

!!$#ifdef SMOOTH_TILDE_A             
       isU = PNI(A_row + NP - 1) + 1
       ieU = PNI(A_row + NP)
!!$#else
!!$       isU = INU(A_row - 1) + 1
!!$       ieU = INU(A_row)
!!$#endif
       DO i = isU, ieU
!!$#ifdef SMOOTH_TILDE_A             
          l = NI(i)
!!$#else
!!$          l = i
!!$#endif
          A_col = IAU(l)+ZERO_ORIGIN
          P_col = Temp_int(A_col)
          IF(P_col == 0) CYCLE

          DO k = 1, Temp_P_row_size(A_row)
             IF( Temp_P(k,A_row) % column == P_col) EXIT
          END DO
          IF(k > Temp_P_row_size(A_row)) THEN
             STOP 'unexpected error : smooth_aggregate()'
          END IF
          
          Temp_P(k, A_row) % value = Temp_P(k, A_row) % value - AU(l)
       END DO
    END DO
    
    !C-- diagonal part
    DO A_row = 1, NP
       DO i= 1, Temp_P_row_size(A_row)
!!$#ifdef SMOOTH_TILDE_A                            
          Temp_P(i, A_row) % value = Temp_P(i, A_row) % value * omega / &
               & tilde_D(A_row)
!!$#else
!!$          Temp_P(i, A_row) % value = Temp_P(i, A_row) % value * omega / &
!!$               & D(A_row)
!!$#endif
       END DO
    END DO

    omega2 = 1. - omega
    DO row_no = 1, N
       P_col = Temp_int(row_no)
       IF(P_col == 0) CYCLE
       DO j = 1, Temp_P_row_size(row_no)
          IF(P_col == Temp_P(j, row_no) % column) EXIT
       END DO
       IF(j > Temp_P_row_size(row_no)) STOP 'unexpected error : smooth_aggregate()'
       Temp_P(j, row_no) % value = Temp_P(j, row_no) % value + omega2
    END DO
    !C-- end: smooth aggregates
    
    DEALLOCATE(Temp_int)
  END SUBROUTINE smooth_aggregate



!---SAAMG subroutines

  

  SUBROUTINE aggressive_coasening(N, PNIT, NIT, PNI, NI, node_index, coarser_level_size, SCORE)
    USE data_structure_for_AMG
    USE queue_mod
    implicit none
    
    INTEGER(kind=kint ), intent(in) :: PNIT(0:), PNI(0:), NIT(:), NI(:), N
    INTEGER(kind=kint ), intent(in) :: SCORE
    INTEGER(kind=kint ), intent(inout) :: node_index(:), coarser_level_size

    INTEGER(kind=kint ), allocatable :: PNI2(:), NI2(:), PNI2T(:), NI2T(:)
    INTEGER(kind=kint ), allocatable :: Temp(:), q(:)
    INTEGER(kind=kint ) :: qsize, NI2_index, re_alloc_times, lower, upper
    INTEGER(kind=kint ) :: n1, n2, i, j, k

    Type(intvec_list), target  :: NI2_list
    Type(intvec_list), pointer :: NI2_list_ptr

    allocate(PNI2(0:2*N), Temp(N), q(coarser_level_size))

    PNI2(0) = 0

    NI2_list_ptr => NI2_list
    allocate(NI2_list_ptr % intvec(N))
    
    Temp = 0
    NI2_index = 0
    re_alloc_times = 0
    DO i = 1, N
       if(node_index(i)==-2) then
          !C 2-time pass through the NI2(:)
          !C qsize, q(:) + Temparray(1:N)ѤƣNItraverse롣
          !C ǻ͡ʣ⤷ϣ˰ʾǤСNI2_list˳ǼƤ
          !C ƱˡPNI2ˤͤ롣
          qsize = 0 
          DO j = PNI(2*i-2)+1, PNI(2*i)
             n1 = NI(j)
             DO k = PNI(2*n1-2)+1, PNI(2*n1)
                n2 = NI(k)

                !C check n2
                if(node_index(n2) == -2) then
                   !C n2 is C-point
                   if(Temp(n2) == 0) then
                      qsize = qsize + 1
                      q(qsize) = n2
                   end if
                   Temp(n2) = Temp(n2) + 1
                end if
                
             END DO

             !C check n1
             if(node_index(n1) == -2) then
                !C n2 is C-point
                if(Temp(n1) == 0) then
                   qsize = qsize + 1
                   q(qsize) = n1
                end if
                Temp(n1) = Temp(n1) + 1
             end if

          END DO

          !C store q(:) to NI2_list and reset Temp(:)
          lower = 0; upper=0
          DO j = 1, qsize
             k = q(j)
             if(Temp(k) >= SCORE .and. k /= i) then
                !C SCORE is 1 or 2.

                NI2_index = NI2_index + 1
                if(NI2_index > N) then
                   !C reallocate temporary list of P_col & P_val
                   NI2_index = NI2_index - N
                   re_alloc_times = re_alloc_times + 1
                   
                   allocate(NI2_list_ptr % next)
                   NI2_list_ptr => NI2_list_ptr % next
                   allocate(NI2_list_ptr % intvec(N))
                end if

                NI2_list_ptr % intvec(NI2_index) = k
                if(k < i) then
                   lower = lower+1
                else 
                   upper = upper+1
                end if
                
             end if
             Temp(k) = 0
          END DO
          PNI2(2*i-1) = lower; PNI2(2*i)   = upper
       else
          !C store 0 for PNI2
          PNI2(2*i-1) = 0; PNI2(2*i) = 0
       end if
    END DO

    DO i = 1, 2*N
       PNI2(i) = PNI2(i-1)+PNI2(i)
    END DO

    allocate(NI2(PNI2(2*N)))

    !C NI2_listNI2˥ԡ
    NI2_list_ptr => NI2_list
    do i = 1, re_alloc_times


       do j = 1,N
          NI2(j+(i-1)*N) = NI2_list_ptr % intvec(j)
       end do
       NI2_list_ptr => NI2_list_ptr % next
    end do
    do i = 1, NI2_index 
       NI2(re_alloc_times * N + i) = NI2_list_ptr % intvec(i)
    end do
    
    !C deallocate NI2_list
    call deallocate_intlist(NI2_list, re_alloc_times)
    deallocate(Temp)
    
    j = PNI2(2*N)
    allocate(PNI2T(0:2*N), NI2T(j))

    call transpose_ni(N, PNI2, NI2, PNI2T, NI2T)

    !C in the case of aggresive coarsening, leave nodes of isolated nodes as C-points
    DO i = 1, N
       j = PNI2T(2*i)-PNI2T(2*i-2)+PNI2(2*i)-PNI2(2*i-2)
       !C reset node_index except for isolated C'points
       if(node_index(i) /= -2 .or. j /= 0) node_index(i) = 0
    END DO
!!$    node_index = 0

!    write(*,*) "aggressive coloring"
    call coloring(N, PNI2T, NI2T, PNI2, NI2, node_index, coarser_level_size)

    deallocate(PNI2, NI2, PNI2T, NI2T)
    
  END SUBROUTINE aggressive_coasening


  SUBROUTINE RAP(LEVEL_NO, Temp_SIZE)

    USE data_structure_for_AMG
    IMPLICIT NONE

    INTEGER(kind=kint), INTENT(in)   :: LEVEL_NO

    TYPE(INTER_LEVEL_OPERATOR), POINTER :: P, R

    INTEGER(kind=kint), INTENT(in) :: Temp_SIZE
!!$    INTEGER(kind=kint), INTENT(in) :: SYM_FLAG

    !C pointers for matrix
    REAL   (kind=kreal), POINTER ::  D(:), AU(:), AL(:)
    REAL   (kind=kreal), POINTER ::  newD(:), newAU(:), newAL(:)
    REAL   (kind=kreal), ALLOCATABLE :: Temp(:,:), Temp_1_col(:)
    REAL   (kind=kreal) :: v

    INTEGER(kind=kint ), POINTER ::  INU(:), IAU(:), INL(:), IAL(:)
    INTEGER(kind=kint ), POINTER ::  newINU(:), newIAU(:), newINL(:), newIAL(:)
    
    INTEGER(kind=kint ), ALLOCATABLE :: Temp_agr(:), Temp_sz(:)
    INTEGER(kind=kint ), ALLOCATABLE :: Temp_nd(:)
    INTEGER(kind=kint ), ALLOCATABLE :: Temp_int(:,:)


    INTEGER(kind=kint) :: i, j, k, l, m, row, col, row2, index_C
    INTEGER(kind=kint) :: coarser_level_size, js, je
    INTEGER(kind=kint) :: N, NP, newNPL, newNPU
    LOGICAL :: cancel_flag, finish_flag, flag

    INTEGER(kind=kint ) :: nd_sz, re_alloc_times


    Type(intvec_list), target  :: C_col, C_row
    Type(dblvec_list), target  :: C_val
    Type(intvec_list), pointer :: C_row_ptr, C_col_ptr
    Type(dblvec_list), pointer :: C_val_ptr
    INTEGER(kind=kint), allocatable :: C_row_num(:), C_SZ(:)

    NP   =  HIERARCHICAL_DATA(LEVEL_NO - 1) % NP 
    N    =  HIERARCHICAL_DATA(LEVEL_NO - 1) % N 
    D    => HIERARCHICAL_DATA(LEVEL_NO - 1) % D 
    INL  => HIERARCHICAL_DATA(LEVEL_NO - 1) % INL  
    INU  => HIERARCHICAL_DATA(LEVEL_NO - 1) % INU 
    IAL  => HIERARCHICAL_DATA(LEVEL_NO - 1) % IAL
    IAU  => HIERARCHICAL_DATA(LEVEL_NO - 1) % IAU
    AL   => HIERARCHICAL_DATA(LEVEL_NO - 1) % AL
    AU   => HIERARCHICAL_DATA(LEVEL_NO - 1) % AU
    P    => HIERARCHICAL_DATA(LEVEL_NO    ) % P
    R    => HIERARCHICAL_DATA(LEVEL_NO    ) % R
    
    coarser_level_size = R % ROW_SIZE

    !C coarser_level
    HIERARCHICAL_DATA(LEVEL_NO) % N  = coarser_level_size
    HIERARCHICAL_DATA(LEVEL_NO) % NP = coarser_level_size

    allocate(HIERARCHICAL_DATA(LEVEL_NO) % INL(0:coarser_level_size), &
         &   HIERARCHICAL_DATA(LEVEL_NO) % INU(0:coarser_level_size))
    newINL  => HIERARCHICAL_DATA(LEVEL_NO) % INL  
    newINU  => HIERARCHICAL_DATA(LEVEL_NO) % INU 

    newINL = 0; newINU = 0
    newNPL = 0; newNPU = 0

    !C Calculation of RAP. This uses R information. 
    !C R % IN must be rewrite @ the latter part.
    !C So Temp_R_IN is used for R % IN(:).
    !C ------
    ALLOCATE(Temp_sz (N), Temp_int(Temp_SIZE,N), Temp(Temp_SIZE,N),          &  
         &   Temp_agr(coarser_level_size), Temp_nd(coarser_level_size),          &
         &   Temp_1_col(coarser_level_size))

    ALLOCATE(C_row % intvec(N), C_col % intvec(N), C_val % dblvec(N))
    C_row_ptr => C_row
    C_col_ptr => C_col
    C_val_ptr => C_val
    index_C = 0
    
    Temp_1_col = 0    
    re_alloc_times = 0
    Temp_agr = 0
    !C-- RA
    finish_flag = .TRUE.
    DO WHILE(finish_flag)
       finish_flag = .FALSE.
       Temp_sz = 0
       
       DO col = 1, coarser_level_size
          IF(Temp_agr(col) > 0) CYCLE
          cancel_flag = .FALSE.
          
          DO i = R % IN(col - 1) + 1, R % IN(col) 
             row = R % CN(i)
             IF(row <= N .AND. Temp_sz(row) >= Temp_SIZE) THEN
                cancel_flag = .TRUE.
                EXIT
             END IF
          END DO
          IF(.NOT.cancel_flag) THEN
             DO i = R % IN(col - 1) + 1, R % IN(col) 
                row = R % CN(i)
                DO j = INL(row - 1) + 1, INL(row)
                   row2 = IAL( j )+ZERO_ORIGIN
                   IF(row2 <= N .AND. Temp_sz(row2) >= Temp_SIZE) THEN
                      cancel_flag = .TRUE.
                      EXIT
                   END IF
                END DO
                IF(cancel_flag) EXIT
                DO j = INU(row - 1) + 1, INU(row)
                   row2 = IAU( j )+ZERO_ORIGIN
                   IF(row2 <= N .AND. Temp_sz(row2) >= Temp_SIZE) THEN
                      cancel_flag = .TRUE.
                      EXIT
                   END IF
                END DO
                IF(cancel_flag) EXIT
             END DO
          END IF

          finish_flag = finish_flag .OR. cancel_flag
          IF(cancel_flag) CYCLE
          
          Temp_agr(col) = 1
          
          DO i = R % IN(col - 1) + 1, R % IN(col) 
             row = R % CN(i)
             v   = R % V(i)
             IF(row <= N) THEN
                k = Temp_sz( row )

                flag = (k==0)
                IF(k>0) THEN
                   IF(Temp_int(k,row) /= col) THEN
                      flag=.TRUE.
                   END IF
                END IF
                IF(flag) THEN                
!!$                IF( k == 0 .OR. Temp_int(k,row) /= col ) THEN                
                   k = k+1
                   Temp_sz ( row ) = k
                   Temp_int(k,row) = col
                   Temp    (k,row) = v * D(row)
                ELSE 
                   Temp(k,row) = Temp(k,row) + v * D(row)
                END IF
             END IF

             DO j = INL(row - 1) +1, INL(row)
                row2 = IAL( j )+ZERO_ORIGIN
                !C --
                IF(row2 > N) THEN
                   CYCLE
                END IF
                !C --

                k = Temp_sz(row2)

                flag = (k==0)
                IF(k > 0) THEN
                   IF(Temp_int(k,row2) /= col) THEN
                      flag = .TRUE.
                   END IF
                END IF
                IF(flag) THEN                
!!$                IF( k == 0 .OR. (k>0 .and. Temp_int(k,row2)/=col) ) THEN
                   k = k + 1
                   Temp_sz(row2) = k
                   Temp_int(k, row2) = col
                   Temp    (k, row2) = v * AL(j)
                ELSE
                   Temp(k, row2) = Temp(k, row2) + v * AL(j)
                END IF
             END DO

             DO j = INU(row - 1) + 1, INU(row)
                row2 = IAU(j)+ZERO_ORIGIN
                !C --
                IF(row2 > N) CYCLE
                !C --
                   
                k = Temp_sz(row2)

                flag = (k==0)
                IF(k > 0) THEN
                   IF(Temp_int(k, row2) /= col) THEN
                      flag = .TRUE.
                   END IF
                END IF
                IF(flag) THEN                
!!$                IF( k==0 .OR. (k>0 .and. Temp_int(k,row2)/=col) ) THEN
                   k = k + 1
                   Temp_sz(row2) = k
                   Temp_int(k, row2) = col
                   Temp    (k, row2) = v * AU(j)
                ELSE
                   Temp(k, row2) = Temp(k, row2) + v * AU(j)
                END IF
             END DO
          END DO
       END DO
       !C-- end: RA


       !C-- temp * P
       DO col = 1, R % ROW_SIZE
          nd_sz = 0
              
          js = R % IN(col - 1) + 1
          je = R % IN(col)
          DO j = js, je
             k = R % CN(j)
             v = R % V(j)

             DO l = 1, Temp_sz(k)
                row = Temp_int(l, k)

!!$                IF(SYM_FLAG /= 1 .OR. col >= row) THEN
                   DO m = 1, nd_sz
                      IF(Temp_nd(m) == row) EXIT
                   END DO
                   IF(m > nd_sz) THEN
                      nd_sz = nd_sz + 1
                      Temp_nd(nd_sz) = row
                   END IF
                   Temp_1_col(row) = Temp_1_col(row) + v * Temp(l, k)
!!$                END IF
             END DO
          END DO
          
          !C-- input
          DO j = 1, nd_sz
             row = Temp_nd(j)
             !C calculate the index of the coarser level matrix
             if(col < row) then
                newNPL = newNPL + 1
                k = row+1
                if(k <= coarser_level_size ) then
                   newINL(k) = newINL(k) + 1
                end if
             else if(col > row) then
                newNPU = newNPU + 1
                k = row+1
                if(k <= coarser_level_size) then
                   newINU(k) = newINU(k) + 1
                end if
             end if
             index_C = index_C + 1
             if(index_C > N) then
                re_alloc_times = re_alloc_times + 1
                allocate(C_row_ptr % next, C_col_ptr % next, C_val_ptr % next)
                C_row_ptr => C_row_ptr % next
                C_col_ptr => C_col_ptr % next
                C_val_ptr => C_val_ptr % next
                allocate(C_row_ptr % intvec(N), C_col_ptr % intvec(N), C_val_ptr % dblvec(N))
                index_C = index_C - N
             end if
             C_row_ptr % intvec(index_C) = row
             C_col_ptr % intvec(index_C) = col
             C_val_ptr % dblvec(index_C) = Temp_1_col(row)
             Temp_1_col(row) = 0
          END DO
          !C-- end: input
       END DO
       !C-- end:temp * P
    END DO

    deallocate(Temp_sz, Temp_int, Temp, Temp_agr, Temp_nd, Temp_1_col)

    !C input C_row, C_col, C_val
    HIERARCHICAL_DATA(LEVEL_NO) % NPL = newNPL
    HIERARCHICAL_DATA(LEVEL_NO) % NPU = newNPU

    DO i = 1, coarser_level_size
       newINL(i) = newINL(i) + newINL(i-1) 
       newINU(i) = newINU(i) + newINU(i-1) 
    END DO

    allocate(HIERARCHICAL_DATA(LEVEL_NO) % IAL(newNPL),          &
         &   HIERARCHICAL_DATA(LEVEL_NO) % AL(newNPL),           &
         &   HIERARCHICAL_DATA(LEVEL_NO) % IAU(newNPU),          &
         &   HIERARCHICAL_DATA(LEVEL_NO) % AU(newNPU),           &
         &   HIERARCHICAL_DATA(LEVEL_NO) % D(coarser_level_size),&
         &   HIERARCHICAL_DATA(LEVEL_NO) % X(coarser_level_size),&
         &   HIERARCHICAL_DATA(LEVEL_NO) % B(coarser_level_size))
    newD    => HIERARCHICAL_DATA(LEVEL_NO) % D 
    newIAL  => HIERARCHICAL_DATA(LEVEL_NO) % IAL
    newIAU  => HIERARCHICAL_DATA(LEVEL_NO) % IAU
    newAL   => HIERARCHICAL_DATA(LEVEL_NO) % AL
    newAU   => HIERARCHICAL_DATA(LEVEL_NO) % AU


    C_row_ptr => C_row
    C_col_ptr => C_col
    C_val_ptr => C_val
    DO i = 1, re_alloc_times
       DO j = 1, N
          row = C_row_ptr % intvec(j); col = C_col_ptr % intvec(j); v = C_val_ptr % dblvec(j)
          if(col > row) then
             !C upper part
             k = newINU(row) + 1; newINU(row) = k;
             newIAU(k) = col; newAU(k)  = v
          else if (col < row) then
             !C lower part
             k = newINL(row) + 1; newINL(row) = k;
             newIAL(k) = col; newAL(k)  = v
          else
             !C diagonal part
             newD(row) = v
          end if
       END DO
       C_row_ptr => C_row_ptr % next; C_col_ptr => C_col_ptr % next; C_val_ptr => C_val_ptr % next
    END DO
    
    DO j = 1, index_C
       row = C_row_ptr % intvec(j); col = C_col_ptr % intvec(j); v = C_val_ptr % dblvec(j)
       if(col > row) then
          !C upper part
          k = newINU(row) + 1; newINU(row) = k; 
          newIAU(k) = col; newAU(k)  = v
       else if (col < row) then
          !C lower part
          k = newINL(row) + 1; newINL(row) = k;
          newIAL(k) = col; newAL(k)  = v
       else
          !C diagonal part
          newD(row) = v
       end if
    end DO


    !C deallocate C_... list
    call deallocate_intlist(C_row, re_alloc_times)
    call deallocate_intlist(C_col, re_alloc_times)
    call deallocate_dbllist(C_val, re_alloc_times)

    

  END SUBROUTINE RAP

  SUBROUTINE posi_couple_check(plus, PNI, NI, PNIT, NIT, node_index, &
       &                       coarser_level_size, LEVEL_NO)
    USE data_structure_for_AMG
    implicit none
    
    INTEGER(kind=kint ), intent(in) :: LEVEL_NO
    REAL(kind=kreal ), intent(in) :: plus
    INTEGER(kind=kint ), pointer    :: PNI(:), NI(:), PNIT(:), NIT(:)
    INTEGER(kind=kint ), intent(inout) :: node_index(:),coarser_level_size

    INTEGER(kind=kint ), allocatable   :: temp_PNI(:), temp_NI(:)
    INTEGER(kind=kint ), pointer   :: new_PNI(:), new_NI(:)

    REAL(kind=kreal )  :: max,  max_v, v
    INTEGER(kind=kint) :: max_in

    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)

    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)

    INTEGER(kind=kint ), allocatable :: vec(:)
    INTEGER(kind=kint ) :: i,j,k,col, pos
    INTEGER(kind=kint ) ::  N, NPL, NPU

    N = HIERARCHICAL_DATA(LEVEL_NO-1) % N 
    NPL = HIERARCHICAL_DATA(LEVEL_NO-1) % NPL 
    NPU = HIERARCHICAL_DATA(LEVEL_NO-1) % NPU 
    D   => HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO-1) % AU


    !C temp_?NI is for positive couplings 
    j = NPL+NPU-PNI(2*N) 
    allocate(temp_PNI(0:2*N), temp_NI(j))
    
    temp_PNI = 0; pos = 0
    DO i = 1, N
       if(node_index(i) /= -2) then
          !C i is the F-point

          !C max_v: max absolute value
          !C max : max positive F-F connection strength
          !C max_in: index of temp_NI(:)
          max_v = 0.0; max = 0.0; max_in = 0
          
          !C determine max_v
          DO j = INL(i-1)+1, INL(i)
             v = dabs(AL(j))
             if(max_v < v) max_v = v
          END DO
          DO j = INU(i-1)+1, INU(i)
             v = dabs(AU(j))
             if(max_v < v) max_v = v
          END DO

          !C count up and store the complement elememts in temp_NI
          !C and max positive coupling is searched
          DO j = INL(i-1)+1, INL(i)
             if(max_v * plus < AL(j)) then
                pos = pos + 1
                temp_NI(pos) = IAL(j)
                if(max < AL(j)) then 
                   max = AL(j); max_in = pos
                end if

             end if
          END DO
          temp_PNI(2*i-1) = pos

          DO j = INU(i-1)+1, INU(i)
             if(max_v * plus < AU(j)) then
                pos = pos + 1
                temp_NI(pos) = IAU(j)
                if(max < AU(j)) then 
                   max = AU(j); max_in = pos
                end if
                
             end if
          END DO
          temp_PNI(2*i) = pos

          !C the variable (the largest positive coupling ) is changed to C-point.
          if(pos - temp_PNI(2*i-2) > 0) then
             
             !C check if there is any C-points representing strong positive connections
             DO j = temp_PNI(2*i-2)+1, temp_PNI(2*i)
                col = node_index( temp_NI(j) )
                if(node_index(col) == -2) exit
             END DO

!!$             if(node_index(col) /= -2) then
             !C if there are no C-point representing the strong positive connections
             if(j > temp_PNI(2*i)) then
                !C the largest positive F-F coupling is removed and C-point is added.
                !C there are positive couplings.
                col = temp_NI(max_in)
                node_index(col) = -2
                coarser_level_size = coarser_level_size + 1
             end if
          end if
          
       else
          temp_PNI(2*i-1) = pos; temp_PNI(2*i) = pos
       end if
    END DO

    if(pos > 0) then
       !C allocate(new NI)
       i = PNI(2*N)+ temp_PNI(2*N)
       allocate(new_NI(i), new_PNI(0:2*N))

       pos = 0; new_PNI(0) = 0
       !C assign new values to allocated area.
       DO i = 1, N
          DO j = PNI(2*i-2)+1, PNI(2*i-1)
             pos = pos + 1
             new_NI(pos) = NI(j)
          END DO
          DO j = temp_PNI(2*i-2)+1, temp_PNI(2*i-1)
             pos = pos + 1
             new_NI(pos) = temp_NI(j)
          END DO
          new_PNI(2*i-1) = pos

          DO j = PNI(2*i-1)+1, PNI(2*i)
             pos = pos + 1
             new_NI(pos) = NI(j)
          END DO
          DO j = temp_PNI(2*i-1)+1, temp_PNI(2*i)
             pos = pos + 1
             new_NI(pos) = temp_NI(j)
          END DO
          new_PNI(2*i) = pos
       END DO


       !C pointer association to new data
       deallocate(PNI, NI)
       PNI => new_PNI; NI => new_NI

       !C PNIT, NIT is changed according to PNI, NI.
       DEALLOCATE(NIT)
       j = PNI(2*N)
       ALLOCATE(NIT(j))
       call transpose_ni(N, PNI, NI, PNIT, NIT)
       
    end if
    deallocate(temp_PNI, temp_NI)
       

    write(*,*) coarser_level_size, " C-points in posi_couple_check()"
  END SUBROUTINE posi_couple_check


  SUBROUTINE coloring(N, PNIT, NIT, PNI, NI, node_index, coarser_level_size)
    USE queue_mod
    implicit none
    include 'precision.inc'
    
    INTEGER(kind=kint ), intent(in) :: PNIT(0:), PNI(0:), NIT(:), NI(:), N
    INTEGER(kind=kint ), intent(inout) :: node_index(:)
    
    INTEGER(kind=kint ), allocatable :: P_queue(:,:)
    INTEGER(kind=kint ) :: i, j, k, s, l1, l2, queue_size
    
    INTEGER(kind=kint ), intent(out) :: coarser_level_size

    coarser_level_size = 0
    allocate(P_queue(2, N))

    !C node_index has 0.
    !C P_queue is initialized -1.
    P_queue(1,:) =  -1
    queue_size = 0
    DO i = 1, N
       s = PNIT(2*i)-PNIT(2*i-2); j = PNI(2*i)-PNI(2*i-2)
       if(node_index(i) == -2) then
          !C i is initially set as coarse node
          coarser_level_size = coarser_level_size + 1
          
       else if(s+j >0) then
          call enqueue(i, s, P_queue, node_index, queue_size)
       end if
    END DO

    
    !C selecting C point
    DO while(P_queue(1,1) /= -1 .and. P_queue(1,1) /= 0)
       coarser_level_size = coarser_level_size + 1
       i = P_queue(2,1)

       !C node_index(i) == -2 :: C point
       call dequeue(i, P_queue, node_index, queue_size)
       node_index(i) = -2

       DO j = PNIT(2*i-2)+1, PNIT(2*i)
          l1 = NIT(j)
          if(node_index(l1) > 0) then
             !C node_index(i) == -1 :: F point
             call dequeue(l1, P_queue, node_index, queue_size)          
             node_index(l1) = -1

             DO k = PNI(2*l1-2)+1, PNI(2*l1)
                l2 = NI(k)
                if(node_index(l2) > 0) call score_up(l2, P_queue, node_index, queue_size)
             END DO
          end if
          
       END DO
    end DO

    write(*,*) coarser_level_size, " C-points are selected from ", N
    deallocate(P_queue)

  END SUBROUTINE coloring


  !C this sub finds f-f connection, whose nodes aren't dependent on any common C-points ,
  !C and reallocates one node in the connection as C-point.
  !C 
  !C if |a(i,j)| > NONZERO .and. |a(j,k)| < NONZERO exist i \in 1..N, exist j \in Fi, forall k \in Ci,
  !C       i is set as C-point. NONZERO is set as EPS*(diagonal value)
  !C here, matrix structure is not used as non-zero mark.
  SUBROUTINE f_f_check(LEVEL_NO, node_index, coarser_level_size, NONZERO_THRSLD)
    USE data_structure_for_AMG
    implicit none
    
    INTEGER(kind=kint), intent(in) :: LEVEL_NO
    INTEGER(kind=kint), intent(inout) :: node_index(:), coarser_level_size
    REAL(kind=kreal), intent(in) :: NONZERO_THRSLD

    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)

    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)

    

    INTEGER(kind=kint ) :: NPL, NPU, N
    
    INTEGER(kind=kint), allocatable :: temp_vec(:), temp_vec2(:)

    INTEGER(kind=kint ) :: i, j, k, l, vec2_size, col
    REAL(kind=kreal) :: NONZERO


    NPU = HIERARCHICAL_DATA(LEVEL_NO-1) % NPU
    NPL = HIERARCHICAL_DATA(LEVEL_NO-1) % NPL
    N   = HIERARCHICAL_DATA(LEVEL_NO-1) % N 
    
    D   => HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO-1) % AU

    allocate(temp_vec(N), temp_vec2(N))

    !C initialize temp_vec 
    temp_vec = 0
    DO i = 1, N
       if(node_index(i) /= -2) then
          NONZERO = NONZERO_THRSLD*dabs(D(i))

          !C mark C-points in temp_vec and record F-points in temp_vec2
          vec2_size = 0
          DO j = INL(i-1)+1, INL(i)
             col = IAL(j)
             if(dabs(AL(j)) > NONZERO) then
                if(node_index(col)==-2) then
                   temp_vec(col) = i
                else
                   !C i-th row's F-points with nonzero value is stored in temp_vec2.
                   vec2_size = vec2_size+1
                   temp_vec2(vec2_size) = col
                END if
             END if
          END DO
          DO j = INU(i-1)+1, INU(i)
             col = IAU(j)
             if(dabs(AU(j)) > NONZERO) then
                if(node_index(col)==-2) then
                   temp_vec(col) = i
                else
                   !C i-th row's F-points with nonzero value is stored in temp_vec2.
                   vec2_size = vec2_size+1
                   temp_vec2(vec2_size) = col
                end if
             end if
          END DO
          
          DO j = 1, vec2_size
             k = temp_vec2(j)
             !C i-k connection is checked here.

             NONZERO = NONZERO_THRSLD*dabs(D(k))       

             DO l = INL(k-1)+1, INL(k)
                col = IAL(l)
                !C 1.common C-point is found
                !C   confirm the the element's value is not 0.
                if(temp_vec(col) == i .and. dabs(AL(l)) > NONZERO) exit
             END DO
             !C case 1.: check next C-point in i-th row
             if(l > INL(k)) then
                
                DO l = INU(k-1)+1, INU(k)
                   col = IAU(l)
                   !C 2.common C-point is found
                   !C   confirm the the element's value is not 0.
                   if(temp_vec(col) == i .and. dabs(AU(l)) > NONZERO) exit
                END DO
                !C case 2.: check next C-point in i-th row          
                if(l > INU(k)) then
                   !C 3. here, F-F connection to remove is found.
                   !C then, F-point i is set as C. 
                   node_index(i) = -2
                   coarser_level_size = coarser_level_size + 1
                   exit 
                end if
             end if
          END DO
       end if
    END DO
    
    deallocate(temp_vec, temp_vec2)
!    write(*,*) "f_f_check:  there are ", coarser_level_size, " C-points."

    
  END SUBROUTINE f_f_check

  !C this sub. can deal with a matrix with zero elements in the non-zero structure.
  !C this sub. add new C-points so that strongly-connected F-points share C-points 
  !C which influence those F-points.
  SUBROUTINE strong_f_f_check(LEVEL_NO, node_index, coarser_level_size, PNI, NI, PNIT, NIT)
    USE data_structure_for_AMG
    implicit none

    INTEGER(kind=kint), intent(in) :: LEVEL_NO, PNI(0:), NI(:), PNIT(0:), NIT(:)
    INTEGER(kind=kint), intent(inout) :: node_index(:), coarser_level_size

    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)

    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)

    INTEGER(kind=kint ) :: NPL, NPU, N

    INTEGER(kind=kint), allocatable :: temp_vec(:), cntd_F_nds(:)

    INTEGER(kind=kint ) :: i, j, k, l, m, col, max, TIMES, delta, count, uppb, lowb



    NPU = HIERARCHICAL_DATA(LEVEL_NO-1) % NPU
    NPL = HIERARCHICAL_DATA(LEVEL_NO-1) % NPL
    N   = HIERARCHICAL_DATA(LEVEL_NO-1) % N 

    D   => HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO-1) % AU

    allocate(temp_vec(N), cntd_F_nds(N))

!!$    TIMES = 16
!!$
!!$    !C countup strong F-F connections in cntd_F_nds(:)
!!$    max = 0
!!$    DO i = 1, N
!!$       if(node_index(i) /= -2) then
!!$
!!$          count = 0
!!$          DO j = PNIT(2*i-2)+1, PNIT(2*i)
!!$             col = NIT(j)
!!$             
!!$             if(node_index(col) /= -2) then
!!$                count = count + 1 
!!$             END if
!!$             
!!$          END DO
!!$
!!$          cntd_F_nds(i) = count
!!$          if(max < count) max = count
!!$          
!!$       else
!!$          
!!$          cntd_F_nds(i) = -1
!!$          
!!$       end if
!!$    END DO
!!$    
!!$
!!$    delta = max / TIMES
!!$    write(*,*) max, TIMES, delta
!!$
!!$    !C initialize temp_vec     
!!$    temp_vec = 0
!!$    DO m = 1, TIMES
!!$       
!!$       lowb = max - delta * m
!!$       uppb = max - delta * (m-1)
!!$
!!$       !C last cycle deals nodes with 0..(max-(TIMES-1)*delta) in cntd_F_nds
!!$       !C C-nodes has -1 in cntd_F_nds, thus, the last cycle deals with only F-nodes
!!$       if(m == 1) lowb = -1
!!$       
!!$       !C always check ith row with plus value in cntd_F_nds(i)
!!$       !C this corresponds to the fact that the node is F-point
!!$       DO i = 1, N
!!$          if( cntd_F_nds(i) > lowb  .and. cntd_F_nds(i) <= uppb) then
!!$             
!!$             !C ith row must be F-node.
!!$             !C this sub changes i,j,k,l,col,temp_vec,coarser_level_size
!!$             call check_ith_row() 
!!$          end if
!!$       END DO
!!$       write(*,*) m, ":", coarser_level_size
!!$    END DO
    
    
    !C initialize temp_vec 
    temp_vec = 0
    DO i = 1, N
       if(node_index(i) /= -2) then
          call check_ith_row()
       end if
    END DO

    deallocate(temp_vec, cntd_F_nds)
!    write(*,*) "strong_f_f_check:  there are ", coarser_level_size, " C-points."

  contains
    subroutine check_ith_row()

      DO j = PNI(2*i-2)+1, PNI(2*i)
         col = NI(j)
         if(node_index(col) == -2) then
            temp_vec(col) = i
         END if
      END DO

      DO j = PNI(2*i-2)+1, PNI(2*i)
         k = NI(j)
         if(node_index(k) /= -2) then

            !C i-k connection is checked here.
            DO l = PNI(2*k-2)+1, PNI(2*k)
               col = NI(l)
               if(temp_vec(col) == i) exit
            END DO

            !C case 2.: check next C-point in i-th row          
            if(l > PNI(2*k)) then
               node_index(i) = -2
               coarser_level_size = coarser_level_size + 1
               exit
            end if
         end if
      END DO

    end subroutine check_ith_row
  END SUBROUTINE strong_f_f_check
  
  
  SUBROUTINE transpose_P_in_R(LEVEL_NO, coarser_level_size)
    USE data_structure_for_AMG
    INTEGER(kind=kint), intent(in) :: LEVEL_NO, coarser_level_size
    TYPE(inter_level_operator), pointer :: R, P
    
    INTEGER(kind=kint) :: i,j,k,col, count
    REAL(kind=kreal) :: val

    !C    initialize R
    ALLOCATE(HIERARCHICAL_DATA(LEVEL_NO) % R)
    P => HIERARCHICAL_DATA(LEVEL_NO) % P
    R => HIERARCHICAL_DATA(LEVEL_NO) % R
    
    ALLOCATE(R % IN(0:coarser_level_size))
    R % IN(0:coarser_level_size) = 0
    R % ROW_SIZE = coarser_level_size

    count = 0
    R % IN(0) = 0
    DO i = 1, P % ROW_SIZE
       DO j = P % IN(i-1)+1, P % IN(i)
          k = P % CN(j) + 1
          count = count + 1
          if(k <= coarser_level_size) then
             R % IN(k) = R % IN(k) + 1
          end if
       END DO
    END DO

    ALLOCATE(R % CN(count), R % V(count))

    DO i = 1, R % ROW_SIZE
       R % IN(i) = R % IN(i) + R % IN(i-1)
    END DO

    DO i = 1, P % ROW_SIZE
       DO j = P % IN(i-1)+1, P % IN(i)
          col = P % CN(j); val = P % V(j)
          k = R % IN(col)
          k = k+1
          R % IN(col) = k
          R % CN(k) = i
          R % V(k) = val
       END DO
    END DO

    
  END SUBROUTINE transpose_P_in_R

  SUBROUTINE expand_by_kth_row(v, k, f_ivec, f_rvec, q, qsize, INL, IAL, AL, &
       & INU, IAU, AU, D)
    implicit none
    include 'precision.inc'
    INTEGER(kind=kint) :: k,l,col,f_ivec(:),qsize,q(:)
    REAL(kind=kreal) :: v,f_rvec(:),val

    INTEGER(kind=kint), INTENT(in) :: INL(0:),INU(0:),IAL(:),IAU(:)
    REAL(kind=kreal),   INTENT(in) :: AL(:),AU(:),D(:)


    DO l = INL(k-1)+1, INL(k)
       col = IAL(l); val = AL(l)
       if(f_ivec(col) > 0) then
          f_rvec(col) = f_rvec(col) -v*val/D(k)
       else
          qsize = qsize + 1; q(qsize) = col
          !C  positive value means col-th element in f_?vec()
          !C is already recorded.
          f_ivec(col) = 1
          f_rvec(col) = -v*val/D(k)          
       end if
    END DO
    DO l = INU(k-1)+1, INU(k)
       col = IAU(l); val = AU(l)
       if(f_ivec(col) > 0) then
          f_rvec(col) = f_rvec(col) -v*val/D(k)
       else
          qsize = qsize + 1; q(qsize) = col
          !C  positive value means col-th element in f_?vec()
          !C is already recorded.
          f_ivec(col) = 1
          f_rvec(col) = -v*val/D(k)                      
       end if
    END DO
                
    
  END SUBROUTINE expand_by_kth_row


  !C D, INL, .., AU are pointers for original mat.
  SUBROUTINE get_back_expand_plus(D, INL, INU, IAL, IAU, AL, AU, &
       & LEVEL_NO)
    USE data_structure_for_AMG
    implicit none

    INTEGER(kind=kint ), intent(in) :: LEVEL_NO
    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)
    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)

    INTEGER(kind=kint ) :: N

    deallocate(                                   &
         HIERARCHICAL_DATA(LEVEL_NO-1) % D   ,    &
         HIERARCHICAL_DATA(LEVEL_NO-1) % INL ,    &
         HIERARCHICAL_DATA(LEVEL_NO-1) % INU ,    &
         HIERARCHICAL_DATA(LEVEL_NO-1) % IAL ,    &
         HIERARCHICAL_DATA(LEVEL_NO-1) % IAU ,    &
         HIERARCHICAL_DATA(LEVEL_NO-1) % AL  ,    &
         HIERARCHICAL_DATA(LEVEL_NO-1) % AU  )

    HIERARCHICAL_DATA(LEVEL_NO-1) % D => D
    HIERARCHICAL_DATA(LEVEL_NO-1) % INL =>INL
    HIERARCHICAL_DATA(LEVEL_NO-1) % INU =>INU
    HIERARCHICAL_DATA(LEVEL_NO-1) % IAL =>IAL
    HIERARCHICAL_DATA(LEVEL_NO-1) % IAU =>IAU
    HIERARCHICAL_DATA(LEVEL_NO-1) % AL => AL
    HIERARCHICAL_DATA(LEVEL_NO-1) % AU => AU

    N = HIERARCHICAL_DATA(LEVEL_NO-1) % N
    HIERARCHICAL_DATA(LEVEL_NO-1) % NPL = INL(N)
    HIERARCHICAL_DATA(LEVEL_NO-1) % NPU = INU(N)

    
  END SUBROUTINE get_back_expand_plus
    

  !C expand large plus elements of matrix rows  Ap
  !C D, INL, .., AU are pointers for backup of original mat.
  SUBROUTINE expand_plus_elements(D, INL, INU, IAL, IAU, AL, AU, &
       & LEVEL_NO, pepsilon)
    USE data_structure_for_AMG
    implicit none

    REAL(kind=kreal), INTENT(in) :: pepsilon

    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO

    Type(intvec_list), target :: Ap_col
    Type(dblvec_list), target :: Ap_val
    Type(intvec_list), pointer :: Ap_col_ptr
    Type(dblvec_list), pointer :: Ap_val_ptr
    INTEGER(kind=kint) :: index, re_alloc_times 
    INTEGER(kind=kint), allocatable :: Ap_row(:)
    INTEGER(kind=kint) :: Ap_row_size
    
    REAL   (kind=kreal), POINTER :: D  (:), PD(:)
    REAL   (kind=kreal), POINTER :: AU (:), PAU(:)
    REAL   (kind=kreal), POINTER :: AL (:), PAL(:)
    INTEGER(kind=kint ), POINTER :: INU(:), PINU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:), PIAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:), PINL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:), PIAL(:)
    INTEGER(kind=kint ) :: N

    INTEGER(kind=kint) :: qsize
    INTEGER(kind=kint), allocatable :: q(:), f_ivec(:)
    REAL(kind=kreal),   allocatable :: qr(:), f_rvec(:)

    INTEGER(kind=kint) :: i,j,k,l,m,col, L_index, U_index
    INTEGER(kind=kint) :: row, rowsize, Ap_row_index
    REAL(kind=kreal) :: val, max, max_p
    

    N  = HIERARCHICAL_DATA(LEVEL_NO-1)   % N 
    D   => HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO-1) % AU
    
    !C initialize list structure
    Ap_col_ptr => Ap_col
    Ap_val_ptr => Ap_val
    Ap_col_ptr % size = N
    allocate(Ap_col_ptr % intvec(N))
    allocate(Ap_val_ptr % dblvec(N))
    index = 0
    re_alloc_times = 0
    allocate(Ap_row(N))
    Ap_row_size = 0

    allocate(q(N),f_ivec(N),f_rvec(N))
    f_ivec = 0

    allocate(PINL(0:N), PINU(0:N))
    PINL = 0; PINU=0

    !C check each row and expand in the list structure of Ap
    DO i = 1, N
       max = 0; max_p = 0
       DO j = INL(i-1)+1, INL(i)
          if(max  >AL(j)) max = AL(j)
          if(max_p<AL(j)) max_p = AL(j)
       END DO
       DO j = INU(i-1)+1, INU(i)
          if(max  >AU(j)) max   = AU(j)
          if(max_p<AU(j)) max_p = AU(j)
       END DO

       !C there is some elements to be expanded
       if(max_p > -pepsilon*max) then
          qsize = 0

          !C record the row number to be expanded in list Ap.
          Ap_row_size = Ap_row_size+1
          Ap_row(Ap_row_size) = i
          
          !C insert the non-zero elements into f_?vec, q?.
          DO j = INL(i-1)+1, INL(i)
             col = IAL(j); val = AL(j)
             if(val <= -pepsilon*max) then
                qsize = qsize+1
                q(qsize) = col; f_ivec(col) = j;
                f_rvec(col) = val
             end if
          END DO
          DO j = INU(i-1)+1, INU(i)
             col = IAU(j); val = AU(j)
             if(val <= -pepsilon*max) then
                qsize = qsize+1
                q(qsize) = col; f_ivec(col) = j;
                f_rvec(col) = val
             end if
          END DO

          !C 1 is dammy value. positive value shows i-th element is already recorded..
          f_ivec(i) = 1; f_rvec(i) = D(i)
          
          !C expand the big plus elements.          
          DO j = INL(i-1)+1, INL(i)
             col = IAL(j); val = AL(j)
             if(val > -pepsilon*max) then
                call expand_by_kth_row(val, col, f_ivec, f_rvec, q, qsize, INL, IAL, AL, INU,&
                     & IAU, AU, D)
             end if
          END DO
          DO j = INU(i-1)+1, INU(i)
             col = IAU(j); val = AU(j)
             if(val > -pepsilon*max) then
                call expand_by_kth_row(val, col, f_ivec, f_rvec, q, qsize, INL, IAL, AL, INU,&
                     & IAU, AU, D)
             end if
          END DO


          !C q(:) doesn't contain diagonal element. diagonal element is treated first.
          col = i; val = f_rvec(i)
          call insert_temp_P(col, val, index, re_alloc_times, Ap_col_ptr, Ap_val_ptr, N)
          f_ivec(i) = 0
          
          !C store the expanded row in list structure Ap
          DO j = 1, qsize
             col = q(j)
             val = f_rvec(col)
             
             call insert_temp_P(col, val, index, re_alloc_times, Ap_col_ptr, Ap_val_ptr, N)
             if(col < i) then
                PINL(i) = PINL(i)+1
             else if(col > i) then
                PINU(i) = PINU(i)+1
             end if
             
             !C reset f_ivec
             f_ivec(col) = 0
          END DO
       else
          PINL(i) = INL(i)-INL(i-1)
          PINU(i) = INU(i)-INU(i-1)
       end if
    END DO

    deallocate(q, f_ivec, f_rvec)
    
    !C neat PIN?(:)
    DO i = 1, N
       PINL(i)=PINL(i) + PINL(i-1)
       PINU(i)=PINU(i) + PINU(i-1)
    END DO

    allocate(PD(N), PAU(PINU(N)),PIAU(PINU(N)), &
         & PAL(PINL(N)),PIAL(PINL(N)))
    !C create expanded matrix in PA from matrix A and row-list structure Ap
    k = 0
    Ap_col_ptr => Ap_col
    Ap_val_ptr => Ap_val
    DO i = 1, Ap_row_size
       row = Ap_row(i)
       L_index = PINL(row-1); U_index = PINU(row-1)

       rowsize = PINU(row)+PINL(row)-PINU(row-1)-PINL(row-1)+1


       !C  k is counting from 0.
       !C So the last element is at k+rowsize-1.
       m = k+rowsize-1

       !C  the row's data is in the range from  k to m
       !C  if m/N /= k/N, then change to the next list at ? mod N == N-1
       if(m/N > k/N) m = (k/N+1)*N-1
       DO j = k, m
          l = mod(j , N)+1
          col = Ap_col_ptr % intvec(l)
          val = Ap_val_ptr % dblvec(l)
          call copy_Ap_to_PA()    
       END DO

       !C change to the next list
       if(mod(m, N) == N-1) then
          
          Ap_col_ptr => Ap_col_ptr % next
          Ap_val_ptr => Ap_val_ptr % next
          
          m = k+rowsize-1; k = j

          DO j = k, m
             l = mod(j , N)+1
             col = Ap_col_ptr % intvec(l)
             val = Ap_val_ptr % dblvec(l)
             call copy_Ap_to_PA()    
          END DO
       end if
       
       k = j
    END DO

    !C copy the matrix A to matrix PA
    if(Ap_row_size < N) Ap_row(Ap_row_size+1)=0 !C end mark
    Ap_row_index = 1
    DO i = 1, N
       row = Ap_row(Ap_row_index)
       if(i == row) then
          Ap_row_index = Ap_row_index + 1
       else
          !C copy i-th row  to PA
          k = INL(i-1)-PINL(i-1)
          DO j = PINL(i-1)+1, PINL(i)
             PAL(j) = AL(j+k)
             PIAL(j) = IAL(j+k)
          END DO
          k = INU(i-1)-PINU(i-1)
          DO j = PINU(i-1)+1, PINU(i)
             PAU(j) = AU(j+k)
             PIAU(j) = IAU(j+k)
          END DO
          PD(i)=D(i)
       end if
    END DO

    write(*,*) "number of extended rows containing plus elements: ", Ap_row_size

    !C clear list structure
    deallocate(Ap_row)
    call deallocate_dbllist(Ap_val, re_alloc_times)
    call deallocate_intlist(Ap_col, re_alloc_times)

    !C change HIERARCHICAL_DATA()
    HIERARCHICAL_DATA(LEVEL_NO-1) % D => PD
    HIERARCHICAL_DATA(LEVEL_NO-1) % INL =>PINL
    HIERARCHICAL_DATA(LEVEL_NO-1) % INU =>PINU
    HIERARCHICAL_DATA(LEVEL_NO-1) % IAL =>PIAL
    HIERARCHICAL_DATA(LEVEL_NO-1) % IAU =>PIAU
    HIERARCHICAL_DATA(LEVEL_NO-1) % AL => PAL
    HIERARCHICAL_DATA(LEVEL_NO-1) % AU => PAU
    
    HIERARCHICAL_DATA(LEVEL_NO-1) % NPL = PINL(N)
    HIERARCHICAL_DATA(LEVEL_NO-1) % NPU = PINU(N)
    
    
  contains 
    subroutine copy_Ap_to_PA()
      if(row > col) then
         L_index = L_index+1
         PIAL(L_index) = col
         PAL(L_index)  = val
      else if(row < col) then
         U_index = U_index+1
         PIAU(U_index) = col
         PAU(U_index)  = val
      else
         PD(row) = val
      end if
    end subroutine copy_Ap_to_PA
  END SUBROUTINE expand_plus_elements

  SUBROUTINE classical_interpolation_FF(LEVEL_NO, PNI, NI, node_index, coarser_level_size, NONZERO_THRSLD)
    USE data_structure_for_AMG
    IMPLICIT NONE
    INTEGER(kind=kint ), INTENT(in) :: node_index(:), PNI(0:), NI(:), coarser_level_size

    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO
    REAL(kind=kreal ), INTENT(in) :: NONZERO_THRSLD
    Type(intvec_list), target :: P_col
    Type(dblvec_list), target :: P_val

    Type(intvec_list), pointer :: P_col_ptr
    Type(dblvec_list), pointer :: P_val_ptr
    INTEGER(kind=kint) :: re_alloc_times, index_p
    
    !C To do: use ADD_C_PT in the same way as standard_interpolation_FF()
    logical :: ADD_C_PT = .false.

    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)
    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)
    INTEGER(kind=kint ) :: N

    INTEGER(kind=kint), allocatable :: fcol(:), fmap(:)
    REAL(kind=kreal),   allocatable :: cval(:), ctemp(:)
    INTEGER(kind=kint) :: csize
    REAL(kind=kreal)   :: sum
    
    INTEGER(kind=kint) :: i, j, k, col, icol
    REAL(kind=kreal)   :: val, NONZERO

    TYPE(inter_level_operator), pointer :: P
    
    N  = HIERARCHICAL_DATA(LEVEL_NO-1) % N 
    D   => HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO-1) % AU
    
    !C initialize P
    ALLOCATE(HIERARCHICAL_DATA(LEVEL_NO) % P)
    P => HIERARCHICAL_DATA(LEVEL_NO) % P
    P % ROW_SIZE = N
    ALLOCATE(P % IN(0:N)) 
    P % IN(0) = 0

    P_col_ptr => P_col
    P_val_ptr => P_val

    !C use list structure
    P_col_ptr % size = N
    allocate(P_col_ptr % intvec(N))
    allocate(P_val_ptr % dblvec(N))
    index_p = 0
    re_alloc_times = 0

    allocate(fmap(N),fcol(coarser_level_size),cval(coarser_level_size), &
         & ctemp(coarser_level_size))

    fmap = 0
    ctemp = 0

    DO i = 1, N
       icol = node_index(i)
       if(icol > 0) then
          !C i is C-point
          val = 1.0
          call insert_temp_P(icol, val, index_p, re_alloc_times, P_col_ptr, P_val_ptr, N) 
       else
          !C i is F-point

          NONZERO = NONZERO_THRSLD * dabs(D(i))

          !C fmap keeps mapping between coarse nodes' fine-node number and 
          !C the index of fcal(:), cval(:), and ctemp(:).
          !C
          !C create fmap(:), fcol(:), csize. 1 pass of i-th row.
          csize = 0
          DO j = INL(i-1)+1, INL(i)
             col = IAL(j);val =AL(j)
             k = node_index(col)
             if(k > 0 .and. dabs(val) > NONZERO) then
                !C a(i,col) \in Ci
                csize = csize+1
                fmap(col) = csize
                fcol(csize) = col
                cval(csize) = val
             end if
          END DO
          DO j = INU(i-1)+1, INU(i)
             col = IAU(j);val =AU(j)
             k = node_index(col)
             if(k > 0 .and. dabs(val) > NONZERO) then
                !C a(i,col) \in Ci
                csize = csize+1
                fmap(col) = csize
                fcol(csize) = col
                cval(csize) = val
             end if
          END DO
          
          !C 2pass
          !C create fmap(:), fcol(:), csize. 1 pass of i-th row.
          sum = 0
          DO j = INL(i-1)+1, INL(i)
             col = IAL(j);val =AL(j)
             k = node_index(col)
             !C when a(i,col) == 0, it's no use calculating col-th row.
             if(k<=0 .and. dabs(val) > NONZERO) then
                !C a(i,col) \in Fi
                !C calculate a(i,col) * a(col, j) in ctemp where j \in Ci
                !C accumulate a(col,j) in sum.
                !C then add ctemp/sum with cval(1..csize). clear ctemp at last of this sub.
                call weigh_nd_in_kth_row(col, val, csize, fmap, ctemp, cval, INL, IAL, &
                     & AL, INU, IAU, AU, D, ADD_C_PT)
             end if
          END DO
          DO j = INU(i-1)+1, INU(i)
             col = IAU(j);val =AU(j)
             k = node_index(col)
             !C when a(i,col) == 0, it's no use calculating col-th row.
             if(k<=0 .and. dabs(val) > NONZERO) then
                !C a(i,col) \in Fi
                !C calculate a(i,col) * a(col, j) in ctemp where j \in Ci
                !C accumulate a(col,j) in sum.
                !C then add ctemp/sum with cval(1..csize). clear ctemp at last of this sub
                call weigh_nd_in_kth_row(col, val, csize, fmap, ctemp, cval, INL, IAL, &
                     & AL, INU, IAU, AU, D, ADD_C_PT)
             end if
          END DO
          
          !C insert the cval to temp_p
          !C clear fmap(:) with fcol.
          DO j = 1, csize
             cval(j) = -cval(j)/D(i)
             k = fcol(j)
             call insert_temp_P(node_index(k), cval(j), index_p, re_alloc_times, P_col_ptr, &
                  & P_val_ptr, N) 
             fmap(k) = 0
          END DO
       END if
       P % IN(i) = index_p + re_alloc_times * N          
    end DO

    deallocate(fmap, fcol, cval, ctemp)    

    !C--- this part should be copied as subroutine
    !C copy & deallocate of temp_P
    allocate(P % CN(re_alloc_times * N + index_p), P % V(re_alloc_times * N + index_p))
    P_col_ptr => P_col
    P_val_ptr => P_val
    
    do i = 1, re_alloc_times
       do j = 1, N
          P % CN(j+(i-1)*N) = P_col_ptr % intvec(j)
          P % V (j+(i-1)*N) = P_val_ptr % dblvec(j)
       end do
       P_col_ptr => P_col_ptr % next
       P_val_ptr => P_val_ptr % next
    end do
    
    do i = 1, index_p
       P % CN(re_alloc_times * N + i) = P_col_ptr % intvec(i)
       P % V (re_alloc_times * N + i) = P_val_ptr % dblvec(i)
    end do

    call deallocate_dbllist(P_val, re_alloc_times)
    call deallocate_intlist(P_col, re_alloc_times)

    if(ADD_C_PT) then
       write(*,*) "There are some problematic F-F connection."
    end if


  end SUBROUTINE classical_interpolation_FF


  SUBROUTINE adaptive_interpolation_FF(LEVEL_NO, PNI, NI, node_index, coarser_level_size, NONZERO_THRSLD, &
       & prototype, ADD_NEW_C_FLAG)
    USE data_structure_for_AMG
    IMPLICIT NONE
    INTEGER(kind=kint ), INTENT(in) :: PNI(0:), NI(:), coarser_level_size
    INTEGER(kind=kint ), INTENT(inout) :: node_index(:)
    REAL   (kind=kreal), INTENT(in) :: prototype(:)
    logical, INTENT(inout) :: ADD_NEW_C_FLAG
    logical :: ADD_C_PT

    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO
    REAL(kind=kreal ), INTENT(in) :: NONZERO_THRSLD
    Type(intvec_list), target :: P_col
    Type(dblvec_list), target :: P_val

    Type(intvec_list), pointer :: P_col_ptr
    Type(dblvec_list), pointer :: P_val_ptr
    INTEGER(kind=kint) :: re_alloc_times, index_p

    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)
    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)
    INTEGER(kind=kint ) :: N

    INTEGER(kind=kint), allocatable :: fcol(:), fmap(:)
    REAL(kind=kreal),   allocatable :: cval(:), ctemp(:)
    INTEGER(kind=kint) :: csize
    REAL(kind=kreal)   :: sum
    
    INTEGER(kind=kint) :: i, j, k, col, icol
    REAL(kind=kreal)   :: val, NONZERO

    TYPE(inter_level_operator), pointer :: P
    
    N  = HIERARCHICAL_DATA(LEVEL_NO-1) % N 
    D   => HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO-1) % AU
    
    !C initialize P
    ALLOCATE(HIERARCHICAL_DATA(LEVEL_NO) % P)
    P => HIERARCHICAL_DATA(LEVEL_NO) % P
    P % ROW_SIZE = N
    ALLOCATE(P % IN(0:N)) 
    P % IN(0) = 0

    P_col_ptr => P_col
    P_val_ptr => P_val

    !C use list structure
    P_col_ptr % size = N
    allocate(P_col_ptr % intvec(N))
    allocate(P_val_ptr % dblvec(N))
    index_p = 0
    re_alloc_times = 0

    allocate(fmap(N),fcol(coarser_level_size),cval(coarser_level_size), &
         & ctemp(coarser_level_size))

    fmap = 0
    ctemp = 0

    DO i = 1, N
       icol = node_index(i)
       if(icol > 0) then
          !C i is C-point
          val = 1.0
          call insert_temp_P(icol, val, index_p, re_alloc_times, P_col_ptr, P_val_ptr, N) 
       else
          !C i is F-point

          NONZERO = NONZERO_THRSLD * dabs(D(i))

          !C fmap keeps mapping between coarse nodes' fine-node number and 
          !C the index of fcal(:), cval(:), and ctemp(:).
          !C
          !C create fmap(:), fcol(:), csize. 1 pass of i-th row.
          csize = 0
          DO j = INL(i-1)+1, INL(i)
             col = IAL(j); val = AL(j)
             k = node_index(col)
             if(k > 0 .and. dabs(val) > NONZERO) then
                !C a(i,col) \in Ci
                csize = csize+1
                fmap(col) = csize
                fcol(csize) = col
                cval(csize) = val
             end if
          END DO
          DO j = INU(i-1)+1, INU(i)
             col = IAU(j);val = AU(j)
             k = node_index(col)
             if(k > 0 .and. dabs(val) > NONZERO) then
                !C a(i,col) \in Ci
                csize = csize+1
                fmap(col) = csize
                fcol(csize) = col
                cval(csize) = val
             end if
          END DO

          !C ADD_C_PT records if this row has problematic F-F connection, or not.
          ADD_C_PT = .false.

          
          !C 2pass
          !C create fmap(:), fcol(:), csize. 1 pass of i-th row.
          sum = 0
          DO j = INL(i-1)+1, INL(i)
             col = IAL(j);val =AL(j)
             k = node_index(col)
             !C when a(i,col) == 0, it's no use calculating col-th row.
             if(k<=0 .and. dabs(val) > NONZERO) then
                !C a(i,col) \in Fi
                !C calculate a(i,col) * a(col, j) in ctemp where j \in Ci
                !C accumulate a(col,j) in sum.
                !C then add ctemp/sum with cval(1..csize). clear ctemp at last of this sub.
                call ad_weigh_nd_in_kth_row(col, val, csize, fmap, ctemp, cval, INL, IAL, &
                     & AL, INU, IAU, AU, D, prototype, ADD_C_PT)
             end if
          END DO
          DO j = INU(i-1)+1, INU(i)
             col = IAU(j);val =AU(j)
             k = node_index(col)
             !C when a(i,col) == 0, it's no use calculating col-th row.
             if(k<=0 .and. dabs(val) > NONZERO) then
                !C a(i,col) \in Fi
                !C calculate a(i,col) * a(col, j) in ctemp where j \in Ci
                !C accumulate a(col,j) in sum.
                !C then add ctemp/sum with cval(1..csize). clear ctemp at last of this sub
                call ad_weigh_nd_in_kth_row(col, val, csize, fmap, ctemp, cval, INL, IAL, &
                     & AL, INU, IAU, AU, D, prototype, ADD_C_PT)
             end if
          END DO

          !C If this row has problem in interpolation, add this row as C-point
          if(ADD_C_PT) then
             node_index(i) = -2
             ADD_NEW_C_FLAG = .true.
          end if
          
          !C insert the cval to temp_p
          !C clear fmap(:) with fcol.
          DO j = 1, csize
             cval(j) = -cval(j)/D(i)
             k = fcol(j)
             call insert_temp_P(node_index(k), cval(j), index_p, re_alloc_times, P_col_ptr, &
                  & P_val_ptr, N) 
             fmap(k) = 0
          END DO
       END if
       P % IN(i) = index_p + re_alloc_times * N          
    end DO

    deallocate(fmap, fcol, cval, ctemp)    

    !C In the ADD_NEW_C_FLAG case, don't allocate and copy the interpolation data.
    if(.not.ADD_NEW_C_FLAG) then
       
       
       !C--- this part should be copied as subroutine
       !C copy & deallocate of temp_P
       allocate(P % CN(re_alloc_times * N + index_p), P % V(re_alloc_times * N + index_p))
       P_col_ptr => P_col
       P_val_ptr => P_val

       do i = 1, re_alloc_times
          do j = 1, N
             P % CN(j+(i-1)*N) = P_col_ptr % intvec(j)
             P % V (j+(i-1)*N) = P_val_ptr % dblvec(j)
          end do
          P_col_ptr => P_col_ptr % next
          P_val_ptr => P_val_ptr % next
       end do

       do i = 1, index_p
          P % CN(re_alloc_times * N + i) = P_col_ptr % intvec(i)
          P % V (re_alloc_times * N + i) = P_val_ptr % dblvec(i)
       end do
    end if

    
    call deallocate_dbllist(P_val, re_alloc_times)
    call deallocate_intlist(P_col, re_alloc_times)

  end SUBROUTINE adaptive_interpolation_FF


  
  SUBROUTINE standard_interpolation_FF(LEVEL_NO, PNI, NI, node_index, coarser_level_size, ADD_NEW_C_FLAG)
    USE data_structure_for_AMG
    IMPLICIT NONE
    INTEGER(kind=kint ), INTENT(in)    :: PNI(0:), NI(:), coarser_level_size
    INTEGER(kind=kint ), INTENT(inout) :: node_index(:)
    logical, INTENT(inout) :: ADD_NEW_C_FLAG
    logical :: ADD_C_PT
    

    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO
    Type(intvec_list), target :: P_col
    Type(dblvec_list), target :: P_val

    Type(intvec_list), pointer :: P_col_ptr
    Type(dblvec_list), pointer :: P_val_ptr
    INTEGER(kind=kint) :: re_alloc_times, index_p

    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)
    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)
    INTEGER(kind=kint ) :: N

    INTEGER(kind=kint), allocatable :: fcol(:), fmap(:), strong_coup(:)
    REAL(kind=kreal),   allocatable :: cval(:), ctemp(:)
    INTEGER(kind=kint) :: csize
    
    INTEGER(kind=kint) :: i, j, k, col, icol
    REAL(kind=kreal)   :: val, diag

    TYPE(inter_level_operator), pointer :: P
    
    N  = HIERARCHICAL_DATA(LEVEL_NO-1) % N 
    D   => HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO-1) % AU
    
    !C initialize P
    ALLOCATE(HIERARCHICAL_DATA(LEVEL_NO) % P)
    P => HIERARCHICAL_DATA(LEVEL_NO) % P
    P % ROW_SIZE = N
    ALLOCATE(P % IN(0:N)) 
    P % IN(0) = 0

    P_col_ptr => P_col
    P_val_ptr => P_val

    !C use list structure
    P_col_ptr % size = N
    allocate(P_col_ptr % intvec(N))
    allocate(P_val_ptr % dblvec(N))
    index_p = 0
    re_alloc_times = 0

    allocate(fmap(N),fcol(coarser_level_size),cval(coarser_level_size), &
         & ctemp(coarser_level_size), strong_coup(N))

    fmap = 0
    ctemp = 0
    strong_coup = 0
    DO i = 1, N
       icol = node_index(i)
       if(icol > 0) then
          !C i is C-point
          val = 1.0
          call insert_temp_P(icol, val, index_p, re_alloc_times, P_col_ptr, P_val_ptr, N) 
       else

          !C i is F-point

          !C record ith-point's strong couplings in strong_coup(:)
          DO j = PNI(2*i-2)+1, PNI(2*i)
             k = NI(j)
             strong_coup(k) = i
          END DO

          diag = D(i)

          !C fmap keeps mapping between coarse nodes' fine-node number and 
          !C the index of fcal(:), cval(:), and ctemp(:).
          !C
          !C create fmap(:), fcol(:), csize. 1 pass of i-th row.
          csize = 0
          DO j = INL(i-1)+1, INL(i)
             col = IAL(j);val =AL(j)
             k = node_index(col)
             if(k > 0 ) then
                if(strong_coup(col) == i) then
                   !C a(i,col) \in Ci
                   csize = csize+1
                   fmap(col) = csize
                   fcol(csize) = col
                   cval(csize) = val
                else
                   diag = diag + val
                end if
             end if
             
             
          END DO

          DO j = INU(i-1)+1, INU(i)
             col = IAU(j);val =AU(j)
             k = node_index(col)
             if(k > 0 ) then
                if(strong_coup(col) == i) then
                   !C a(i,col) \in Ci
                   csize = csize+1
                   fmap(col) = csize
                   fcol(csize) = col
                   cval(csize) = val
                else
                   diag = diag + val
                end if
             end if

          END DO

          !C ADD_C_PT records if this row has problematic F-F connection, or not.
          ADD_C_PT = .false.
          

          !C 2 pass of the i-th row
          !C   strong coupling F point: accumulate weight on C-points \in Ni
          !C   weak coupling F point:   accumulate nonzero value on diag
          !C   C point : do nothing
          DO j = INL(i-1)+1, INL(i)
             col = IAL(j);val = AL(j)
             k = node_index(col)

             !C when a(i,col) == 0, it's no use calculating col-th row.
             if(k<=0 .and. val /= 0) then

                !C a(i,col) \in Fi
                if(strong_coup(col) == i) then
                   !C a(i,col) \in Fi_s
                   !C calculate a(i,col) * a(col, j) in ctemp where j \in Ci
                   !C accumulate a(col,j) in sum.
                   !C then add ctemp/sum with cval(1..csize). clear ctemp at last of this sub.
                   call weigh_nd_in_kth_row(col, val, csize, fmap, ctemp, cval, INL, IAL, &
                        & AL, INU, IAU, AU, D, ADD_C_PT)
                else
                   diag = diag + val
                end if
             end if
          END DO
          
          DO j = INU(i-1)+1, INU(i)
             col = IAU(j); val =AU(j)
             k = node_index(col)
             !C when a(i,col) == 0, it's no use calculating col-th row.
             if(k<=0 .and. val /= 0) then
                !C a(i,col) \in Fi
                if(strong_coup(col) == i) then
                   !C a(i,col) \in Fi_s
                   !C calculate a(i,col) * a(col, j) in ctemp where j \in Ci
                   !C accumulate a(col,j) in sum.
                   !C then add ctemp/sum with cval(1..csize). clear ctemp at last of this sub
                   call weigh_nd_in_kth_row(col, val, csize, fmap, ctemp, cval, INL, IAL, &
                        & AL, INU, IAU, AU, D, ADD_C_PT)
                else
                   diag = diag + val
                end if
             end if
          END DO

          !C If this row has problem in interpolation, add this row as C-point
          if(ADD_C_PT) then
             node_index(i) = -2
             ADD_NEW_C_FLAG = .true.
          end if

          !C insert the cval to temp_p
          !C clear fmap(:) with fcol.
          DO j = 1, csize
             
             cval(j) = -cval(j)/diag
             k = fcol(j)
             call insert_temp_P(node_index(k), cval(j), index_p, re_alloc_times, P_col_ptr, &
                  & P_val_ptr, N) 
             fmap(k) = 0
          END DO
             
       END if
       P % IN(i) = index_p + re_alloc_times * N          

    end DO

    deallocate(fmap, fcol, cval, ctemp, strong_coup)    

    !C In the ADD_NEW_C_FLAG case, don't allocate and copy the interpolation data.
    if(.not.ADD_NEW_C_FLAG) then
       !C--- this part should be copied as subroutine
       !C copy & deallocate of temp_P
       allocate(P % CN(re_alloc_times * N + index_p), P % V(re_alloc_times * N + index_p))
       P_col_ptr => P_col
       P_val_ptr => P_val
    
       do i = 1, re_alloc_times
          do j = 1, N
             P % CN(j+(i-1)*N) = P_col_ptr % intvec(j)
             P % V (j+(i-1)*N) = P_val_ptr % dblvec(j)
          end do
          P_col_ptr => P_col_ptr % next
          P_val_ptr => P_val_ptr % next
       end do
    
       do i = 1, index_p
          P % CN(re_alloc_times * N + i) = P_col_ptr % intvec(i)
          P % V (re_alloc_times * N + i) = P_val_ptr % dblvec(i)
       end do
    end if

    call deallocate_dbllist(P_val, re_alloc_times)
    call deallocate_intlist(P_col, re_alloc_times)

  end SUBROUTINE standard_interpolation_FF


  subroutine weigh_nd_in_kth_row(row, val, csize, fmap, ctemp, cval, INL, IAL, &
       & AL, INU, IAU, AU, D, ADD_C_PT)
    implicit none
    include 'precision.inc'
    
    INTEGER(kind=kint), INTENT(in) :: INL(0:),INU(0:),IAL(:),IAU(:),fmap(:),csize,row
    REAL(kind=kreal),   INTENT(in) :: AL(:),AU(:),D(:),val
    REAL(kind=kreal), INTENT(inout) :: ctemp(:), cval(:)
    logical, INTENT(inout) :: ADD_C_PT

    REAL(kind=kreal)   :: sum, v, max
    INTEGER(kind=kint) :: i,j,k, c 

    sum = 0
    !C max stores max value
    max = 0

    DO j = INL(row-1)+1, INL(row)
       c = IAL(j);  k = fmap(c)
!!$       c = IAL(j);  k = fmap(c); v = AL(j) 
!!$       if(k>0 .and. v < 0) then
       if(k>0) then
          v = AL(j) 
          sum = sum+v
          ctemp(k) = v

          if(max < dabs(v)) max = dabs(v)

       end if
    END DO
    DO j = INU(row-1)+1, INU(row)
       c = IAU(j); k = fmap(c)
!!$       c = IAU(j); k = fmap(c); v = AU(j)
!!$       if(k>0 .and. v<0) then
       if(k>0) then
          v = AU(j)
          sum = sum+v
          ctemp(k) = v

          if(max < dabs(v)) max = dabs(v)
       end if
    END DO

    !C In the case of sum == 0,  error? or not.
    if(dabs(sum/max) <  1.0E-8) then
       write(*,*) ctemp(1:csize), sum
       
       !C 'Error : F-F connections without any common C-point'
       !C add F-point of row as C point.
       ADD_C_PT = .true.
    end if
    
    !C clear ctemp(:)
    DO j = 1, csize
       cval(j) = cval(j) + (ctemp(j)/sum) * val  
       ctemp(j) = 0
    END DO

  end subroutine weigh_nd_in_kth_row
  
  !C max variable is added to check orthogonal relation with prototype vector.
  subroutine ad_weigh_nd_in_kth_row(row, val, csize, fmap, ctemp, cval, INL, IAL, &
       & AL, INU, IAU, AU, D, prototype, ADD_C_PT)
    implicit none
    include 'precision.inc'
    
    INTEGER(kind=kint), INTENT(in) :: INL(0:),INU(0:),IAL(:),IAU(:),fmap(:),csize,row
    REAL(kind=kreal),   INTENT(in) :: AL(:),AU(:),D(:),val,prototype(:)
    REAL(kind=kreal), INTENT(inout) :: ctemp(:), cval(:)
    logical, INTENT(inout) :: ADD_C_PT

    REAL(kind=kreal)   :: sum, v,w, max, vk
    INTEGER(kind=kint) :: i,j,k, c 

    sum = 0
    !C max stores max value
    max = 0

    vk = prototype(row)
    DO j = INL(row-1)+1, INL(row)
       c = IAL(j);  k = fmap(c); v = AL(j)
!!$       c = IAL(j);  k = fmap(c); v = AL(j) 
!!$       if(k>0 .and. v < 0) then
       if(k>0) then
          sum = sum + v * prototype(c)
          w   =  v * vk
          ctemp(k) = w
          
          if(max < dabs(w)) max = dabs(w)

       end if
    END DO
    DO j = INU(row-1)+1, INU(row)
       c = IAU(j); k = fmap(c); v = AU(j)
!!$       c = IAU(j); k = fmap(c); v = AU(j)
!!$       if(k>0 .and. v<0) then
       if(k>0) then
          sum = sum + v * prototype(c)
          w   = v * vk
          ctemp(k) = w
          
          if(max < dabs(w)) max = dabs(w)
       end if
    END DO

    !C In the case of sum == 0,  error? or not.
    if(dabs(sum/max) <  1.0E-8) then
       write(*,*) csize, ":", ctemp(1:csize), sum
       !C 'Error : F-F connections which is orthogonal with  prototype vector'
       !C add F-point of row as C point.
       ADD_C_PT = .true.

    end if
    
    !C clear ctemp(:)
    DO j = 1, csize
       cval(j) = cval(j) + (ctemp(j)/sum) * val  
       ctemp(j) = 0
    END DO

  end subroutine ad_weigh_nd_in_kth_row
  

  SUBROUTINE standard_interpolation(LEVEL_NO, PNI, NI, node_index, epsilon)
    USE data_structure_for_AMG
    IMPLICIT NONE
    INTEGER(kind=kint ), INTENT(in) :: node_index(:), PNI(0:), NI(:)
    REAL(kind=kreal), INTENT(in) :: epsilon

    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO
    Type(intvec_list), target :: P_col
    Type(dblvec_list), target :: P_val

    Type(intvec_list), pointer :: P_col_ptr
    Type(dblvec_list), pointer :: P_val_ptr


    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)
    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)
    INTEGER(kind=kint ) :: N

    INTEGER(kind=kint), allocatable :: q(:), f_ivec(:)
    REAL(kind=kreal),   allocatable :: qr(:), f_rvec(:)
    INTEGER(kind=kint ) :: qsize
    REAL(kind=kreal)    :: diag

    INTEGER(kind=kint ) :: i,j,k,l,m,index_p, re_alloc_times
    INTEGER(kind=kint ) :: icol, col, sk
    REAL(kind=kreal) :: val, v

    TYPE(inter_level_operator), pointer :: P
    
    N  = HIERARCHICAL_DATA(LEVEL_NO-1) % N 
    D   => HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO-1) % AU
    
    !C initialize P
    ALLOCATE(HIERARCHICAL_DATA(LEVEL_NO) % P)
    P => HIERARCHICAL_DATA(LEVEL_NO) % P
    P % ROW_SIZE = N
    ALLOCATE(P % IN(0:N)) 
    P % IN(0) = 0

    P_col_ptr => P_col
    P_val_ptr => P_val

    !C use list structure
    P_col_ptr % size = N
    allocate(P_col_ptr % intvec(N))
    allocate(P_val_ptr % dblvec(N))
    index_p = 0
    re_alloc_times = 0

    allocate(q(N),qr(N),f_ivec(N),f_rvec(N))
    f_ivec = 0
    
    DO i = 1, N
       icol = node_index(i)
       if(icol > 0) then
          !C i is C-point
          val = 1.0
          call insert_temp_P(icol, val, index_p, re_alloc_times, P_col_ptr, P_val_ptr, N) 
       else
          !C i is F-point
          !C expand the row of j in NI(PNI(i-1)+1:PNI(i)) on f_ivec, f_rvec, then
          !C pack the data into q(), and qr()
          qsize = 0
          DO j = INL(i-1)+1, INL(i)
             col = IAL(j);val =AL(j)
             qsize = qsize+1
             q(qsize) = col; f_ivec(col) = j; 
             f_rvec(col) = val
          END DO
          DO j = INU(i-1)+1, INU(i)
             col = IAU(j);val =AU(j)
             qsize = qsize + 1
             q(qsize) = col; f_ivec(col) = j; 
             f_rvec(col) = val
          END DO
          !C 1 is dammy value. positive value shows i-th element is already recorded..
          f_ivec(i) = 1; f_rvec(i) = D(i)

          DO j = PNI(2*i-2)+1, PNI(2*i)
             k = NI(j)
             !C will-be eliminated non-zero element set as 0
             if(node_index(k) < 0) f_rvec(k) = 0
          END DO

          !C expand for F-F couplings
          DO j = PNI(2*i-2)+1, PNI(2*i)
             !C that element is (i,k) == col=>IA?(m), val=>A?(m)
             k = NI(j); m = f_ivec(k)
             if(node_index(k) <= 0) then
                !C expand k-th row
                !C v is the associated node's value
                if(j > PNI(2*i-1))then
                   !C-- in case of m=0, set v = 0
                   v = AU(m)
                else
                   v = AL(m)
                end if
                
                call expand_by_kth_row(v, k, f_ivec, f_rvec, q, qsize, INL, IAL, AL, INU,&
                     & IAU, AU, D)
                
             end if
          END DO

          !C diagonal element is treated first.
          diag = f_rvec(i); f_rvec(i) = 0; f_ivec(i) = 0;
          
          !C in order to skip the 0 elememts, count those elements
          sk = 0
          DO j = 1, qsize
             !C reset f_ivec
             col = q(j)
             
             f_ivec(col) = 0
             val = f_rvec(col)
             if(val == 0) then
                sk = sk + 1
             else
                qr(j-sk) = val
                q (j-sk) = col
             end if
             
          END DO
          qsize = qsize - sk

          !C diag may be changed in the following subroutine.
          call interpolate_f_point(qsize, q, qr, diag, node_index, index_p, &
               & re_alloc_times, P_col_ptr, P_val_ptr, N, epsilon)

       end if
       P % IN(i) = index_p + re_alloc_times * N
       
    END DO

    deallocate(q, qr, f_ivec, f_rvec)

    !C copy & deallocate
    allocate(P % CN(re_alloc_times * N + index_p), P % V(re_alloc_times * N + index_p))
    P_col_ptr => P_col
    P_val_ptr => P_val
    
    do i = 1, re_alloc_times
       do j = 1, N
          P % CN(j+(i-1)*N) = P_col_ptr % intvec(j)
          P % V (j+(i-1)*N) = P_val_ptr % dblvec(j)
       end do
       P_col_ptr => P_col_ptr % next
       P_val_ptr => P_val_ptr % next
    end do
    
    do i = 1, index_p
       P % CN(re_alloc_times * N + i) = P_col_ptr % intvec(i)
       P % V (re_alloc_times * N + i) = P_val_ptr % dblvec(i)
    end do

    call deallocate_dbllist(P_val, re_alloc_times)
    call deallocate_intlist(P_col, re_alloc_times)

  END SUBROUTINE standard_interpolation

  
  SUBROUTINE direct_interpolation(LEVEL_NO, node_index)
    USE data_structure_for_AMG
    IMPLICIT NONE
    INTEGER(kind=kint ), INTENT(in) :: node_index(:)

    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO

    INTEGER(kind=kint ) :: index_p, re_alloc_times
    Type(intvec_list), target :: P_col
    Type(dblvec_list), target :: P_val

    Type(intvec_list), pointer :: P_col_ptr
    Type(dblvec_list), pointer :: P_val_ptr

    TYPE(inter_level_operator), pointer :: P
    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)

    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)
    INTEGER(kind=kint ) :: N

    INTEGER(kind=kint), allocatable :: q(:)
    REAL(kind=kreal),   allocatable :: qr(:)
    INTEGER(kind=kint ) :: qsize
    REAL(kind=kreal)    :: diag

    INTEGER(kind=kint ) :: i,j,icol
    REAL(kind=kreal)    :: val

    N  = HIERARCHICAL_DATA(LEVEL_NO-1) % N 
    D   => HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO-1) % AU
    
    !C initialize P
    ALLOCATE(HIERARCHICAL_DATA(LEVEL_NO) % P)
    P => HIERARCHICAL_DATA(LEVEL_NO) % P
    P % ROW_SIZE = N
    ALLOCATE(P % IN(0:N)) 
    P % IN(0) = 0

    P_col_ptr => P_col
    P_val_ptr => P_val

    !C use list structure
    P_col_ptr % size = N
    allocate(P_col_ptr % intvec(N))
    allocate(P_val_ptr % dblvec(N))
    index_p = 0
    re_alloc_times = 0
    
    allocate(q(N),qr(N))

    DO i = 1, N
       icol = node_index(i)
       if(icol > 0) then
          !C i is C-point
          val = 1.0
          call insert_temp_P(icol, val, index_p, re_alloc_times, P_col_ptr, P_val_ptr, N) 
       else
          !C i is F-point
    
          !C copy the row data into q(),qr()with diag
          qsize = 0
          DO j = INL(i-1)+1, INL(i)
             qsize = qsize+1
             q(qsize) = IAL(j);qr(qsize) = AL(j)

          END DO
          DO j = INU(i-1)+1, INU(i)
             qsize = qsize+1
             q(qsize) = IAU(j);qr(qsize) = AU(j)

          END DO
          diag = D(i)

          !C diag may be changed in the following subroutine.
          val = 0.0
          call interpolate_f_point(qsize, q, qr, diag, node_index, index_p, &
               & re_alloc_times, P_col_ptr, P_val_ptr, N, val)
          
       end if
       P % IN(i) = index_p + re_alloc_times * N
    END DO

    !C copy & deallocate
    allocate(P % CN(re_alloc_times*N+index_p), P % V(re_alloc_times*N+index_p))
    P_col_ptr => P_col
    P_val_ptr => P_val
    
    do i = 1, re_alloc_times
       do j = 1, N
          P % CN(j+(i-1)*N) = P_col_ptr % intvec(j)
          P % V(j+(i-1)*N) = P_val_ptr % dblvec(j)
       end do
       P_col_ptr => P_col_ptr % next
       P_val_ptr => P_val_ptr % next
    end do
    
    do i = 1, index_p
       P % CN(re_alloc_times * N + i) = P_col_ptr % intvec(i)
       P % V (re_alloc_times * N + i) = P_val_ptr % dblvec(i)
    end do

    !C deallocate
    call deallocate_dbllist(P_val, re_alloc_times)
    call deallocate_intlist(P_col, re_alloc_times)

  END SUBROUTINE direct_interpolation

  SUBROUTINE interpolate_f_point(qsize, q, qr, diag, node_index, index_p, &
       & re_alloc_times, P_col_ptr, P_val_ptr, N, epsilon)
    USE data_structure_for_AMG
    implicit none
    INTEGER(kind=kint), intent(in) :: node_index(:), N
    INTEGER(kind=kint), intent(inout) :: qsize, q(:)
    REAL(kind=kreal),   intent(inout) :: qr(:)
    REAL(kind=kreal),   intent(in)    :: epsilon
    REAL(kind=kreal),   intent(inout) :: diag
    
    INTEGER(kind=kint), intent(inout) :: index_p, re_alloc_times

    Type(intvec_list), pointer :: P_col_ptr
    Type(dblvec_list), pointer :: P_val_ptr

    
    INTEGER(kind=kint) :: j,col
    REAL(kind=kreal) :: val, sum_c, sum_cp, sum, sum_p
    REAL(kind=kreal) :: sum_tr, sum_tr_p, max_p, max
    REAL(kind=kreal) :: alpha, alpha_p
    
    sum = 0; sum_p = 0; sum_c = 0; sum_cp = 0;max_p=0; max = 0
    sum_tr_p = 0; sum_tr = 0;
    DO j = 1, qsize
       col = node_index(q(j))
       q(j) = col
       val = qr(j)
       if(val > 0) then
          sum_p = sum_p + val
       else
          sum = sum + val
       end if
       if(col > 0) then 
          !C C-points
          if(val > 0) then
             sum_cp = sum_cp + val
             if(max_p < val) max_p = val
          else
             sum_c = sum_c + val
             if(max > val) max = val
          end if
       end if
    END DO


    DO j = 1, qsize
       col = q(j)
       if(col > 0) then
          val = qr(j)
          if(val > 0) then
             !C positive C-point
             if(val < max_p*epsilon) then
                sum_tr_p = sum_tr_p+val
                !C truncation
                q(j) = -1
             end if
          else
             !C negative C-point
             if(val > max*epsilon) then
                sum_tr = sum_tr+val
                !C truncation
                q(j) = -1
             end if
          end if
       end if
    END DO


    if(sum_cp == 0) then
       !C No positive C-points case
       diag = diag + sum_p
    else
       !C some positive C-points case
       alpha_p = -sum_p/(sum_cp-sum_tr_p)/diag
    end if

    !C some negative C-points case
    if(sum_c /= 0) alpha = -sum/(sum_c-sum_tr)/diag
    
    if(sum /= 0 .or. sum_p /= 0) then
       !C 2 pass
       DO j = 1, qsize
          col = q(j)
          if(col > 0) then
             val = qr(j)
             if(val > 0) then
                !C positive C-point case
                val = alpha_p*val
             else
                !C negative C-point case
                val = alpha * val
             end if
             
             call insert_temp_P(col, val, index_p, re_alloc_times, P_col_ptr, P_val_ptr, N)
             
          end if
       END DO
    end if
    
  end subroutine interpolate_f_point

  subroutine deallocate_dbllist(dbl_vec_list, re_alloc_times)
    use data_structure_for_AMG
    implicit none
    Type(dblvec_list), target  :: dbl_vec_list
    Type(dblvec_list), pointer :: dbl_vec_list_ptr,dbl_vec_list_ptr2
    integer(kind=kint) :: i, re_alloc_times

    deallocate(dbl_vec_list % dblvec)    
    dbl_vec_list_ptr => dbl_vec_list % next
    do i = 1, re_alloc_times
       dbl_vec_list_ptr2 => dbl_vec_list_ptr % next
       deallocate(dbl_vec_list_ptr % dblvec)
       deallocate(dbl_vec_list_ptr)
       dbl_vec_list_ptr => dbl_vec_list_ptr2
    end do
  end subroutine deallocate_dbllist

  subroutine deallocate_intlist(int_vec_list, re_alloc_times)
    use data_structure_for_AMG
    implicit none
    Type(intvec_list), target  :: int_vec_list
    Type(intvec_list), pointer :: int_vec_list_ptr,int_vec_list_ptr2
    integer(kind=kint) :: i, re_alloc_times
    
    deallocate(int_vec_list % intvec)
    int_vec_list_ptr => int_vec_list % next
    do i = 1, re_alloc_times
       int_vec_list_ptr2 => int_vec_list_ptr % next
       deallocate(int_vec_list_ptr % intvec)
       deallocate(int_vec_list_ptr)
       int_vec_list_ptr => int_vec_list_ptr2
    end do
  END SUBROUTINE deallocate_intlist
  
  SUBROUTINE multipath_interpolation(LEVEL_NO, PNIT, NIT, PNI, NI, node_index, coarser_level_size)
    USE data_structure_for_AMG
    IMPLICIT NONE
    INTEGER(kind=kint), intent(in) :: LEVEL_NO, PNIT(0:), NIT(:), PNI(0:), NI(:)
    INTEGER(kind=kint), intent(in) :: node_index(:), coarser_level_size
    
    INTEGER(kind=kint) :: i,j,k,l,m,node,js,je,index_p,re_alloc_times, qsize, ex_node, col, row
    REAL(kind=kreal)   :: sum, sum_p, sum_c, sum_cp, diag, val

    INTEGER(kind=kint), allocatable :: layer_node(:), index_layer(:), node_order(:)
    INTEGER(kind=kint), allocatable :: f_ivec(:), P_IN(:), q(:), cf_vec(:)
    REAL(kind=kreal),   allocatable :: f_rvec(:)
    INTEGER(kind=kint) :: layer_size, layer_nd_sz, isolated_points, MAXLAYER=10

    Type(intvec_list), target :: P_col
    Type(dblvec_list), target :: P_val

    Type(intvec_list), pointer :: P_col_ptr
    Type(dblvec_list), pointer :: P_val_ptr
    TYPE(inter_level_operator), pointer :: P

    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)
    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)
    INTEGER(kind=kint ) :: N
    
    N  = HIERARCHICAL_DATA(LEVEL_NO-1) % N 
    D   => HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO-1) % AU

    !C calculate and check not-C isolated points.  
    !C node_order(?) = -1 for isolated point
    allocate(node_order(N))
    
    node_order = 0
    isolated_points = 0
    DO i = 1, N
!!$       k = PNI(2*i) - PNI(2*i-2) + PNIT(2*i) - PNIT(2*i-2)
       k = PNI(2*i) - PNI(2*i-2)
       if(k==0 .and. node_index(i) <= 0) then
          isolated_points = isolated_points + 1
          node_order(i) = -1
       end if
    END DO
    
    !C layer check
    !C node_order(?) shows order in layer_node
    allocate(layer_node(N), index_layer(0:MAXLAYER))
    index_layer(0) = 0; layer_nd_sz=0

    !C layer 1
    DO i = 1, N
       if(node_index(i) > 0) then
          layer_nd_sz = layer_nd_sz + 1
          layer_node(layer_nd_sz) = i
          node_order(i) = layer_nd_sz
       end if
    END DO
    index_layer(1) = layer_nd_sz    

    !C layer 2...MAXLAYER
    DO layer_size = 2, MAXLAYER
       DO i = index_layer(layer_size-2)+1, index_layer(layer_size-1)
          node = layer_node(i)
          DO j = PNIT(2*node-2)+1, PNIT(2*node)
             k = NIT(j)
             !C k 
             if(node_order(k) == 0) then
                layer_nd_sz = layer_nd_sz + 1
                layer_node(layer_nd_sz) = k
                node_order(k) = layer_nd_sz
             end if
          END DO
       END DO
       if(layer_nd_sz == index_layer(layer_size-1)) exit
       index_layer(layer_size) = layer_nd_sz
    END DO
    
    if(layer_size == MAXLAYER) stop "MAXLAYER should be enlarged."
!    write(*,*) "#######", layer_size, "########"

    layer_size = layer_size - 1    

    if(layer_nd_sz + isolated_points /= N) then
       write(*,*) "Error!", layer_nd_sz, isolated_points, N
       stop "not-C isolated points + C + F /= N"
    end if

    !C calculate interpolation for each layer
    allocate(f_rvec(N), f_ivec(N), q(N), cf_vec(coarser_level_size))
    allocate(P_col % intvec(N), P_val % dblvec(N), P_IN(0:N))
    P_col_ptr => P_col; P_val_ptr => P_val; P_IN(0) = 0

    !C create cf_vec
    DO i = 1, N
       if(node_index(i) > 0) cf_vec(node_index(i)) = i
    END DO

    !C layer 1: the node is C-point
    DO j = 1, index_layer(1)
       node = layer_node(j)
       P_col_ptr % intvec(j) = node_index(node)
       P_val_ptr % dblvec(j) = 1.0
       P_IN(j) = P_IN(j-1) + 1
    END DO
    index_p = P_IN(index_layer(1)); re_alloc_times = 0

    !C layer 2..: the node is F-point
    f_ivec = 0
    DO i = 2, layer_size
       js = index_layer(i-1)+1; je = index_layer(i)
       DO j = js, je
          node = layer_node(j)
          qsize = 0
          
          DO k = PNI(2*node-2)+1, PNI(2*node)
             ex_node = NI(k)
             if(node_order(ex_node) < js .and. node_index(ex_node) < 1) then
                !C only check F-point for expansion
                !C expansion mark is "j" for jth row of layer_node(:)
                f_ivec(ex_node) = j
             end if
          end DO

          DO k = INL(node-1)+1, INL(node)
             col = IAL(k); val = AL(k)

             if(f_ivec(col) == j .and. node_index(col) < 1) then
                !C F-point's value wont be changed even after expansion
                !C This shows the second time accessing F-point is marked
                !C for expansion

                !C expansion. Expanded non-zero elements are assumed to be C-points.
                call expand(j, col, val, f_rvec, f_ivec, qsize, q, P_IN, P_col, P_val, &
                     & node_order, cf_vec, N)
             else
                !C assgin a non-zero value
                if(f_ivec(col) /= j) then
                   f_ivec(col) = j 
                   qsize = qsize + 1; q(qsize) = col
                   f_rvec(col) = val
                else
                   f_rvec(col) = f_rvec(col)+val
                end if
             end if
          END DO

          DO k = INU(node-1)+1, INU(node)
             col = IAU(k); val = AU(k)
             if(f_ivec(col) == j .and. node_index(col) < 1) then
                !C expansion
                call expand(j, col, val, f_rvec, f_ivec, qsize, q, P_IN, P_col, P_val, &
                     & node_order, cf_vec, N)
             else
                !C assgin a non-zero value
                if(f_ivec(col) /= j) then
                   f_ivec(col) = j
                   qsize = qsize + 1; q(qsize) = col
                   f_rvec(col) = val
                else
                   f_rvec(col) = f_rvec(col) + val
                end if
             end if
          END DO
          
          !C calculation P elements
          sum_p = 0; sum = 0; sum_c = 0; sum_cp = 0
          DO k = 1, qsize 
             l = q(k)
             val = f_rvec(l)
             col = node_index(l)
             if(val > 0) then
                sum_p = sum_p + val
             else
                sum = sum + val
             end if
             if(col > 0) then
                if(val > 0) then
                   sum_cp = sum_cp + val
                else
                   sum_c = sum_c + val
                end if
             end if
          END DO
          if(sum_cp > 0) then
             diag = D(node)
          else
             diag = D(node) + sum_p
          end if
          DO k = 1, qsize
             l = q(k)
             col = node_index(l)
             val = f_rvec(l) 
             if(col > 0) then
                if(val > 0) then
                   val = -sum_p/sum_cp*val/diag
                else
                   val = -sum/sum_c*val/diag
                end if
                call insert_temp_P(col, val, index_p, re_alloc_times, P_col_ptr, P_val_ptr, N)
             end if
          end DO

          P_IN(j) = index_p + re_alloc_times * N
       END DO
    END DO

    !C copy 
    !C initialize P
    ALLOCATE(HIERARCHICAL_DATA(LEVEL_NO) % P)
    P => HIERARCHICAL_DATA(LEVEL_NO) % P
    P % ROW_SIZE = N
    ALLOCATE(P % IN(0:N)) 
    P % IN(0) = 0
    allocate(P % CN(re_alloc_times*N+index_p), P % V(re_alloc_times*N+index_p))

    !C P % IN(:) is calculated
    DO i = 1, N
       !C node_order(?) = -1 for isolated point
       if(node_order(i) == -1) then
          P % IN(i) = P % IN(i-1)
       else
          j = node_order(i)
          P % IN(i) = P_IN(j) - P_IN(j-1) + P % IN(i-1)
       end if
    END DO

    !C P % CN, V are calculated
    !C transpose node_order(:) on f_ivec as temporary space
    DO i = 1, N
       if(node_order(i) > 0) then
          f_ivec(node_order(i)) = i
       end if
    END DO
    P_col_ptr => P_col
    P_val_ptr => P_val
    DO i = 1, layer_nd_sz
       !C row is the row num in P.
       row = f_ivec(i); l = P % IN(row - 1)
       DO j = P_IN(i-1) + 1, P_IN(i)
          !C j is changed to the range of 1..N 
          k = mod(j - 1, N) + 1
          if(k == 1 .and. j > N) then
             P_col_ptr => P_col_ptr % next
             P_val_ptr => P_val_ptr % next
          end if

          l = l+1
          P % CN(l) = P_col_ptr % intvec(k)
          P % V (l) = P_val_ptr % dblvec(k)
       END DO
    END DO

    !C deallocate
    deallocate(P_IN, f_rvec, f_ivec, q, layer_node, index_layer, node_order)

    call deallocate_intlist(P_col, re_alloc_times)
    call deallocate_dbllist(P_val, re_alloc_times)
    
  END SUBROUTINE multipath_interpolation
  
  subroutine insert_temp_P(col, val, index_p, re_alloc_times, P_col_ptr, P_val_ptr, N)
    USE data_structure_for_AMG
    IMPLICIT none
    INTEGER(kind=kint) :: col, index_p, re_alloc_times, N
    REAL(kind=kreal)   :: val

    TYPE(intvec_list), pointer :: P_col_ptr
    TYPE(dblvec_list), pointer :: P_val_ptr
    
    index_p = index_p + 1
    if(index_p > N) then
       !C reallocate temporary list of P_col & P_val
       index_p = index_p - N
       re_alloc_times = re_alloc_times + 1                
       allocate(P_col_ptr % next, P_val_ptr % next)
       P_col_ptr => P_col_ptr % next; P_val_ptr => P_val_ptr % next
       allocate(P_col_ptr % intvec(N), P_val_ptr % dblvec(N))
    end if
    
    P_col_ptr % intvec(index_p) = col
    P_val_ptr % dblvec(index_p) = val

  end subroutine insert_temp_P
  
  subroutine expand(mark, col, val, f_rvec, f_ivec, qsize, q, P_IN, P_col, P_val, node_order, cf_vec, N)
    USE data_structure_for_AMG
    IMPLICIT none
    
    INTEGER(kind=kint) :: mark, col, f_ivec(:), qsize, q(:), P_IN(0:), node_order(:), cf_vec(:), N
    REAL(kind=kreal)  :: val, f_rvec(:)
    TYPE(intvec_list), target :: P_col
    TYPE(dblvec_list), target :: P_val
    
    TYPE(intvec_list), pointer :: P_col_ptr
    TYPE(dblvec_list), pointer :: P_val_ptr
    INTEGER(kind=kint) :: i, j, k, l, c
    REAL(kind=kreal)   :: v

    P_col_ptr => P_col; P_val_ptr => P_val    

    !C l is the number in layer_node()
    l = node_order(col)
    
    !C j shows the number of vectors before the target data
    j = P_IN(l-1) / N
    DO i = 1, j
       P_col_ptr => P_col_ptr % next; P_val_ptr => P_val_ptr % next
    END DO

    k = (P_IN(l)-1) / N
    if(k > j) then
       !C the target row data is divided into to vectors
       DO i = mod(P_IN(l-1),  N) + 1, N
          call store_in_vec()
       END DO
       P_col_ptr => P_col_ptr % next; P_val_ptr => P_val_ptr % next       
       DO i = 1, mod(P_IN(l)-1, N) + 1
          call store_in_vec()
       END DO
    else
       !C the target row data is stored in one vector
       DO i = mod(P_IN(l-1), N) + 1, mod(P_IN(l)-1, N) + 1
          call store_in_vec()
       END DO
    end if
    contains 
      subroutine store_in_vec()
          c = cf_vec(P_col_ptr % intvec(i)); v = P_val_ptr % dblvec(i)
          if(f_ivec(c) /= mark) then
             f_ivec(c) = mark
             qsize = qsize + 1; q(qsize) = c
             f_rvec(c) =  val * v
          else
             f_rvec(c) = f_rvec(c) + val * v
          end if
        end subroutine store_in_vec
    end subroutine expand

  SUBROUTINE jacobi_relaxation(LEVEL_NO, coarser_level_size, node_index, epsilon)
    USE data_structure_for_AMG
    IMPLICIT NONE
    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO, coarser_level_size, node_index(:)
    REAL   (kind=kreal), INTENT(in) :: epsilon
    
    Type(intvec_list), target :: P_col
    Type(dblvec_list), target :: P_val
    Type(intvec_list), pointer :: P_col_ptr
    Type(dblvec_list), pointer :: P_val_ptr

    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)
    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)
    INTEGER(kind=kint ) :: N

    TYPE(inter_level_operator), pointer :: P

    INTEGER(kind=kint ), allocatable :: c_ivec(:), q(:), Temp_P_IN(:)
    REAL(kind=kreal ),   allocatable :: c_rvec(:), qr(:)
    INTEGER(kind=kint ) :: i,j,k,l,col,index_p,re_alloc_times,qsize
    INTEGER(kind=kint ) :: col2
    REAL(kind=kreal) :: val,val2,alpha_p,alpha_m
    REAL(kind=kreal) :: max_p,max_m, sum_p,sum_m, sum_tr_p, sum_tr_m
    


    N  = HIERARCHICAL_DATA(LEVEL_NO-1) % N 
    D   => HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO-1) % AU

    P => HIERARCHICAL_DATA(LEVEL_NO) % P
    
    P_col_ptr => P_col
    P_val_ptr => P_val
    P_col_ptr % size = N
    allocate(P_col_ptr % intvec(N))
    allocate(P_val_ptr % dblvec(N))

    allocate(c_ivec(coarser_level_size),c_rvec(coarser_level_size),&
         &   q(coarser_level_size),qr(coarser_level_size),Temp_P_IN(N))
    c_ivec = 0 
    qsize = 0
    index_p = 0
    re_alloc_times = 0


    DO i = 1, N
       col = node_index(i)
       if(col > 0) then
          val = 1.0
          call insert_temp_P(col, val, index_p, re_alloc_times, P_col_ptr, P_val_ptr, N)            
       else
          !C i is F-point
          DO j = INL(i-1)+1, INL(i)
             k = IAL(j); col = node_index(k); val = AL(j)
             DO l = P % IN(k-1)+1, P % IN(k)
                col2 = P % CN(l)
                val2 = P % V(l)
                if(c_ivec(col2) /= i) then
                   !C c_vec(:)
                   c_ivec(col2) = i; c_rvec(col2) = val*val2
                   !C q(:)
                   qsize = qsize + 1; q(qsize) = col2
                else
                   !C c_vec(:)
                   c_rvec(col2) = c_rvec(col2) + val*val2
                end if
             END DO
          END DO

          DO j = INU(i-1)+1, INU(i)
             k = IAU(j); col = node_index(k); val = AU(j)

             DO l = P % IN(k-1)+1, P % IN(k)
                col2 = P % CN(l)
                val2 = P % V(l)
                if(c_ivec(col2) /= i) then
                   !C c_vec(:)
                   c_ivec(col2) = i; c_rvec(col2) = val*val2
                   !C q(:)
                   qsize = qsize + 1; q(qsize) = col2
                else
                   !C c_vec(:)
                   c_rvec(col2) = c_rvec(col2) + val*val2
                end if
             END DO
          END DO

          if(qsize > 0) then
             sum_p=0;sum_m=0; sum_tr_p=0; sum_tr_m=0; max_p=0;max_m=0
             !C 2 pass sum and max value is calculated and 
             !C data is shifted into continuous space.
             DO j = 1, qsize
                col = q(j); val = -c_rvec(col)/D(i)
                qr(j) = val
                if(val > 0) then
                   sum_p = sum_p + val
                   if(max_p<val) max_p = val
                else
                   sum_m = sum_m + val
                   if(max_m>val) max_m = val
                end if
             END DO

             !C 3 pass on continuous space
             DO j = 1, qsize
                val = qr(j)
                if(val > 0) then
                   if(val < max_p*epsilon) then
                      sum_tr_p = sum_tr_p+val
                       qr(j) = 0
                   end if
                else
                   if(val > max_m*epsilon) then
                      sum_tr_m = sum_tr_m+val
                      qr(j) = 0
                   end if
                end if
             END DO
             alpha_p = sum_p/(sum_p-sum_tr_p)
             alpha_m = sum_m/(sum_m-sum_tr_m)

             !C 3 pass
             DO j = 1, qsize
                col = q(j); val = qr(j)

                if(val /= 0) then
                   if(val>0) then
                      val = alpha_p*val
                   else
                      val = alpha_m*val
                   end if

                   call insert_temp_P(col, val, index_p, re_alloc_times, P_col_ptr, P_val_ptr, N)        
                end if
             END DO
          end if
          qsize = 0
       end if
       TEMP_P_IN(i) = index_p + re_alloc_times * N
    END DO

    !C reallocate and chage P(:)
    deallocate(c_ivec,c_rvec,q,qr,   &
         & P % CN, P % V)
    
    i = Temp_P_IN(N)
    allocate(P % CN(i), P % V(i))
    
    DO i = 1, N
       P % IN(i) = Temp_P_IN(i)
    END DO
    deallocate(Temp_P_IN)
    
    P_col_ptr => P_col
    P_val_ptr => P_val
    
    do i = 1, re_alloc_times
       do j = 1, N
          P % CN(j+(i-1)*N) = P_col_ptr % intvec(j)
          P % V(j+(i-1)*N) = P_val_ptr % dblvec(j)
       end do
       P_col_ptr => P_col_ptr % next
       P_val_ptr => P_val_ptr % next
    end do
    do i = 1, index_p
       P % CN(re_alloc_times * N + i) = P_col_ptr % intvec(i)
       P % V (re_alloc_times * N + i) = P_val_ptr % dblvec(i)
    end do

    !C deallocate
    call deallocate_dbllist(P_val, re_alloc_times)
    call deallocate_intlist(P_col, re_alloc_times)
    
  END SUBROUTINE jacobi_relaxation

  SUBROUTINE stu_neighbors(PNI, NI, theta, LEVEL_NO)
    USE data_structure_for_AMG
    IMPLICIT NONE
    
    INTEGER(kind=kint ) :: N, NP
    REAL   (kind=kreal), INTENT(in):: theta
    
    INTEGER(kind=kint ), INTENT(out) :: NI  (:)
    INTEGER(kind=kint ), INTENT(out) :: PNI(0:)
    
    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO
    
    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)

    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)

    INTEGER(kind=kint) :: i,j,isU,ieU,isL,ieL,inod,nni,size
    INTEGER(kind=kint) :: NPL, NPU
    REAL   (kind=kreal):: t, max 
    REAL   (kind=kreal):: max_p, epsilon_p=0.2

    NPU = HIERARCHICAL_DATA(LEVEL_NO-1) % NPU
    NPL = HIERARCHICAL_DATA(LEVEL_NO-1) % NPL
    NP = HIERARCHICAL_DATA(LEVEL_NO-1) % NP
    N  = HIERARCHICAL_DATA(LEVEL_NO-1) % N 
    
    D   => HIERARCHICAL_DATA(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA(LEVEL_NO-1) % AU
    
    
    PNI(0) = 0
    nni    = 0

    DO j = 1, N
       isL = INL(j-1)+1
       ieL = INL(j)
       isU = INU(j-1)+1
       ieU = INU(j)

       !C negative-coupling
       max = 0;
       max_p = 0
!C--- memo: Expnd_Posi_IN(0:N), Expnd_Posi_CN(:), Expnd_Posi_VAL(:)
       
       DO i = isL, ieL
          if(max   < -AL(i)) max = -AL(i)
          if(max_p < AL(i)) max_p = AL(i)
       END DO
       DO i = isU, ieU
          if(max   < -AU(i))  max   = -AU(i)
          if(max_p < AU(i)) max_p = AU(i)
       END DO

!!$       if(max_p > epsilon_p*max) then
!!$          !C elimination positive couplings

!!$          DO i= isL, ieL
!!$             k = IAL(i) + ZERO_ORIGIN
!!$             v = AL(i)
!!$             IF(t > epsilon_p*max) THEN
!!$                !C expand! by v, k
       !C call expand_
!!$             ENDIF
!!$          ENDDO
!!$          DO i= isU, ieU
!!$             k = IAU(i) + ZERO_ORIGIN
!!$             v = AU(i)
!!$             IF(t > epsilon_p*max) THEN
!!$                !C expand! by v, k
       !C call expand_
!!$             ENDIF
!!$          ENDDO
!!$
!!$          PNI(2*j-1)=nni
!!$
!!$          DO i = isU, ieU
!!$             inod = IAU(i)+ZERO_ORIGIN
!!$             t = -AU(i)
!!$             
!!$             IF(t > max*theta) THEN
!!$                nni = nni+1
!!$                NI(nni) = inod
!!$             END IF
!!$          ENDDO
!!$          PNI(2*j)=nni
!!$
!!$
!!$
!!$
!!$       
!!$          
!!$       else
          DO i= isL, ieL
             inod = IAL(i) + ZERO_ORIGIN
             t = -AL(i)

             IF(t > max*theta) THEN
                nni = nni+1
                NI(nni) = inod
             ENDIF
          ENDDO

          PNI(2*j-1)=nni

          DO i = isU, ieU
             inod = IAU(i)+ZERO_ORIGIN
             t = -AU(i)
             
             IF(t > max*theta) THEN
                nni = nni+1
                NI(nni) = inod
             END IF
          ENDDO
          PNI(2*j)=nni
!!$       end if
    END DO
    
!!$    DO j = 1, N
!!$       size = PNI(2*j)-PNI(2*j-2)
!!$       IF(size == 0) THEN
!!$          node_index(j) = -1
!!$       END IF
!!$    ENDDO

  END SUBROUTINE stu_neighbors


  SUBROUTINE transpose_ni(N, PNI, NI, PNIT, NIT)
    implicit none
    include 'precision.inc'

    INTEGER(kind=kint), intent(in) :: PNI(0:), NI(:), N
    INTEGER(kind=kint), intent(out) :: PNIT(0:), NIT(:)
    INTEGER(kind=kint) :: i,j,k,col
    
    PNIT(0:2*N) = 0
    !C calculate PNIT
    DO i = 1, N
       !C lower part of NI
       DO j = PNI(2*i-2)+1, PNI(2*i-1)
          col = NI(j)
          if(col < N) then
             PNIT(2*col+1) = PNIT(2*col+1) + 1
          end if
       END DO
       
       !C upper part of NI
       DO j = PNI(2*i-1)+1, PNI(2*i)
          col = NI(j)
          PNIT(2*col) = PNIT(2*col) + 1
       END DO
    END DO
    DO i = 1, 2*N
       PNIT(i) = PNIT(i-1) + PNIT(i)
    END DO
    
    !C determine the column numbers of NIT
    DO i = 1, N
       !C lower part of NI
       DO j = PNI(2*i-2)+1, PNI(2*i-1)
          col = NI(j)
          k = PNIT(2*col)
          k = k + 1
          PNIT(2*col) = k
          NIT(k) = i
       END DO

       !C upper part of NI
       DO j = PNI(2*i-1)+1, PNI(2*i)
          col = NI(j)
          k = PNIT(2*col-1)
          k = k + 1
          PNIT(2*col-1) = k
          NIT(k) = i
       END DO
    END DO
    
  END SUBROUTINE transpose_ni


  subroutine matrix_out(B)
    use data_structure_for_AMG
    use problem_matrix, only: LEVEL_NUM
    IMPLICIT NONE
    REAL(kind=kreal) :: B(:)
    INTEGER(kind=kint) :: i, k, N, NP, NPL, NPU, R_NP, P_NP, R_NPLU, P_NPLU

    
    open(15, file="matrix_structure", status='unknown', form='formatted')
    N  = HIERARCHICAL_DATA(1) % N
    NP = HIERARCHICAL_DATA(1) % NP
    NPL = HIERARCHICAL_DATA(1) % NPL
    NPU = HIERARCHICAL_DATA(1) % NPU
    write (15,'(10i10)') LEVEL_NUM
    write (15,'(10i10)') N, NP, NPL, NPU
    write (15,'(10i10)') (HIERARCHICAL_DATA(1) % INL(k), k= 1, NP)
    write (15,'(10i10)') (HIERARCHICAL_DATA(1) % INU(k), k= 1, NP)
    write (15,'(10i10)') (HIERARCHICAL_DATA(1) % IAL(k), k= 1, NPL)
    write (15,'(10i10)') (HIERARCHICAL_DATA(1) % IAU(k), k= 1, NPU)
    write (15,'(5e27.20)') (HIERARCHICAL_DATA(1) % AL(k), k= 1, NPL)
    write (15,'(5e27.20)') (HIERARCHICAL_DATA(1) % AU(k), k= 1, NPU)
    write (15,'(5e27.20)') (HIERARCHICAL_DATA(1) %  D(k), k= 1, NP)
    write (15,'(5e27.20)') (B(k), k= 1, NP)
    
    DO i=2, LEVEL_NUM
       !C Problem matrix: level i
       N =  HIERARCHICAL_DATA(i) % N
       NP = HIERARCHICAL_DATA(i) % NP
       NPL = HIERARCHICAL_DATA(i) % NPL
       NPU = HIERARCHICAL_DATA(i) % NPU
       write(15, *) "LEVEL", i
       write (15,'(10i10)') N, NP, NPL, NPU
       write (15,'(10i10)') (HIERARCHICAL_DATA(i) % INL(k), k= 1, NP)
       write (15,'(10i10)') (HIERARCHICAL_DATA(i) % INU(k), k= 1, NP)
       write (15,'(10i10)') (HIERARCHICAL_DATA(i) % IAL(k), k= 1, NPL)
       write (15,'(10i10)') (HIERARCHICAL_DATA(i) % IAU(k), k= 1, NPU)
       write (15,'(5e27.20)') (HIERARCHICAL_DATA(i) % AL(k), k= 1, NPL)
       write (15,'(5e27.20)') (HIERARCHICAL_DATA(i) % AU(k), k= 1, NPU)
       write (15,'(5e27.20)') (HIERARCHICAL_DATA(i) %  D(k), k= 1, NP)

       !C Restriction Operator: level i-1 -> level i
       R_NP = HIERARCHICAL_DATA(i) % NP
       R_NPLU = HIERARCHICAL_DATA(i) % R % IN(R_NP)
       write (15,'(10i10)')  R_NP, R_NPLU
       write (15,'(10i10)') (HIERARCHICAL_DATA(i) % R % IN(k), k= 1, R_NP)
       write (15,'(10i10)') (HIERARCHICAL_DATA(i) % R % CN(k), k= 1, R_NPLU)
       write (15,'(5e27.20)') (HIERARCHICAL_DATA(i) % R % V(k), k= 1, R_NPLU)

       !C Prolongation Operator: level i -> level i-1
       P_NP = HIERARCHICAL_DATA(i-1) % NP
       P_NPLU = HIERARCHICAL_DATA(i) % P % IN(P_NP)
       write (15,'(10i10)')  P_NP, P_NPLU
       write (15,'(10i10)') (HIERARCHICAL_DATA(i) % P % IN(k), k= 1, P_NP)
       write (15,'(10i10)') (HIERARCHICAL_DATA(i) % P % CN(k), k= 1, P_NPLU)
       write (15,'(5e27.20)') (HIERARCHICAL_DATA(i) % P % V(k), k= 1, P_NPLU)
    END DO
    close(15)
  end subroutine matrix_out
  
  subroutine clear_multi_lev_matrix(LEVEL_NUM)
    use data_structure_for_AMG

    IMPLICIT NONE
    INTEGER(kind=kint) :: i, LEVEL_NUM

    NULLIFY(       HIERARCHICAL_DATA(1) % INU,      &
            &      HIERARCHICAL_DATA(1) % INL,      &
            &      HIERARCHICAL_DATA(1) % IAU,      &
            &      HIERARCHICAL_DATA(1) % IAL,      &
            &      HIERARCHICAL_DATA(1) % AU,       &
            &      HIERARCHICAL_DATA(1) % AL,       &
            &      HIERARCHICAL_DATA(1) % D,        &
            &      HIERARCHICAL_DATA(1) % X,        &
            &      HIERARCHICAL_DATA(1) % B         )

    DO i=2,LEVEL_NUM
       deallocate(HIERARCHICAL_DATA(i) % INU,      &
            &     HIERARCHICAL_DATA(i) % INL,      &
            &     HIERARCHICAL_DATA(i) % IAU,      &
            &     HIERARCHICAL_DATA(i) % IAL,      &
            &     HIERARCHICAL_DATA(i) % AU,       &
            &     HIERARCHICAL_DATA(i) % AL,       &
            &     HIERARCHICAL_DATA(i) % D,        &
            &     HIERARCHICAL_DATA(i) % X,        &
            &     HIERARCHICAL_DATA(i) % B,        &
            &     HIERARCHICAL_DATA(i) % R % IN,   &
            &     HIERARCHICAL_DATA(i) % R % CN,   &
            &     HIERARCHICAL_DATA(i) % R % V,    &
            &     HIERARCHICAL_DATA(i) % P % IN,   &
            &     HIERARCHICAL_DATA(i) % P % CN,   &
            &     HIERARCHICAL_DATA(i) % P % V    &
            & )
       deallocate(HIERARCHICAL_DATA(i) % P,     &
            &     HIERARCHICAL_DATA(i) % R     &
            & )

    END DO

    if(ILU0_SMOOTH_FL) then
       DO i = 1, LEVEL_NUM
          deallocate( HIERARCHICAL_DATA(i) % ILU0_D,  &
               &      HIERARCHICAL_DATA(i) % ILU0_AL, &
               &      HIERARCHICAL_DATA(i) % ILU0_AU  )
       END DO
    end if
    
    LEVEL_NUM=0

  end subroutine clear_multi_lev_matrix

  subroutine clear_coarse_matrices(LEVEL_NUM)
    use data_structure_for_AMG

    IMPLICIT NONE
    INTEGER(kind=kint) :: i, LEVEL_NUM

    DO i=2,LEVEL_NUM
       deallocate(HIERARCHICAL_DATA(i) % INU,      &
            &     HIERARCHICAL_DATA(i) % INL,      &
            &     HIERARCHICAL_DATA(i) % IAU,      &
            &     HIERARCHICAL_DATA(i) % IAL,      &
            &     HIERARCHICAL_DATA(i) % AU,       &
            &     HIERARCHICAL_DATA(i) % AL,       &
            &     HIERARCHICAL_DATA(i) % D,        &
            &     HIERARCHICAL_DATA(i) % X,        &
            &     HIERARCHICAL_DATA(i) % B,        &
            &     HIERARCHICAL_DATA(i) % R % IN,   &
            &     HIERARCHICAL_DATA(i) % R % CN,   &
            &     HIERARCHICAL_DATA(i) % R % V,    &
            &     HIERARCHICAL_DATA(i) % P % IN,   &
            &     HIERARCHICAL_DATA(i) % P % CN,   &
            &     HIERARCHICAL_DATA(i) % P % V    &
            & )
       deallocate(HIERARCHICAL_DATA(i) % P,     &
            &     HIERARCHICAL_DATA(i) % R     &
            & )

    END DO

  end subroutine clear_coarse_matrices
  
END MODULE  data_creation_stuben
