!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

MODULE data_structure_for_AMG
  IMPLICIT NONE
  INCLUDE 'precision.inc'
  
  INTEGER, parameter :: SWITCH_SIZE    = 1700
  !C creating only for level 2
!C  INTEGER, parameter :: MAX_LEVEL_SIZE = 20
!C  INTEGER, parameter :: MIN_NODE_SIZE  = 90
  INTEGER, parameter :: TIME_KINDS     = 10
  LOGICAL :: ILU0_SMOOTH_FL = .false.
    
  
  TYPE row_node
     INTEGER(kind=kint) :: column
     REAL(kind=kreal)   :: value
  END TYPE row_node

  TYPE intvec_list
     integer(kind=kint) :: size
     integer(kind=kint), pointer :: intvec(:)
     Type(intvec_list), pointer :: next
  END TYPE intvec_list

  TYPE dblvec_list
     integer(kind=kint) :: size
     real(kind=kreal), pointer :: dblvec(:)
     Type(dblvec_list), pointer :: next
  END TYPE dblvec_list
  
  TYPE INTER_LEVEL_OPERATOR
     integer(kind=kint) :: ROW_SIZE
     integer(kind=kint),  pointer :: IN(:)
     integer(kind=kint),  pointer :: CN(:)
     real   (kind=kreal), pointer :: V(:)
  END TYPE INTER_LEVEL_OPERATOR
  
!!$  type COMMUNICATION_TABLE
!!$     integer(kind=kint) :: NEIBPETOT
!!$     integer(kind=kint), pointer :: NEIBPE(:)
!!$     integer(kind=kint), pointer :: STACK_IMPORT(:)
!!$     integer(kind=kint), pointer :: STACK_EXPORT(:)
!!$     integer(kind=kint), pointer :: NOD_IMPORT(:)
!!$     integer(kind=kint), pointer :: NOD_EXPORT(:)
!!$  end type COMMUNICATION_TABLE
  
  TYPE DATA_FOR_EACH_LEV
     integer(kind=kint ) :: N, NP, NPL, NPU
     integer(kind=kint ), pointer :: INU(:), INL(:)
     integer(kind=kint ), pointer :: IAU(:), IAL(:)
#ifdef COMPLEXMODE
     complex(kind=kcomp), pointer :: AU(:), AL(:), D(:), X(:), B(:)
#else
     real   (kind=kreal), pointer :: AU(:), AL(:), D(:), X(:), B(:)
#endif

     real   (kind=kreal), pointer :: ILU0_AU(:), ILU0_AL(:), ILU0_D(:)
     
!!$     type(COMMUNICATION_TABLE) :: COMM_TABLE
!!$     type(COMMUNICATION_TABLE) :: INT_LVL_TABLE
     TYPE(INTER_LEVEL_OPERATOR), pointer :: R, P
  END TYPE DATA_FOR_EACH_LEV
  
  TYPE (DATA_FOR_EACH_LEV) :: HIERARCHICAL_DATA(20)
  
!C Rate_Of_Space * mesh size is the width of the next coarser 
!C level's restriction
!!$  real (kind=kreal) :: RATE_OF_SPACE
!!$  real (kind=kreal),parameter :: EPS = 1.E-13
  real (kind=kreal),  parameter :: EPS = 1.E-10
  real (kind=kreal),  parameter :: EPS_COARSEST = 1.E-5
  INTEGER(kind=kint), parameter :: NODE_RECORD_SIZE=300
  
end module data_structure_for_AMG







