!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

MODULE ilu_relaxation
CONTAINS
  
  SUBROUTINE ilu(N, D, ILU_D, AL, ILU_AL, INL, IAL, AU, ILU_AU, INU, IAU, B, X, TEMP)

    USE data_structure_for_AMG
    IMPLICIT NONE

    INTEGER(kind=kint),INTENT(in )  :: N
    REAL   (kind=kreal), DIMENSION(:)   ::  B
    REAL   (kind=kreal), DIMENSION(:)   ::  X, TEMP
    
    REAL   (kind=kreal), DIMENSION(:)   ::  ILU_D, D
    REAL   (kind=kreal), DIMENSION(:)   ::  ILU_AU, AU
    REAL   (kind=kreal), DIMENSION(:)   ::  ILU_AL, AL
    INTEGER(kind=kint ), DIMENSION(0:)  ::  INU
    INTEGER(kind=kint ), DIMENSION(:)   ::  IAU
    INTEGER(kind=kint ), DIMENSION(0:)  ::  INL
    INTEGER(kind=kint ), DIMENSION(:)   ::  IAL

    REAL   (kind=kreal) :: v
    INTEGER(kind=kint)  :: isL,ieL,isU,ieU,i,j,col

    !C Residual vector is stored in TEMP(:)
    DO i = 1, N
       v = B(i)-D(i)*X(i)
       
       DO j = INL(i-1)+1, INL(i)
          col = IAL(j)
          v = v - AL(j)*X(col)
       END DO

       DO j = INU(i-1)+1, INU(i)
          col = IAU(j)
          v = v - AU(j)*X(col)
       END DO
       TEMP(i) = v
       
    END DO

    DO i = 1, N 
       v = TEMP(i)
       isL = INL(i-1)+1; ieL = INL(i)
       DO j = isL, ieL
          col = IAL(j)
          v = v - X(col) * ILU_AL(j)
       END DO
       
       X(i) = v
    END DO

    DO i = N, 1, -1
       v = X(i)
       isU = INU(i-1)+1; ieU = INU(i)
       DO j = isU, ieU
          col = IAU(j)
          v = v - X(col) * ILU_AU(j)
       END DO
       X(i) = v * ILU_D(i)
    END DO

  END SUBROUTINE ilu

  !C ILU_AL(:), ILU_AU(:), ILU_D(:) has the same date of AL, AU, D
  SUBROUTINE ilu0_setup(N, ILU_D, ILU_AL, INL, IAL, ILU_AU, INU, IAU)
    USE data_structure_for_AMG
    IMPLICIT NONE
    
    INTEGER(kind=kint),INTENT(in )  :: N
    REAL   (kind=kreal), DIMENSION(:)   ::  ILU_D
    REAL   (kind=kreal), DIMENSION(:)   ::  ILU_AU
    REAL   (kind=kreal), DIMENSION(:)   ::  ILU_AL

    INTEGER(kind=kint ), DIMENSION(0:)  ::  INU
    INTEGER(kind=kint ), DIMENSION(:)   ::  IAU
    INTEGER(kind=kint ), DIMENSION(0:)  ::  INL
    INTEGER(kind=kint ), DIMENSION(:)   ::  IAL

    REAL   (kind=kreal) :: t1
    
    INTEGER(kind=kint) :: i,j,k,jj,jrow,jcol,jw,j1,j2
    INTEGER(kind=kint), allocatable :: iw(:)

    REAL   (kind=kreal) :: delta = 1.0, rii
    

    !C iw is temporary vecor 
    allocate(iw(N))

    iw = 0

    !C first row is not changed except diagonal element.
    ILU_D(1) = 1.0 / ILU_D(1)

    do k = 2, N
       !C set k-th row's nonzero pattern in iw(:)
       j1 = INL(k-1)+1; j2 = INL(k)
       do j = j1, j2
          !C lower matrix elements are recorded as minus values
          iw(IAL(j)) = -j
       end do
       j1 = INU(k-1)+1; j2 = INU(k)
       do j = j1, j2
          iw(IAU(j)) = j
       end do

       !C
       rii = 0

       !C gauss elimination on k-th row
       j1 = INL(k-1)+1; j2 = INL(k)
       do j = j1, j2
          !C jrow is the pivot row
          jrow = IAL(j)
          t1 = ILU_AL(j) * ILU_D(jrow)
          ILU_AL(j) = t1

          do jj = INU(jrow-1)+1, INU(jrow)
             jcol = IAU(jj)
             jw = iw(jcol)
             if(jw > 0) then
                ILU_AU(jw) = ILU_AU(jw) - t1 * ILU_AU(jj)
             else if(jw < 0) then
                ILU_AL(-jw) = ILU_AL(-jw) - t1 * ILU_AU(jj)
             else if(jcol == k) then
                ILU_D(k) = ILU_D(k) - t1 * ILU_AU(jj)
             else
                rii = rii + abs(t1*ILU_AU(jj))
             end if
          end do
       end do

       !C modified ILU
       ILU_D(k) = ILU_D(k) + delta * rii

       if(ILU_D(k) == 0.0) stop "Diagonal element comes to 0.0"
       ILU_D(k) = 1.0 / ILU_D(k)
       
       !C reset iw(:)
       j1 = INL(k-1)+1; j2 = INL(k)
       do j = j1, j2
          iw(IAL(j)) = 0
       end do
       j1 = INU(k-1)+1; j2 = INU(k)
       do j = j1, j2
          iw(IAU(j)) = 0
       end do
    end do

    deallocate(iw)
  END SUBROUTINE ilu0_setup


END MODULE  ilu_relaxation



