/*
   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
   3. Neither the name of the project nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "mySparseMatrix.h"

#include "myCUDA.h"


crs::crs()
{
	dsp = NULL;
	col = NULL;
	val = NULL;
}


lu_crs::lu_crs()
{
	inl = NULL;
	inu = NULL;
	ial = NULL;
	iau = NULL;

	al  = NULL;
	au  = NULL;
	d   = NULL;
}


void lu_crs::malloc(int n, int np, int npl, int npu)
{
	this->n   = n;
	this->np  = np;
	this->npl = npl;
	this->npu = npu;

	this->inl = (int *)CUDA_MALLOC_HOST(sizeof(int) * n,   "lu_crs inl");
	this->inu = (int *)CUDA_MALLOC_HOST(sizeof(int) * n,   "lu_crs inu");
	this->ial = (int *)CUDA_MALLOC_HOST(sizeof(int) * npl, "lu_crs ial");
	this->iau = (int *)CUDA_MALLOC_HOST(sizeof(int) * npu, "lu_crs iau");

	this->al = (double *)CUDA_MALLOC_HOST(sizeof(double) * npl,"lu_crs al");
	this->au = (double *)CUDA_MALLOC_HOST(sizeof(double) * npu,"lu_crs au");
	this->d  = (double *)CUDA_MALLOC_HOST(sizeof(double) * n,  "lu_crs d" );
}


void lu_crs::free(void)
{
	CUDA_FREE_HOST(inl);
	CUDA_FREE_HOST(inu);
	CUDA_FREE_HOST(ial);
	CUDA_FREE_HOST(iau);

	CUDA_FREE_HOST(al );
	CUDA_FREE_HOST(au );
	CUDA_FREE_HOST(d  );
}


icrs::icrs()
{
	original = true;

	cnt = NULL;
	col = NULL;
	val = NULL;
}


icrs::icrs(const icrs& other)
{
	original = false;

	n    = other.n;
	np   = other.np;
	nplu = other.nplu;
	row  = other.row;

	cnt = other.cnt;
	col = other.col;
	val = other.val;
}


icrs::~icrs()
{
	if(original){
		if(cnt) CUDA_FREE(cnt);
		if(col) CUDA_FREE(col);
		if(val) CUDA_FREE(val);
	}
}


lu_icrs::lu_icrs()
{
	original = true;

	cnt = NULL;
	col = NULL;
	val = NULL;
	d   = NULL;
}


lu_icrs::lu_icrs(const lu_icrs& other)
{
	original = false;

	n    = other.n;
	np   = other.np;
	nplu = other.nplu;
	row  = other.row;

	cnt = other.cnt;
	col = other.col;
	val = other.val;
	d   = other.d;
}

lu_icrs::~lu_icrs()
{
	if(original){
		if(cnt) CUDA_FREE(cnt);
		if(col) CUDA_FREE(col);
		if(val) CUDA_FREE(val);
		if(d  ) CUDA_FREE(d  );
	}
}
