!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

module parameter_at
  include 'precision.inc'

  !C public: at_parameter_setting(), AMG_solver_include_AT()
  
  real(kind=kreal) :: tv
  
  logical :: at_flag = .false.
  logical :: offline_fl = .false.

  integer(kind=kint) :: expected_call_times = -1

  logical :: ATMODULE_FL = .false.

  
contains
  
  SUBROUTINE set_expected_call_times(count)
    implicit none
    integer(kind=kint) :: count
    
    expected_call_times = count
    
  END SUBROUTINE set_expected_call_times
  
  SUBROUTINE AMG_solver_include_AT(X, B, iter, solver_setup_fl)
    use solver, only: AMG_solver, conv_state
    use problem_matrix, only: LEVEL_NUM, NP
    use at_mod, only:  keisoku_fl,  offline_tuned_solver
    use amg_parameter, only: AT_init
    use terminal_condition, only: BNRM2_set
    use omp_lib
    
    implicit none
    integer(kind=kint), intent(inout) :: iter
    real(kind=kreal), intent(inout)   :: B(:)
    real(kind=kreal), intent(inout)   :: X(:)

    logical, intent(in)               :: solver_setup_fl
    logical :: CONV_FL
    real(kind=kreal) :: setup_time, iter_time
    real(kind=kreal) :: rt1, rt2
    integer :: t1, t2, t_max, t_rate

    call BNRM2_set(B, NP)

    !C when AT_init == 0, at_flag is set.
    !C no AT is used when AT_init is less than 0 at first
    if(AT_init > 0) then
       AT_init = AT_init - 1
    else if(.not.at_flag .and. AT_init == 0) then
       AT_init = AT_init - 1
       at_flag = .true.
    end if
    
    if(at_flag) call at_preprocess(B, X, solver_setup_fl, setup_time)

!!$    call system_clock(t1, t_rate, t_max)
!!$    call cpu_time(rt1)
    rt1 = omp_get_wtime()
    
    if(at_flag .and. keisoku_fl .and. offline_fl) then
       call offline_tuned_solver(X, B, iter, solver_setup_fl)
    else if(at_flag .and. keisoku_fl) then
       call AMG_solver(X, B, iter, .false.)
    else 
       call AMG_solver(X, B, iter, solver_setup_fl)
    end if

!!$    call system_clock(t2, t_rate, t_max)
!!$    call cpu_time(rt2)
    rt2 = omp_get_wtime()
    iter_time = rt2-rt1
    
    !C iter == 0: solver is not called
    if(at_flag .and. iter /= 0) call at_postprocess(iter, X, solver_setup_fl, setup_time, iter_time)

    call conv_state(CONV_FL)
    
  END SUBROUTINE AMG_solver_include_AT

  !C this subroutine is different from  at_preprocess() in amgs_p
  !C especially, tv is determined in different way.
  !C out: tv, X_bak, keisoku_fl, t1
  subroutine at_preprocess(B, X, solver_setup_fl, setup_time)
    use problem_matrix, only: NP, LEVEL_NUM
    use data_creation_stuben, only: setup_phase, clear_multi_lev_matrix
    use terminal_condition, only: terminal_kind, conv_criterion_vec, &
         & conv_criterion_val, terminal_cond_check, BNRM2
    use at_mod, only: keisoku_fl,  setupflag_for_AT, setup_param_changed_fl, &
         & solver_choice, parameter_type, set_parameter, initial_resid
    use residual_calculation, only:cal_res_vec
    use ATmodule, only: ATChooseParam
    use omp_lib

    implicit none
    real(kind=kreal), intent(inout)   :: B(:)
    real(kind=kreal), intent(inout)   :: X(:)
    real(kind=kreal) :: BNRM, RNRM, v, w, setup_time, conv_vec_norm

    real(kind=kreal) :: rt1, rt2
    integer :: t1, t2, t_max, t_rate

    logical :: solver_setup_fl, fl
    integer(kind=kint), save :: X_bak_sz=0
    real(kind=kreal), pointer, save :: X_bak(:)
    
    type(parameter_type) :: param

    integer(kind=kint) :: i

     
    if(keisoku_fl) then
       if(X_bak_sz == 0) then
          X_bak_sz = NP
          allocate(X_bak(NP))
       else if(X_bak_sz < NP) then
          deallocate(X_bak)
          allocate(X_bak(NP))
          X_bak_sz = NP
       end if
       !
       call cal_res_vec(X_bak, X, B)
       !

!!$       BNRM = 0
!!$       do i = 1, NP
!!$          BNRM = BNRM+B(i)**2
!!$       end do
!!$       BNRM = dsqrt(BNRM)
!C       initial_resid = RNRM/BNRM
       
!!$       call BNRM2_set(B, NP)
       
       !
       if(terminal_kind == 1) then
          
          !C conv_vec_norm contains infinity norm of conv_criterion_vec(:)
          conv_vec_norm =  dabs(conv_criterion_vec(1))
          do i = 2, NP
             v = dabs(conv_criterion_vec(i))
             if(conv_vec_norm < v) conv_vec_norm = v
          end do

          call terminal_cond_check(fl, tv, X_bak, NP)

          initial_resid = tv
          
          tv = conv_vec_norm / (tv*BNRM2)
          
       else if(terminal_kind == 0) then

          call terminal_cond_check(fl, tv, X_bak, NP)
          
          initial_resid = tv

          !C conv_criterion_val is assumed to have relative residual norm to rhs vec.
          write(*,*) conv_criterion_val, tv, "uhe-"
          tv = conv_criterion_val / tv
       else
          stop 'unknown terminal_kind'
       end if
       
    end if

#ifdef PRINT_REZ    
    write(*,*) "******", tv, "******"
#endif
    
    if(.not.offline_fl) then
       if(ATMODULE_FL) then
          call ATChooseParam(tv, solver_setup_fl, expected_call_times, param)
       else
          if(tv > 0.9) tv = 0.9
          call solver_choice(tv, setupflag_for_AT, expected_call_times, param)
       end if
       call set_parameter(param)
       
    end if

    if(keisoku_fl .and. .not.offline_fl) then
       !C--check if multi-level data is set appropriately.
       !C       solver_setup_fl       -> create multi-level data everytime in AMG_solver()
       !C       .not.solver_setup_fl  -> call setup() only when theta is changed

       
       if(.not.solver_setup_fl .and. (LEVEL_NUM == 0 .or. setup_param_changed_fl) .or. &
            & solver_setup_fl .and. (LEVEL_NUM == 0)) then
!!$    call system_clock(t1, t_rate, t_max)
!!$    call cpu_time(rt1)
          rt1 = omp_get_wtime()

          call setup_phase()

!!$    call system_clock(t2, t_rate, t_max)
!!$    call cpu_time(rt2)
          rt2 = omp_get_wtime()
          setup_time = rt2-rt1

          !C unset setup_param_changed_fl
          setup_param_changed_fl = .false.
       end if

       !C if keisoku_fl == true, then setup phase is done here. so the following line is commented out.
       !C if(solver_setup_fl .and. LEVEL_NUM > 0) call clear_multi_lev_matrix(LEVEL_NUM)

       !C unset setup_param_changed_fl
       if(solver_setup_fl) setup_param_changed_fl = .false.
    end if

  end subroutine at_preprocess

  subroutine reset_B_X(B, X, X_bak)
    USE problem_matrix
    real(kind=kreal)   :: B(:), X(:), w, X_bak(:)
    integer(kind=kint) :: c,i,j,js, je
!
    do i = 1, NP
       w = B(i)
       
       js = INL(i-1)+1
       je = INL(i)
       do j = js, je
          c = IAL(j)
          w = w-AL(j)*X(c)
       end do
       
       w = w-D(i)*X(i)
       
       js = INU(i-1)+1
       je = INU(i)
       do j = js, je
          c = IAU(j)
          w = w-AU(j)*X(c)
       end do
       B(i) = w
    end do
!
    do i = 1, NP
       X_bak(i) = X(i)
       X(i) = 0
    end do

  end subroutine reset_B_X
  
  !C out: t2, X, keisoku_fl, iter
  subroutine at_postprocess(iter, X, solver_setup_fl, setup_time, iter_time) 
    use solver, only: conv_state, resid
    use amg_parameter, only: conv_check_fl
    use problem_matrix, only: NP, LEVEL_NUM
    use at_mod, only: keisoku_fl, update_time, check, initial_resid
    use data_creation_stuben, only: clear_multi_lev_matrix
    use ATmodule, only: ATReportCost
    use terminal_condition, only: BNRM2, terminal_kind
    use omp_lib
    implicit none
    real(kind=kreal), intent(inout)   :: X(:)
    logical  :: reach_conv, solver_setup_fl
    integer(kind=kint) :: iter, i
    real(kind=kreal)   :: rt, setup_time, iter_time 

    !C this sub should be included in amgs_solver()
    call conv_state(reach_conv)
    
    !C --
    if(keisoku_fl .and. .not.offline_fl) then
       !C  In the case of divergence.
       !C     when problem matrix is changing(solver_setup_fl is true), then it return to level 0.
       !C     Otherwise, it return to level 1.
       
!!$       write(*,*) "debug: updating time:", (t2-t1)/dble(t_rate), reach_conv
!!$       call update_time((t2-t1)/dble(t_rate), tv, reach_conv, solver_setup_fl)
       
       !C -- correct tv is
       !C tv = resid / initial_resid
!!$       if(conv_check_fl) then
!!$          !C -- correct tv is calculated..
!!$          if(terminal_kind==0) then
!!$             !C BNRM2 contains sum of the squares of rhs vector elements
!!$             tv = resid * dsqrt(BNRM2) / initial_resid
!!$          else if(terminal_kind==1) then
!!$             !C BNRM2 contains infinity norm of rhs vector
!!$             tv = resid * BNRM2 / initial_resid
!!$          end if
!!$       end if
       
       write(*,*) "updating time:", iter, tv, setup_time, iter_time, reach_conv
       if(ATMODULE_FL) then
          call ATReportCost(setup_time, iter_time, iter, tv, reach_conv, solver_setup_fl)
       else
          call update_time(setup_time, iter_time, iter, tv, reach_conv, solver_setup_fl)
       end if
       
    else
       !C if iteration number increases by some fixed value or reach_conv == .false.,  then
       !C keisoku_fl is set
       if(.not. offline_fl) then
          call check(iter, reach_conv, keisoku_fl, offline_fl)
       end if
       
    end if

    expected_call_times = expected_call_times-1
    
    !C-- after this sub, assigning X(:) to PP(:).
    if(solver_setup_fl .and. keisoku_fl) call clear_multi_lev_matrix(LEVEL_NUM)
    
  end subroutine at_postprocess
  
  subroutine at_parameter_finish()
    at_flag = .false.
  end subroutine at_parameter_finish
  
  subroutine amg_parameter_set2(                    &
       & cycle_shape, sm_type, iter_down, iter_up,  &
       & iter_bottom, F_cycle_iter_down,            &
       & strong_connect_thresld, dump_jacobi_coeff, &
       & dump_gauss_seidel_coeff, solver_str,       &
       & num_of_vecs, conv_chk )
    use amg_parameter
    !C vector problem is not considered here.
    integer(kind=kint) :: cycle_shape, sm_type, iter_down, iter_up, iter_bottom
    integer(kind=kint) :: F_cycle_iter_down, num_of_vecs, conv_chk
    real(kind=kreal) :: strong_connect_thresld, dump_jacobi_coeff, &
         & dump_gauss_seidel_coeff
    character(*) :: solver_str

    cycle_kind = cycle_shape
    gs_kind = sm_type
    iteration_number_doen = iter_down
    iteration_number_up = iter_up
    iteration_number_botton = iter_bottom
    f_first_iteration_number = F_cycle_iter_down
    theta = strong_connect_thresld
    dump_jacobi_weight = dump_jacobi_coeff
    dump_gauss_seidel_weight = dump_gauss_seidel_coeff
!C    mstring = "SA"
    solver_string = solver_str
    number_of_vectors = num_of_vecs

    conv_check_fl = (conv_chk == 1)
    AT_init = -1
    

  end subroutine amg_parameter_set2


  subroutine amg_parameter_set(Nunit)
    use amg_parameter
    use at_mod, only: at_parameter_setting, at_parameter_setting_from_file
    implicit none
    integer :: Nunit , stat
    character(len=20) :: mstring="SA"
    open(Nunit,FILE='AMG.dat',FORM='FORMATTED',STATUS='old', iostat=stat)
    if(stat /= 0) then
       write(*,*) "No AMG parameter file: AMG.dat. Default setting is used."        
    end if


    read(Nunit, *, iostat=stat) cycle_kind 
    if(stat /= 0) goto 120
    read(Nunit, *, iostat=stat) gs_kind
    if(stat /= 0) goto 120
    read(Nunit, *, iostat=stat) iteration_number_down
    if(stat /= 0) goto 120
    read(Nunit, *, iostat=stat) iteration_number_up
    if(stat /= 0) goto 120
    read(Nunit, *, iostat=stat) iteration_number_bottom
    if(stat /= 0) goto 120
    read(Nunit, *, iostat=stat) f_first_iteration_number
    if(stat /= 0) goto 120
    read(Nunit, *, iostat=stat) theta
    if(stat /= 0) goto 120
    read(Nunit, *, iostat=stat) dump_jacobi_weight
    if(stat /= 0) goto 120
    read(Nunit, *, iostat=stat) dump_gauss_seidel_weight
    if(stat /= 0) goto 120
    read(Nunit, *, iostat=stat) mstring
    if(stat /= 0) goto 120
    read(Nunit, *, iostat=stat) solver_string 
    if(stat /= 0) goto 120
!!$    if(solver_string(1:1)=='G' .and. solver_string(2:2)=='M' .and. &
!!$         &  solver_string(3:3)=='R' .and. solver_string(4:4)=='E' .or. &
!!$         &  solver_string(1:1)=='I' .and. solver_string(2:2)=='D' .and. &
!!$         &  solver_string(3:3)=='R'.or. solver_string(1:1)=='R' .and.&
!!$         &  solver_string(2:2)=='E' .and. solver_string(3:3)=='S' .and. &
!!$         &  solver_string(4:4)=='C') then
    read(Nunit, *, iostat=stat) number_of_vectors
    if(stat /= 0) goto 120
!!$    end if
    read(Nunit, *, iostat=stat) dummy
    if(stat /= 0) goto 120
    read(Nunit, *, iostat=stat) conv_check_str
    conv_check_fl = conv_check_str(1:1)=='C' .and. conv_check_str(2:2)=='H' .and. &
         &  conv_check_str(3:3)=='E' .and. conv_check_str(4:4)=='C' 

    read(Nunit, *, iostat=stat) AT_string, AT_interval, AT_diff, AT_init, AT_setup
    if(stat /= 0) goto 120

    close(Nunit)


    if(cycle_kind==0) then 
       write(*,*) "V-cycle"
    else if(cycle_kind==1) then
       write(*,*) "F-cycle"
    else if(cycle_kind==2) then
       write(*,*) "W-cycle"
    else 
       write(*,*) "Error in Line 1 of AMG.dat"
       stop
    end if

    if(mstring(1:1)=='S' .and. mstring(2:2)=='A') then
       write(*,*) "SA"
       method=22
    else if(mstring(1:1)=='S' .and. mstring(2:2)=='T'.and. mstring(3:3)=='1') then
       write(*,*) "Standard method of Stuben solver"
       method=0
    else if(mstring(1:1)=='S' .and. mstring(2:2)=='T'.and. mstring(3:3)=='2') then
       write(*,*) "aggressive coarsening version of Stuben solver"
       method=9
    else if(mstring(1:1)=='S' .and. mstring(2:2)=='T'.and. mstring(3:3)=='3') then
       write(*,*) "Standard AMG method"
       method=-1
    else if(mstring(1:1)=='A' .and. mstring(2:2)=='D') then
       write(*,*) "Adaptive AMG method"
       method=-2
    else
       write(*,*) "Error in Line 10 of AMG.dat"
       stop
    end if
    
    if(AT_string(1:1)=='A' .and. AT_string(2:2)=='T') then
       if(AT_init <= 0) then
          at_flag = .true.
          AT_init = -1
       end if
       
       !C inter_val < 0 means offline tuning.
       if(AT_interval <= 0) then
          offline_fl =.true.
       else
          offline_fl =.false.
       end if

       if(AT_setup == 0)then
!!$          call at_parameter_setting(AT_interval, AT_diff, AT_init, .false.) 
          call at_parameter_setting_from_file(AT_interval, AT_diff, 11, .false.) 
       else if(AT_setup == 1) then
!!$          call at_parameter_setting(AT_interval, AT_diff, AT_init, .true.) 
          call at_parameter_setting_from_file(AT_interval, AT_diff, 11, .true.) 
       end if
    else
       !C if AT_string is not equal to AT, then AT_init is reset to negative value.
       AT_init = -1
    end if
    
    return

120 write(*,*) "Error in AMG parameter file: AMG.dat"
    close(Nunit)
    stop

  end subroutine amg_parameter_set


  
  subroutine parameter_string_input(x)
    implicit none
    character(*) :: x
    character(50) :: key, value
    integer :: h,t,e 

    h = index(x,"-")
    do 
       t = h+index(x(h+1: ), "-") - 1
       if(t<h) then
          e = h+index(x(h+1:), "=")
       else
          e = h+index(x(h+1:t), "=")
       end if

       key = trim(adjustl(x(h+1:e-1)))

       if(t<h) then
          value= trim(adjustl(x(e+1:)))
       else
          value= trim(adjustl(x(e+1:t)))
       end if

       call parameter_value_set(key, value)

       if(t<h) exit
       h=t+1
    end do

  end subroutine parameter_string_input


  subroutine amg_parameter_option_file(Nunit, filename)
    implicit none

    character(*) :: filename
    character(len=1024) :: x,xbuf
    character(len=512) :: linebuf
    integer :: Nunit 
    logical :: NO_FILE_FL

    NO_FILE_FL=.false.
    x = " "
    open(Nunit, FILE=filename, status='old', &
         & FORM='FORMATTED', err=122)
    do 
       read(Nunit, '(a)', end=998) linebuf
       xbuf = trim(x)//" "//trim(linebuf)
       x    = xbuf
    end do
    
    !C file open error
    if(.false.) then
122    NO_FILE_FL=.true.
       write(*,*) "No option file: ", filename
    end if

    IF(.not. NO_FILE_FL) then

       !C input optsion string 
998    write(*,*) "input option:", trim(x)
       
       call parameter_string_input(x)

    end IF
    
  end subroutine amg_parameter_option_file

  subroutine print_amg_option_range(arg)
    implicit none

    integer :: arg
    
    if(arg == 0) then
       !C parameter range is output
       write(*,*) " "
       write(*,*) "----- parameter option list ------"       
       write(*,*) "# means a number, the inital value at each line is default setting."
       write(*,*) "-cycle:                0,1,2 :: V,F,W"
       write(*,*) "-smoother:             sgs,gs,mcsgs,mcgs"
       write(*,*) "-iter_down:            #: 1"
       write(*,*) "-iter_up:              #: 1"
       write(*,*) "-iter_bottom:          #: 30"
       write(*,*) "-f_first_iter_down:    #"
       write(*,*) "-strong_con_threshold: #: 0.1: 0.0-0.15"
       write(*,*) "-threshold_reduction_rate: #: 0.8: 0.5-1.0"
       write(*,*) "-strong_con_threshold_?: #: it can set direct value for the level ?"
       write(*,*) "-dump_jacobi_coef:     #: 0.66: 0.5-0.8"
       write(*,*) "-dump_jacobi_coef_?:     #: it can set direct value for the level ?"
       write(*,*) "-smoother_accel_coef:  #: 0.8: 0.2-1.4"
       write(*,*) "-krylov_solver:        BICGSTAB, NO, CG, RESCUT, GPBICG"
       write(*,*) "-restart:              #: 2-10"
       write(*,*) "-convergence_check:    #: 1 or 0"
       write(*,*) "-parameter_at:         NOAT, AT"
       write(*,*) "-at_init:              #"
       write(*,*) "-at_diff:              #"
       write(*,*) "-at_interval:          #"
       write(*,*) "-at_setup:             #: 1 or 0"
       write(*,*) "-min_node_size:        #: 100:100-700"
       write(*,*) "-max_level_size:       #: 10:2-10"
       write(*,*) " "
    else if(arg == 1) then
       !C current parameter setting
       write(*,*) " "
       write(*,*) " --- current parameter setting ---"
       call print_current_parameter_set()
       write(*,*) " "
    end if

  end subroutine print_amg_option_range




  subroutine parameter_value_set(key, value)
    use amg_parameter,  only: cycle_kind, gs_kind, iteration_number_down, iteration_number_up, &
         &                    iteration_number_bottom, f_first_iteration_number, theta,        &
         &                    theta_reduction_rate, dump_jacobi_weight,                        &
         &                    dump_gauss_seidel_weight, solver_string,                         &
         &                    number_of_vectors, conv_check_str, conv_check_fl,                &
         &                    AT_string, AT_init, AT_interval, AT_diff, AT_setup,              &
         &                    max_level_size, min_node_size, method
    use at_mod, only: at_parameter_setting, at_parameter_setting_from_file
    implicit none

    character(50) :: key, value

    integer :: i,l
    logical :: flag

    l=len_trim(key)

    if(l==5) then  
       if(key(1:l).eq."cycle") then
!C          write(*,*)"cycle : ", trim(value)
          read(value,*) cycle_kind
       else
          write(*,*) "### unknown option!:", key
       end if

    else if(l==7) then
       if(key(1:l).eq."iter_up") then
!C          write(*,*)"iter_up : ", value
          read(value,*) iteration_number_up

       else if(key(1:l).eq."at_diff") then
!C          write(*,*)"at_diff : ", value
          read(value,*) AT_diff

       else if(key(1:l).eq."at_init") then
!C          write(*,*)"at_init : ", value
          read(value,*) AT_init
          
       else if(key(1:l).eq."restart") then
!C          write(*,*)"restart : ", value
          read(value,*) number_of_vectors
       else
          write(*,*) "### unknown option!:", key
       end if

    else if(l==8)then
       if(key(1:l).eq."smoother") then
!C          write(*,*)"smoother : ", trim(value)
          if(value(1:1)=='g' .and. value(2:2)=='s') then
             gs_kind=1
          else if(value(1:1)=='s' .and. &
               &  value(2:2)=='g' .and. &
               &  value(3:3)=='s') then
             gs_kind=0
          end if
             
       else if(key(1:l).eq."at_setup") then
!C          write(*,*)"at_setup : ", value
          read(value,*) AT_setup
       else if(key(1:l).eq."amg_type") then
          if(value(1:2).eq."SA") then
             write(*,*) "SA solver is set"
             method=22
          else if(value(1:3).eq."ST1") then
             write(*,*) "Standard method of Stuben solver"
             method=0
          else if(value(1:3).eq."ST2") then
             write(*,*) "aggressive coarsening version of Stuben solver"
             method=9
          else if(value(1:3).eq."ST3") then
             write(*,*) "Standard AMG method"
             method=-1

          else if(value(1:3).eq."AD") then
             write(*,*) "Adaptive AMG method"
             method=-2
          else 
             write(*,*) "unknown amg_type in AMGS_OPTION file!!!"
          end if
       else 
          write(*,*) "### unknown option!:", key
       end if


    else if(l==9) then
       if(key(1:l).eq."iter_down") then
!C          write(*,*)"iter_down : ", value
          read(value, *) iteration_number_down
       else 
          write(*,*) "### unknown option!:", key
       end if

    else if(l==11) then
       if(key(1:l).eq."iter_bottom") then
!C          write(*,*)"iter_bottom : ", value
          read(value,*) iteration_number_bottom
       else if(key(1:l).eq."at_interval") then
!C          write(*,*)"at_interval : ", value
          read(value,*) AT_interval
       else 
          write(*,*) "### unknown option!:", key
       end if
       
    else if(l==12) then
       if(key(1:l) .eq. "parameter_at") then
!C          write(*,*)"parameter_at : ", value
          read(value,*) i
          if(i==1) then
             AT_string="AT"
          else
             AT_string="NOAT"
          end if
       else 
          write(*,*) "### unknown option!:", key
       end if

    else if(l==13) then
       if(key(1:l).eq."krylov_solver") then
!C          write(*,*)"krylov_solver : ", value
          read(value,*) solver_string
       else if(key(1:l).eq."min_node_size") then
          read(value,*) min_node_size
       else
          write(*,*) "### unknown option!:", key
       end if

    else if(l==14) then
       if(key(1:l).eq."max_level_size") then
          read(value,*) max_level_size
       else 
          write(*,*) "### unknown option!:", key
       end if

    else if(l==16) then
       if(key(1:l).eq."dump_jacobi_coef") then
!C          write(*,*)"dump_jacobi_coef : ", value
          read(value, *) dump_jacobi_weight
       else
          write(*,*) "### unknown option!:", key
       end if


    else if(l==17) then
       if(key(1:l).eq."convergence_check") then
!C          write(*,*)"convergence_check : ", value
          read(value,*) i
          if(i /= 0) conv_check_fl=.true.
       else if(key(1:l).eq."f_first_iter_down") then
!C          write(*,*)"f_first_iter_down : ", value
          read(value,*) f_first_iteration_number
       else
          write(*,*) "### unknown option!:", key
       end if


    else if(l==19) then
       if(key(1:l).eq."smoother_accel_coef") then
!C          write(*,*)"smoother_accel_coef : ", value
          read(value,*) dump_gauss_seidel_weight
       else
          write(*,*) "### unknown option!:", key
       end if

    else if(l==20) then
       if(key(1:l).eq."strong_con_threshold") then
!C          write(*,*)"strong_con_threshold : ", value
          read(value,*) theta
       else
          write(*,*) "### unknown option!:", key
       end if
    else if(l==24) then
       if(key(1:l).eq."threshold_reduction_rate") then
          read(value,*) theta_reduction_rate
       else
          write(*,*) "### unknown option!:", key
       end if
       
    else
       write(*,*) "### unknown option!:", key
    end if
    

  end subroutine parameter_value_set


  subroutine print_current_parameter_set()
    use amg_parameter,  only: cycle_kind, gs_kind, iteration_number_down, iteration_number_up, &
         &                    iteration_number_bottom, f_first_iteration_number, theta,        &
         &                    theta_reduction_rate, dump_jacobi_weight,                        &
         &                    dump_gauss_seidel_weight, solver_string,                         &
         &                    number_of_vectors, conv_check_str, conv_check_fl,                &
         &                    AT_string, AT_init, AT_interval, AT_diff, AT_setup,              &
         &                    method,          &
         &                    max_level_size, min_node_size
    use at_mod, only: at_parameter_setting, at_parameter_setting_from_file
    implicit none
    
    integer, parameter :: string_sz=1000
    character(string_sz) :: string_buf
    character(200) :: sbuf
    
    write(sbuf,*) cycle_kind
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -cycle="//sbuf
    string_buf=sbuf
    string_buf=trim(string_buf)

    if(gs_kind==1) then
       sbuf=" -smoother="//"gs"
    else if(gs_kind==0) then
       sbuf=" -smoother="//"sgs"       
    else if(gs_kind==2) then
       sbuf=" -smoother="//"mcsgs"       
    else if(gs_kind==3) then
       sbuf=" -smoother="//"mcgs"       
    else
       sbuf=" -smoother= unknown!"
    end if
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,*) iteration_number_down
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -iter_down="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,*) iteration_number_up
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -iter_up="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,*) iteration_number_bottom
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -iter_bottom="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,*) f_first_iteration_number
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -f_first_iter_down="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,'(f6.3)') theta
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -strong_connection_threshold="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,'(f6.3)') theta_reduction_rate
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -threshold_reduction_rate="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,'(f6.3)') dump_jacobi_weight
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -dump_jacobi_coef="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,'(f6.3)') dump_gauss_seidel_weight
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -smoother_accel_coef="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    sbuf=adjustl(solver_string); sbuf=trim(sbuf)
    sbuf=" -krylov_solver="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf, *) number_of_vectors
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -restart="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    ! sbuf=adjustl(dst_string); sbuf=trim(sbuf)
    ! sbuf=" -dist_method="//sbuf
    ! string_buf = string_buf(1:len_trim(string_buf))//sbuf

    if(conv_check_fl) then
       write(sbuf, *) 1
    else
       write(sbuf, *) 0
    end if
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -convergence_check="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    if(AT_string(1:1)=='A' .and. AT_string(2:2)=='T') then
       write(sbuf, *) 1
    else 
       write(sbuf, *) 0
    end if
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -parameter_at="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf
    if(AT_string(1:1)=='A' .and. AT_string(2:2)=='T') then
       
       write(sbuf, *) AT_init
       sbuf=adjustl(sbuf); sbuf=trim(sbuf)
       sbuf=" -at_init="//sbuf
       string_buf = string_buf(1:len_trim(string_buf))//sbuf

       write(sbuf, *) AT_diff
       sbuf=adjustl(sbuf); sbuf=trim(sbuf)
       sbuf=" -at_diff="//sbuf
       string_buf = string_buf(1:len_trim(string_buf))//sbuf

       write(sbuf, *) AT_interval
       sbuf=adjustl(sbuf); sbuf=trim(sbuf)
       sbuf=" -at_interval="//sbuf
       string_buf = string_buf(1:len_trim(string_buf))//sbuf

       write(sbuf, *) AT_setup
       sbuf=adjustl(sbuf); sbuf=trim(sbuf)
       sbuf=" -at_setup="//sbuf
       string_buf = string_buf(1:len_trim(string_buf))//sbuf
    end if
    
    ! write(sbuf, *) NB
    ! sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    ! sbuf=" -vector_problem="//sbuf
    ! string_buf = string_buf(1:len_trim(string_buf))//sbuf
    
    ! if(aggregation_strategy==1) then
    !    sbuf = "coupled_w_lu"
    ! else if(aggregation_strategy==2) then
    !    sbuf = "coupled_w_o_lu"
    ! else if(aggregation_strategy==3) then
    !    sbuf = "independent_w_lu"
    ! else if(aggregation_strategy==4) then
    !    sbuf = "independent_w_o_lu"
    ! else if(aggregation_strategy==5) then
    !    sbuf = "old_independent_w_lu"
    ! else if(aggregation_strategy==6) then
    !    sbuf = "old_independent_w_o_lu"
    ! end if
    ! sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    ! sbuf=" -aggregation="//sbuf
    ! string_buf = string_buf(1:len_trim(string_buf))//sbuf

    ! write(sbuf, *) process_aggre_threshold
    ! sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    ! sbuf=" -process_aggre_threshold="//sbuf
    ! string_buf = string_buf(1:len_trim(string_buf))//sbuf

    if(method == 22) then
       sbuf="SA"
    else if(method==0) then
       sbuf="ST1"
    else if(method==9) then
       sbuf="ST2"
    else if(method==-1) then
       sbuf="ST3"
    else if(method==-2) then
       sbuf="AD"
    end if
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -amg_type="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf



    write(sbuf, *) min_node_size
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -min_node_size="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf, *) max_level_size
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -max_level_size="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    ! write(sbuf, *) comm_opt_for_CGA
    ! sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    ! sbuf=" -comm_opt_for_CGA="//sbuf
    ! string_buf = string_buf(1:len_trim(string_buf))//sbuf

    if(string_sz < len_trim(string_buf)) then
       write(*,*) "Error!: string buffer is not sufficient in print_current_parameter_set()"
    end if
    
    write(*,*) string_buf(1:len_trim(string_buf))
    
  end subroutine print_current_parameter_set





  
end module parameter_at
