!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

MODULE problem_matrix
  IMPLICIT NONE
  INCLUDE 'precision.inc'
  
  INTEGER(kind=kint) :: NP, NPL, NPU, LEVEL_NUM = 0
  INTEGER(kind=kint), pointer :: INL(:), INU(:), IAL(:), IAU(:)
  REAL(kind=kreal), pointer   :: AL(:), AU(:), D(:)
  
  LOGICAL :: NULL_SPACE_CONSTANT = .false.

contains

  SUBROUTINE set_null_space_constant
    implicit none
    
    NULL_SPACE_CONSTANT = .true.
  end SUBROUTINE set_null_space_constant
  

  subroutine cal_residual_norm(X,B,r)
    implicit none
    integer(kind=kint) :: i,j
    real(kind=kreal) :: r, v, X(:), B(:)
    
    v=0
    do i = 1, NP
       v = B(i)
       do j = INL(i-1)+1, INL(i)
          v = v-AL(j)*X(IAL(j))
       end do
       do j = INU(i-1)+1, INU(i)
          v = v-AU(j)*X(IAU(j))
       end do
       v = v- D(i)*X(i)
       
       r = r+v*v
    end do
    r = dsqrt(r)
    
  end subroutine cal_residual_norm
  
  subroutine store_problem_matrix(crs_index, crs_column, crs_val, rowsize)
    REAL(kind=kreal) :: crs_val(:)
    INTEGER(kind=kint) :: crs_column(:), crs_index(0:), rowsize
    
    REAL(kind=kreal) :: v
    INTEGER(kind=kint) :: i,j,il,iu,row,col


    NP = rowsize; NPL=0; NPU=0
    do row=1,NP
       do j=crs_index(row-1)+1,crs_index(row)
          col = crs_column(j)
          if(col .ne. row) then
             if(col < row) then
                NPL=NPL+1
             else
                NPU=NPU+1
             end if
          end if
       end do
    end do

    allocate(INL(0:NP),INU(0:NP),IAU(NPU),IAL(NPL),AU(NPU),AL(NPL),&
         & D(NP))

    il=0; iu=0
    INL(0)=0; INU(0)=0
    do row=1,NP
       !
       do j = crs_index(row-1)+1,crs_index(row)
          col = crs_column(j)
          v = crs_val(j)
          if( col == row ) then
             D(row) = v
          else if(col < row) then
             il=il+1; IAL(il)=col; AL(il) = v            
          else 
             iu=iu+1; IAU(iu)=col; AU(iu) = v
          end if
       end do
       INL(row) = il; INU(row) = iu
    end do
  end subroutine store_problem_matrix
  
end module problem_matrix
