!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


module residual_calculation
contains
!C    subroutine cal_relative_res_norm(res_norm, X, B, WS, WR, WSIZE)
  subroutine cal_relative_res_norm(res_norm, R, X, B)
!C    use solver_SR2      
    use problem_matrix
    use terminal_condition, only: BNRM2
!C    !$ USE omp_lib
    implicit none
!C    include 'mpif.h'


    REAL(kind=kreal),intent(out) :: res_norm, R(:)
!C    REAL(kind=kreal) :: WVAL, res_norm0, X(:), B(:), WS(:), WR(:)
    REAL(kind=kreal) :: WVAL, res_norm0, X(:), B(:)
    INTEGER(kind=kint) :: isU,ieU,isL,ieL, i,j,inod, ierr,WSIZE


!C    call SOLVER_SEND_RECV2                                       &
!C         &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT,&
!C         &     NOD_IMPORT, STACK_EXPORT, NOD_EXPORT,             &
!C         &     WS(1:WSIZE), WR(1:WSIZE), X(1:NP) , SOLVER_COMM,  &
!C         &     my_rank)


    res_norm0 = 0.0

!C    !$omp parallel do private(WVAL, isU, ieU, isL, ieL, i, inod) reduction(+:res_norm0)
!C    do j= 1, N
    do j= 1, NP
       WVAL= B(j) - D(j) * X(j)
       isU= INU(j-1) + 1
       ieU= INU(j  ) 
       do i= isU, ieU
          inod= IAU(i)
          WVAL= WVAL - AU(i) * X(inod)
       enddo

       isL= INL(j-1) + 1
       ieL= INL(j  ) 
       do i= isL, ieL
          inod= IAL(i)
          WVAL= WVAL - AL(i) * X(inod)
       enddo
       !C res(1:N)
       res_norm0 = res_norm0+WVAL*WVAL
       R(j) = WVAL
    enddo
!C    !$omp end parallel do

!C    CALL MPI_ALLREDUCE(res_norm0, res_norm, 1, MPI_DOUBLE_PRECISION, &
!C         &             MPI_SUM, SOLVER_COMM, ierr)

!C    res_norm = dsqrt(res_norm)/BNRM2
    res_norm = dsqrt(res_norm0/BNRM2)

  end subroutine cal_relative_res_norm

  subroutine cal_res_vec(R, X, B)
!C    use solver_SR2      
    use problem_matrix
!C    !$ USE omp_lib
    implicit none
!C    include 'mpif.h'


    REAL(kind=kreal),intent(out) :: R(:)
!C    REAL(kind=kreal) :: WVAL, res_norm0, X(:), B(:), WS(:), WR(:)
    REAL(kind=kreal) :: WVAL, X(:), B(:)
    INTEGER(kind=kint) :: isU,ieU,isL,ieL, i, j, inod, ierr, WSIZE


!C    call SOLVER_SEND_RECV2                                       &
!C         &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT,&
!C         &     NOD_IMPORT, STACK_EXPORT, NOD_EXPORT,             &
!C         &     WS(1:WSIZE), WR(1:WSIZE), X(1:NP) , SOLVER_COMM,  &
!C         &     my_rank)



!C    !$omp parallel do private(WVAL, isU, ieU, isL, ieL, i, inod) reduction(+:res_norm0)
!C    do j= 1, N
    do j= 1, NP
       WVAL= B(j) - D(j) * X(j)
       
       isU= INU(j-1) + 1
       ieU= INU(j  ) 
       do i= isU, ieU
          inod= IAU(i)
          WVAL= WVAL - AU(i) * X(inod)
       enddo

       isL= INL(j-1) + 1
       ieL= INL(j  ) 
       do i= isL, ieL
          inod= IAL(i)
          WVAL= WVAL - AL(i) * X(inod)
       enddo
       !C res(1:N)
       R(j) = WVAL
    enddo
!C    !$omp end parallel do

  end subroutine cal_res_vec


end module residual_calculation
