!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif

MODULE sgs_relaxation
CONTAINS

  !C modified..  
  SUBROUTINE sgs(N, NP, D, AL, INL, IAL, AU, INU, IAU, B, X, count, LEVEL_NO, kind)

    USE data_structure_for_AMG
    USE amg_parameter, only: dump_gauss_seidel_weight
    IMPLICIT NONE

    !C kind == 0,1,2 :: sgs, gs, gs in reverse order.
    INTEGER(kind=kint),INTENT(in )  :: N,NP,count, kind
    REAL   (kind=kreal), DIMENSION(:)   ::  D
    REAL   (kind=kreal), DIMENSION(:)   ::  B
    REAL   (kind=kreal), DIMENSION(:)   ::  X

    REAL   (kind=kreal), DIMENSION(:)   ::  AU
    REAL   (kind=kreal), DIMENSION(:)   ::  AL
    REAL   (kind=kreal) :: R_val,R_norm,GR_norm,v

    INTEGER(kind=kint ), DIMENSION(0:)  ::  INU
    INTEGER(kind=kint ), DIMENSION(:)   ::  IAU
    INTEGER(kind=kint ), DIMENSION(0:)  ::  INL
    INTEGER(kind=kint ), DIMENSION(:)   ::  IAL

    INTEGER(kind=kint), intent(in) :: LEVEL_NO

    INTEGER(kind=kint)  :: cnt,isL,ieL,isU,ieU,i,j,inod
    REAL(kind=kreal) :: omega
    omega = dump_gauss_seidel_weight

    DO cnt=1,count
       if(kind == 1 .or. kind==0) then
          DO i=1,N
             v = B(i)

             isL=INL(i-1)+1
             ieL=INL(i)
             DO j=isL,ieL
                inod=IAL(j)+ZERO_ORIGIN
                v = v - AL(j)*X(inod)
             END DO

             isU=INU(i-1)+1
             ieU=INU(i)
             DO j=isU,ieU
                inod=IAU(j)+ZERO_ORIGIN
                v = v - AU(j)*X(inod)
             END DO

             X(i) = omega * v / D(i) + (1-omega)*X(i)
          END DO
       end if

       if(kind==2 .or. kind==0) then
          DO i=N,1,-1
             v = B(i)

             isU=INU(i-1)+1
             ieU=INU(i)
             DO j=isU,ieU
                inod=IAU(j)+ZERO_ORIGIN
                v = v - AU(j)*X(inod)
             END DO

             isL=INL(i-1)+1
             ieL=INL(i)
             DO j=isL,ieL
                inod=IAL(j)+ZERO_ORIGIN
                v = v - AL(j)*X(inod)
             END DO
             X(i) = omega * v / D(i) + (1-omega)*X(i)
          END DO
       end if

    END DO
  END SUBROUTINE sgs

  SUBROUTINE jacobi(N, NP, D, AL, INL, IAL, AU, INU, IAU, B, X, count, LEVEL_NO)

    USE data_structure_for_AMG
    USE amg_parameter, only: dump_gauss_seidel_weight
    IMPLICIT NONE

    INTEGER(kind=kint),INTENT(in )  :: N,NP,count
    REAL   (kind=kreal), DIMENSION(:)   ::  D
    REAL   (kind=kreal), DIMENSION(:)   ::  B
    REAL   (kind=kreal), DIMENSION(:)   ::  X

    REAL   (kind=kreal), DIMENSION(:)   ::  AU
    REAL   (kind=kreal), DIMENSION(:)   ::  AL
    REAL   (kind=kreal) :: R_val,R_norm,GR_norm,v

    INTEGER(kind=kint ), DIMENSION(0:)  ::  INU
    INTEGER(kind=kint ), DIMENSION(:)   ::  IAU
    INTEGER(kind=kint ), DIMENSION(0:)  ::  INL
    INTEGER(kind=kint ), DIMENSION(:)   ::  IAL

    INTEGER(kind=kint), intent(in) :: LEVEL_NO

    INTEGER(kind=kint)  :: cnt,isL,ieL,isU,ieU,i,j,inod

    real(kind=kreal), allocatable :: temp(:)
    real(kind=kreal) :: omega

    omega = dump_gauss_seidel_weight
    allocate(temp(N))

    temp = X
    
    DO cnt=1,count

       DO i=1,N
          v = B(i)

          isL=INL(i-1)+1
          ieL=INL(i)
          DO j=isL,ieL
             inod=IAL(j)+ZERO_ORIGIN
             v = v - AL(j)*temp(inod)
          END DO

          isU=INU(i-1)+1
          ieU=INU(i)
          DO j=isU,ieU
             inod=IAU(j)+ZERO_ORIGIN
             v = v - AU(j)*temp(inod)
          END DO

          X(i) = omega * v / D(i)+(1-omega)*temp(i)
       END DO
    END DO

    deallocate(temp)
  END SUBROUTINE jacobi



  SUBROUTINE sgs_zero(N, NP, D, AL, INL, IAL, AU, INU, IAU,  X, count)
    
    USE data_structure_for_AMG
    IMPLICIT NONE

    INTEGER(kind=kint),INTENT(in )  :: N,NP,count
    REAL   (kind=kreal), DIMENSION(:)   ::  D
    REAL   (kind=kreal), DIMENSION(:)   ::  X

    REAL   (kind=kreal), DIMENSION(:)   ::  AU
    REAL   (kind=kreal), DIMENSION(:)   ::  AL
    REAL   (kind=kreal) :: R_val,R_norm,GR_norm,v

    INTEGER(kind=kint ), DIMENSION(0:)  ::  INU
    INTEGER(kind=kint ), DIMENSION(:)   ::  IAU
    INTEGER(kind=kint ), DIMENSION(0:)  ::  INL
    INTEGER(kind=kint ), DIMENSION(:)   ::  IAL

    INTEGER(kind=kint)  :: cnt,isL,ieL,isU,ieU,i,j,inod

    DO cnt=1,count
       DO i=1,N
          v = 0

          isL=INL(i-1)+1
          ieL=INL(i)
          DO j=isL,ieL
             inod=IAL(j)+ZERO_ORIGIN
             v = v - AL(j)*X(inod)
          END DO

          isU=INU(i-1)+1
          ieU=INU(i)
          DO j=isU,ieU
             inod=IAU(j)+ZERO_ORIGIN
             v = v - AU(j)*X(inod)
          END DO

          X(i) = v / D(i)
       END DO
       DO i=N,1,-1
          v = 0

          isL=INL(i-1)+1
          ieL=INL(i)
          DO j=isL,ieL
             inod=IAL(j)+ZERO_ORIGIN
             v = v - AL(j)*X(inod)
          END DO

          isU=INU(i-1)+1
          ieU=INU(i)
          DO j=isU,ieU
             inod=IAU(j)+ZERO_ORIGIN
             v = v - AU(j)*X(inod)
          END DO

          X(i) = v / D(i)
       END DO
    END DO
  END SUBROUTINE sgs_zero

  SUBROUTINE sgs_cr(count, LEVEL_NO, node_index, mu)

    USE data_structure_for_AMG
    IMPLICIT NONE

    INTEGER(kind=kint),INTENT(in )  :: count

    REAL   (kind=kreal), DIMENSION(:),allocatable   ::  X

    REAL   (kind=kreal) :: R_val,R_norm,GR_norm,v 
    REAL   (kind=kreal) :: e_nrm, e_nrm2



    REAL   (kind=kreal), DIMENSION(:), pointer  ::  AU
    REAL   (kind=kreal), DIMENSION(:), pointer  ::  AL
    REAL   (kind=kreal), DIMENSION(:), pointer  ::  D
    INTEGER(kind=kint ), DIMENSION(:), pointer  ::  INU
    INTEGER(kind=kint ), DIMENSION(:), pointer  ::  IAU
    INTEGER(kind=kint ), DIMENSION(:), pointer  ::  INL
    INTEGER(kind=kint ), DIMENSION(:), pointer  ::  IAL

    INTEGER(kind=kint), intent(in) :: LEVEL_NO

    INTEGER(kind=kint), intent(in) :: node_index(:)
    REAL(kind=kreal), intent(out)  :: mu
    
    INTEGER(kind=kint)  :: cnt,isL,ieL,isU,ieU,i,j,inod, N,NP,c

    REAL  :: rv
    
    N=   HIERARCHICAL_DATA(LEVEL_NO) % N
    NP=  HIERARCHICAL_DATA(LEVEL_NO) % NP
    INU=>HIERARCHICAL_DATA(LEVEL_NO) % INU 
    INL=>HIERARCHICAL_DATA(LEVEL_NO) % INL 
    D => HIERARCHICAL_DATA(LEVEL_NO) % D
    IAU=>HIERARCHICAL_DATA(LEVEL_NO) % IAU
    IAL=>HIERARCHICAL_DATA(LEVEL_NO) % IAL
    AU =>HIERARCHICAL_DATA(LEVEL_NO) % AU
    AL =>HIERARCHICAL_DATA(LEVEL_NO) % AL

    allocate(X(NP))

    !C random value is set to X(:)
    call random_seed()
    DO i=1, NP
       if(node_index(i) == 0) then
          call RANDOM_NUMBER(rv)
          X(i) = 0.5+1.0/2.0*rv
       else
          X(i) = 0.0
       end if

    END DO

    DO cnt=1,count

       DO i=1,N
          if(node_index(i) == 0) then
             v = 0.0

             isL=INL(i-1)+1
             ieL=INL(i)
             DO j=isL,ieL
                inod=IAL(j)+ZERO_ORIGIN
                v = v - AL(j)*X(inod)
             END DO

             isU=INU(i-1)+1
             ieU=INU(i)
             DO j=isU,ieU
                inod=IAU(j)+ZERO_ORIGIN
                v = v - AU(j)*X(inod)
             END DO

             X(i) = v / D(i)
          end if
       END DO
       DO i=N,1,-1
          if(node_index(i) == 0) then
             v = 0.0

             isL=INL(i-1)+1
             ieL=INL(i)
             DO j=isL,ieL
                inod=IAL(j)+ZERO_ORIGIN
                v = v - AL(j)*X(inod)
             END DO

             isU=INU(i-1)+1
             ieU=INU(i)
             DO j=isU,ieU
                inod=IAU(j)+ZERO_ORIGIN
                v = v - AU(j)*X(inod)
             END DO

             X(i) = v / D(i)
          end if
       END DO
       
       if(cnt >= count-1) then
          e_nrm = 0.0
          DO i = 1, NP
             e_nrm = e_nrm + X(i)*X(i)
          END DO
          e_nrm = dsqrt(e_nrm)
       end if

       if(cnt == count-1) e_nrm2=e_nrm
       
    END DO


    if(e_nrm2 == 0.0) then
       mu = 0.1
    else

       mu = e_nrm/e_nrm2
       if(mu<0.1) mu = 0.1
       
    end if

    deallocate(X)
    
  END SUBROUTINE sgs_cr


END MODULE  sgs_relaxation



