!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif

MODULE solver
  include 'precision.inc'
  logical :: REACH_CONV_FL
  real(kind=kreal) :: resid
  
CONTAINS
  SUBROUTINE CONV_STATE(FL)
    implicit none
    logical, intent(inout) :: FL
    
    FL = REACH_CONV_FL
    
  end SUBROUTINE CONV_STATE
  

  SUBROUTINE AMG_solver(X, B, iter, SETUP_FL)
    USE data_structure_for_AMG, only: HIERARCHICAL_DATA
    USE data_creation_stuben, only: clear_multi_lev_matrix, setup_phase
    USE problem_matrix
    USE amg_parameter, only: solver_string, number_of_vectors, &
         & dump_gauss_seidel_weight, cycle_kind, theta, gs_kind
    use solver_AMG  
    use solver_AMGCG
    use solver_AMGGMRES
    use solver_AMGBiCGSTAB
    use solver_IDR_s_GS
    use solver_IDR_s_AMG
    use solver_AMGRESCUT
    use terminal_condition, only: terminal_cond_check
    use residual_calculation, only: cal_res_vec
    !C ###########

#ifdef GPU
    use solver_GPU, only: mat_and_solver_set_on_GPU, solver_on_GPU
#endif

    implicit none
    
    integer(kind=kint), intent(inout) :: iter
    real(kind=kreal), intent(inout)   :: B(:)
    real(kind=kreal), intent(inout)   :: X(:)
    logical, intent(in)              :: SETUP_FL
!C    real(kind=kreal)                 :: resid
    
    integer(kind=kint) :: j
    real(kind=kreal) :: v
    
    !C-- for GPU
    INTEGER(kind=kint)  :: INT_DUM(1)
    REAL   (kind=kreal) :: REAL_DUM(1)
    integer(kind=kint) :: i,  R_NPLU, P_NPLU, R_NP, P_NP
    integer(kind=kint) :: sw=1, gpu_sol

    real(kind=kreal), allocatable :: res_vec(:)


    allocate(res_vec(NP))
    call cal_res_vec(res_vec, X, B)

    call terminal_cond_check(REACH_CONV_FL, resid, res_vec, NP)
    deallocate(res_vec)

    
    
    if(REACH_CONV_FL) then
#ifdef PRINT_REZ
       write(*,*) "reach convergence without calling solver"
#endif         
       iter = 0
       return
    end if
    


    if(SETUP_FL) then
       if(LEVEL_NUM > 0) then
          call clear_multi_lev_matrix(LEVEL_NUM)
       end if
    else
       if(LEVEL_NUM == 0) then
          call setup_phase()

#ifdef GPU          
          if(    solver_string(1:1) == 'G' .and. &
               & solver_string(2:2) == 'P' .and. &
               & solver_string(3:3) == 'U' .and. &
               & solver_string(4:4) == 'B' .and. &
               & solver_string(5:5) == 'I' .and. &
               & solver_string(6:6) == 'C') then
             gpu_sol = 1
             call mat_and_solver_set_on_GPU(LEVEL_NUM, gpu_sol, dump_gauss_seidel_weight)
          else if(solver_string(1:1) == 'G' .and. &
               & solver_string(2:2) == 'P' .and. &
               & solver_string(3:3) == 'U') then
             gpu_sol = 0
             call mat_and_solver_set_on_GPU(LEVEL_NUM, gpu_sol, dump_gauss_seidel_weight)
          end if
#endif
       end if
    end if

    REACH_CONV_FL = .false.

    
    if(solver_string(1:1)=='N' .and. solver_string(2:2)=='O') then
#ifdef PRINT_REZ
       if(gs_kind == 0) then
          write(*,*) "AMG", " sgs", " GS dump parameter=", dump_gauss_seidel_weight, &
               & " CYCLE: 0:V,1:F,2:W : ", cycle_kind, "theta:", theta
       else
          write(*,*) "AMG", " gs", " GS dump parameter=", dump_gauss_seidel_weight, &
               & " CYCLE: 0:V,1:F,2:W : ", cycle_kind, "theta:", theta
       end if
#endif
       call AMG(NP,  NPL, NPU,                           &
            &   D,  AL, INL, IAL, AU, INU, IAU,          &
            &   B,  X, resid, iter, SETUP_FL, LEVEL_NUM, REACH_CONV_FL)
       
    else if(solver_string(1:1)=='C' .and. solver_string(2:2)=='G') then
#ifdef PRINT_REZ
       if(gs_kind == 0) then
          write(*,*) "AMGCG"," sgs", " GS dump parameter=", dump_gauss_seidel_weight,&
               & " CYCLE: 0:V,1:F,2:W : ", cycle_kind, "theta:", theta
       else
          write(*,*) "AMGCG"," gs", " GS dump parameter=", dump_gauss_seidel_weight,&
               & " CYCLE: 0:V,1:F,2:W : ", cycle_kind, "theta:", theta
       end if
#endif


       call AMGCG(NP,  NPL, NPU,                         &
            &   D,  AL, INL, IAL, AU, INU, IAU,          &
            &   B,  X, resid, iter, SETUP_FL, LEVEL_NUM, NULL_SPACE_CONSTANT, &
            & REACH_CONV_FL)
       
    else if(solver_string(1:1)=='G' .and. solver_string(2:2)=='M' .and. &
         &  solver_string(3:3)=='R' .and. solver_string(4:4)=='E') then
#ifdef PRINT_REZ
       if(gs_kind == 0) then
          write(*,*) "AMGGMRES(",number_of_vectors, ")"," sgs", " GS dump parameter=", &
               & dump_gauss_seidel_weight," CYCLE: V,F,W : ", cycle_kind, "theta:", theta
       else
          write(*,*) "AMGGMRES(",number_of_vectors, ")"," gs", " GS dump parameter=", &
               & dump_gauss_seidel_weight," CYCLE: V,F,W : ", cycle_kind, "theta:", theta
          
       end if
#endif
       call AMGGMRES(number_of_vectors, NP,  NPL, NPU,   &
            &   D,  AL, INL, IAL, AU, INU, IAU,          &
            &   B,  X, resid, iter, SETUP_FL, LEVEL_NUM,REACH_CONV_FL)
    else if(solver_string(1:1)=='B' .and. solver_string(2:2)=='I' .and. &
         &  solver_string(3:3)=='C' .and. solver_string(4:4)=='G' .and. &
         &  solver_string(5:5)=='S' .and. solver_string(6:6)=='T') then
#ifdef PRINT_REZ       
       if(gs_kind == 0) then
          write(*,*) "AMGBICGSTAB"," sgs", " GS dump parameter=", &
               & dump_gauss_seidel_weight," CYCLE: 0:V,1:F,2:W : ", cycle_kind, "theta:", theta
       else
          write(*,*) "AMGBICGSTAB"," gs", " GS dump parameter=", &
               & dump_gauss_seidel_weight," CYCLE: 0:V,1:F,2:W : ", cycle_kind, "theta:", theta
       end if
#endif
          
       call AMGBiCGSTAB(NP,  NPL, NPU,                   &
            &   D,  AL, INL, IAL, AU, INU, IAU,          &
            &   B,  X, resid, iter, SETUP_FL, LEVEL_NUM, REACH_CONV_FL)
    else if(solver_string(1:1)=='I' .and. solver_string(2:2)=='D' .and. &
         &  solver_string(3:3)=='R')  then
       
#ifdef PRINT_REZ
       if(gs_kind == 0) then
          write(*,*) "AMG-based-IDR(",number_of_vectors, ")", &
               & " sgs", " GS dump parameter=", dump_gauss_seidel_weight,&
               & " CYCLE: 0:V,1:F,2:W : ", cycle_kind, "theta:", theta
       else
          write(*,*) "AMG-based-IDR(",number_of_vectors, ")", &
               & " gs", " GS dump parameter=", dump_gauss_seidel_weight,&
               & " CYCLE: 0:V,1:F,2:W : ", cycle_kind, "theta:", theta
       end if
#endif

       call IDR_s_AMG(number_of_vectors, NP,  NPL, NPU,  &
            &   D,  AL, INL, IAL, AU, INU, IAU,          &
            &   B,  X, resid, iter, SETUP_FL, LEVEL_NUM, REACH_CONV_FL)
    else if(solver_string(1:1)=='R' .and. solver_string(2:2)=='E' .and. &
         &  solver_string(3:3)=='S')  then

#ifdef PRINT_REZ       
       if(gs_kind == 0) then
          write(*,*) "AMGRESCUT(",number_of_vectors, ")", &
               & " sgs"," GS dump parameter=", dump_gauss_seidel_weight,&
               & " CYCLE: 0:V,1:F,2:W : ", cycle_kind, "theta:", theta
       else
          write(*,*) "AMGRESCUT(",number_of_vectors, ")", &
               & " gs"," GS dump parameter=", dump_gauss_seidel_weight,&
               & " CYCLE: 0:V,1:F,2:W : ", cycle_kind, "theta:", theta
       end if
#endif
       
       call AMGRESCUT(number_of_vectors, NP,  NPL, NPU,  &
            &   D,  AL, INL, IAL, AU, INU, IAU,          &
            &   B,  X, resid, iter, SETUP_FL, LEVEL_NUM,REACH_CONV_FL)

#ifdef GPU       
    else if(solver_string(1:1)=='G' .and. solver_string(2:2)=='P' .and. &
         &  solver_string(3:3)=='U')  then
       
       call solver_on_GPU(NP,  NPL, NPU,                               &
            &   D,  AL, INL, IAL, AU, INU, IAU,                        &
            &   B,  X, resid, iter, SETUP_FL, LEVEL_NUM, REACH_CONV_FL)
#endif


    else
       write(*,*) "No solver is specified by ", solver_string 
    end if



    !C-- remove the constant component of Solution X
    if(NULL_SPACE_CONSTANT) then
       v = 0
       do j = 1, NP
          v = v + X(j)
       end do
       v = v/NP
       do j = 1, NP
          X(j) = X(j) - v
       end do
    end if
    
  END SUBROUTINE AMG_SOLVER
END MODULE  solver
