!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

module solver_AMGBiCGSTAB
contains

  SUBROUTINE AMGBiCGSTAB    (N,  NPL, NPU,                                &
       &                  D,  AL, INL, IAL, AU, INU, IAU,                 &
       &                  B,  X, resid,iter, SETUP_FL, LEVEL_NUM,         &
       &                  REACH_CONV_FL)
    
    use data_structure_for_AMG
    USE data_creation_stuben
    USE AMG_cycle
    USE terminal_condition, only:terminal_cond_vec_set, terminal_cond_val_set, &
         & terminal_cond_check, BNRM2, terminal_kind,conv_criterion_val
    USE residual_calculation
    USE amg_parameter, only: conv_check_fl

    implicit none

    logical, intent(inout) :: REACH_CONV_FL
    integer(kind=kint ),                   intent(in   )::  N
    integer(kind=kint ),                   intent(in   )::  NPU, NPL

    real   (kind=kreal),                   intent(inout)::  RESID
    integer(kind=kint ),                   intent(inout)::  ITER

    logical, intent(in)  :: SETUP_FL

    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INU
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAU
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INL
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAL
    REAL   (kind=kreal), DIMENSION(:), pointer ::  AU
    REAL   (kind=kreal), DIMENSION(:), pointer ::  AL

    REAL   (kind=kreal), DIMENSION(: ),  pointer ::  D
    REAL   (kind=kreal), DIMENSION(: )  ::  B
!!$    REAL   (kind=kreal), DIMENSION(NP )  ::  X
    REAL   (kind=kreal), DIMENSION(: )  ::  X


    real(kind=kreal), allocatable, save ::  WW(:,:)

    real(kind=kreal)                    :: C0(2), CG(2)

    integer(kind=kint ) :: IFLAG, R, RT, P, PT, S, ST, T, V, MAXIT,ierr
    integer(kind=kint ) :: i,j,k,l,m
    real   (kind=kreal) :: TOL, W, SS, WVAL, RHO0,RHO,OMEGA,ALPHA,RHO1
    real   (kind=kreal) :: BETA,C2, BNRM20

    real (kind=kreal) :: val, bval
    integer(kind=kint) :: LEVEL_NUM, SISIZE, SESIZE, t1,t2,t_rate,t_max
    integer, save :: N_SIZE=0
    logical :: amgprecond, CONV_FL
    
    amgprecond = .true.
    REACH_CONV_FL = .false.

    if(N_SIZE < N) then
       if(allocated(WW)) deallocate(WW)
       allocate (WW(N,8))
    end if


    R=  1; RT= 2; P=  3; PT= 4; S=  5; ST= 6; T=  7; V=  8

    MAXIT = ITER
!    TOL   = RESID
!    conv_criterion_val = RESID

    if(amgprecond .and. SETUP_FL ) then
       call system_clock(t1, t_rate, t_max)       
       CALL data_creation(N, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, LEVEL_NUM)
       call system_clock(t2, t_rate, t_max)
    end if
    

    do j= 1, N
       WVAL = B(j) - D(j)*X(j)

       do k= INL(j-1)+1, INL(j)
          i= IAL(k)
          WVAL= WVAL - AL(k)*X(i)
       enddo

       do k= INU(j-1)+1, INU(j)
          i= IAU(k)
          WVAL= WVAL - AU(k)*X(i)
       enddo

       WW(j  ,R )= WVAL
       WW(j  ,RT)= WVAL

       WW(j  ,PT)= 0.d0
       WW(j  ,ST)= 0.d0
    enddo
    
!!$    if(terminal_kind == 0) then
!!$       BNRM20 = 0.d0
!!$       do i= 1, N
!!$          BNRM20 = BNRM20 + B(i)**2
!!$       enddo
!!$       
!!$       BNRM2=BNRM20
!!$    else if(terminal_kind == 1) then
!!$       !C record infinity norm
!!$       BNRM20 = 0.d0
!!$       do i = 1, N
!!$          val = dabs(B(i))
!!$          if(v > BNRM20) BNRM20 = val
!!$       end do
!!$       
!!$       BNRM2 = BNRM20
!!$    end if
!!$
!!$    if (BNRM2.eq.0.d0) BNRM2= 1.d0
    

    do iter= 1, MAXIT
       RHO= 0.d0
       do j= 1, N
          RHO= RHO+WW(j,RT)*WW(j,R)
       enddo

       if (iter.gt.1) then
          BETA= (RHO/RHO1) * (ALPHA/OMEGA)
          do j= 1, N
             WW(j  ,P)= WW(j  ,R)+BETA*(WW(j  ,P)-OMEGA*WW(j  ,V))
          enddo
!!$          !C-debug
!!$          write(*,*) "--------ITER--------:", iter
!!$          write(*,*) "BETA:", BETA
!!$          write(*,*) "OMEGA:", OMEGA
!!$          val = 0
!!$          DO j = 1, N
!!$             val = val+WW(j  ,R)**2
!!$          END DO
!!$          write(*,*) ,"R:", dsqrt(val)
!!$          val = 0
!!$          DO j = 1, N
!!$             val = val+WW(j  ,P)**2
!!$          END DO
!!$          write(*,*) ,"P :", dsqrt(val)
!!$          val = 0
!!$          DO j = 1, N
!!$             val = val+WW(j  ,V)**2
!!$          END DO
!!$          write(*,*) ,"V:", dsqrt(val)
!!$          !C-debug
          
       else
          do j= 1, N
             WW(j  ,P)= WW(j  ,R)
          enddo
       endif
       if(amgprecond) then
          do j= 1, N
             WW(j,PT) = 0.0d0
          end do


!!$          !C- debug
!!$          val = 0
!!$          DO j = 1, N
!!$             val = val+WW(j  ,PT)**2
!!$          END DO
!!$          write(*,*) ,"PT before AMG:", dsqrt(val)
!!$          !C- debug
!!$          !C- debug
!!$          val = 0
!!$          DO j = 1, N
!!$             val = val+WW(j  ,P)**2
!!$          END DO
!!$          write(*,*) ,"P before AMG:", dsqrt(val)
!!$          !C- debug


          CALL amg_cycles(N, WW(:,P), WW(:,PT), LEVEL_NUM, WW(:,V))          

!!$          !C- debug
!!$          val = 0
!!$          DO j = 1, N
!!$             val = val+WW(j  ,PT)**2
!!$          END DO
!!$          write(*,*) ,"PT after AMG:", dsqrt(val)
!!$          !C- debug
!!$          !C- debug
!!$          val = 0
!!$          DO j = 1, N
!!$             val = val+WW(j  ,P)**2
!!$          END DO
!!$          write(*,*) ,"P after AMG:", dsqrt(val)
!!$          !C- debug

          
       else 
          do j= 1, N
             WW(j  ,PT)= WW(j  ,P)
          enddo
       end if
       
       do j= 1, N
          WVAL= D(j)*WW(j,PT)
          do k= INL(j-1)+1, INL(j)
             i= IAL(k)
             WVAL= WVAL + AL(k)*WW(i,PT)
          enddo
          do k= INU(j-1)+1, INU(j)
             i= IAU(k)
             WVAL= WVAL + AU(k)*WW(i,PT)
          enddo

          WW(j,V)= WVAL
       enddo

       C2= 0.d0
       do j= 1, N
          C2= C2 + WW(j,RT)*WW(j,V) 
       enddo

       ALPHA= RHO / C2

       do j= 1, N
          WW(j  ,S)= WW(j  ,R) - ALPHA*WW(j  ,V)
       enddo


       

       if(amgprecond) then
          do j = 1, N
             WW(j, ST) = 0.0d0
          end do

!!$          !C- debug
!!$          val = 0
!!$          DO j = 1, N
!!$             val = val+WW(j  ,ST)**2
!!$          END DO
!!$          write(*,*) ,"ST before AMG:", dsqrt(val)
!!$          !C- debug
!!$          !C- debug
!!$          val = 0
!!$          DO j = 1, N
!!$             val = val+WW(j  ,S)**2
!!$          END DO
!!$          write(*,*) ,"S before AMG:", dsqrt(val)
!!$          !C- debug


          CALL amg_cycles(N, WW(:,S), WW(:,ST), LEVEL_NUM, WW(:,T))          

!!$          !C- debug
!!$          val = 0
!!$          DO j = 1, N
!!$             val = val+WW(j  ,ST)**2
!!$          END DO
!!$          write(*,*) ,"ST after AMG:", dsqrt(val)
!!$          !C- debug
!!$          !C- debug
!!$          val = 0
!!$          DO j = 1, N
!!$             val = val+WW(j  ,S)**2
!!$          END DO
!!$          write(*,*) ,"S after AMG:", dsqrt(val)
!!$          !C- debug

          
       else
          do j= 1, N
             WW(j  ,ST)= WW(j  ,S)
          enddo
       end if
       
       do j= 1, N
          WVAL= D(j)*WW(j  ,ST)
          do k= INL(j-1)+1, INL(j)
             i= IAL(k)
             WVAL= WVAL + AL(k)*WW(i  ,ST)
          enddo
          do k= INU(j-1)+1, INU(j)
             i= IAU(k)
             WVAL= WVAL + AU(k)*WW(i  ,ST)
          enddo

          WW(j  ,T)= WVAL
       enddo

       C0(1)= 0.d0; C0(2)= 0.d0; CG(1)= 0.d0; CG(2)= 0.d0

       do j= 1, N
          C0(1)= C0(1) + WW(j,T)*WW(j,S)
          C0(2)= C0(2) + WW(j,T)*WW(j,T)
       enddo

       CG(1) = C0(1); CG(2)=C0(2)
       
       OMEGA= CG(1) / CG(2)
       do j= 1, N
          X (j  )= X(j  ) + ALPHA*WW(j  ,PT) + OMEGA*WW(j  ,ST)
          WW(j  ,R)= WW(j  ,S) - OMEGA*WW(j  ,T)
       enddo

       RHO1= RHO

!!$       !C- debug
!!$       bval = 0; val = 0
!!$       DO j = 1, N
!!$          val = val+WW(j  ,R)**2
!!$          bval = bval + B(j)**2
!!$       END DO
!!$       write(*,*) ,"Res 2norm", iter, ":", dsqrt(val), val
!!$       !C- debug

       
       call terminal_cond_check(CONV_FL, RESID, WW(:,R), N)
       
       
       
#ifdef PRINT_REZ
       if (terminal_kind==0) write (*, 1000) ITER, RESID
1000   format (i5, 1pe16.6)
#endif
       if(terminal_kind==1) then
          RESID = 0 
          do j = 1, N
             RESID = RESID + WW(j,R)**2
          end do
          RESID = dsqrt(RESID)
       end if
             
       
       if(CONV_FL) then
          REACH_CONV_FL = .true.
          
#ifdef PRINT_REZ       
          if(SETUP_FL) then
             write(*,*) "Setup time: ", (t2-t1)/dble(t_rate)          
          end if
#endif          
          exit
!!$       else if(isNAN(RESID)) then
!!$          exit
       end if
       
    enddo

    !C check convergence by real residual
!!$    if(.not.CONV_FL .or. conv_check_fl) then
    if(conv_check_fl) then
       call cal_res_vec(WW(:,R), X, B)
       call terminal_cond_check(CONV_FL, RESID, WW(:,R), N)
       if(.not.CONV_FL) then
          REACH_CONV_FL = .false.
          if(RESID > 1.0) then
             iter = -1
          end if
       else
          REACH_CONV_FL = .true.
       end if
    end if

    if(SETUP_FL) then
       deallocate (WW)
       N_SIZE = 0
    end if

    if(amgprecond .and. SETUP_FL) then
       call clear_multi_lev_matrix(LEVEL_NUM)
    end if
    
  end subroutine        AMGBiCGSTAB
end module     solver_AMGBiCGSTAB

