!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif


MODULE solver_AMGCG
CONTAINS
  !C SETUP_FL shows if this subroutine contains the setup phase
  !C and deletes the matrix
  SUBROUTINE AMGCG    (N,  NPL, NPU,                                  &
       &                  D,  AL, INL, IAL, AU, INU, IAU,             &
       &                  B,  X, resid,iter, SETUP_FL, LEVEL_NUM,     &
       &                  NULL_SPACE_CONSTANT,REACH_CONV_FL)
    USE data_structure_for_AMG
    USE data_creation_stuben
    USE sgs_relaxation
    USE ilu_relaxation
    USE AMG_cycle
    USE terminal_condition, only:terminal_cond_vec_set, terminal_cond_val_set, &
         & terminal_cond_check, BNRM2, terminal_kind, conv_criterion_val
    USE residual_calculation
    USE amg_parameter, only: conv_check_fl
    
    IMPLICIT NONE
    LOGICAL, intent(in) :: SETUP_FL, NULL_SPACE_CONSTANT
    LOGICAL, intent(inout) :: REACH_CONV_FL
    INTEGER(kind=kint ), intent(inout) :: LEVEL_NUM
    INTEGER(kind=kint ), intent(in)  :: N,NPL,NPU
    REAL   (kind=kreal)              ::  RESID, TOL
    INTEGER(kind=kint ), intent(inout) ::  ITER

    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INU
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAU
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INL
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAL

    REAL   (kind=kreal), DIMENSION(: ),  pointer ::  D
    REAL   (kind=kreal), DIMENSION(: )  ::  B
!!$    REAL   (kind=kreal), DIMENSION(NP )  ::  X
    REAL   (kind=kreal), DIMENSION(: )  ::  X

    !C pointers for ILU
    REAL   (kind=kreal), DIMENSION(:), pointer ::  ILU_D, ILU_AL, ILU_AU

    REAL   (kind=kreal), DIMENSION(:), pointer ::  AU
    REAL   (kind=kreal), DIMENSION(:), pointer ::  AL
    REAL   (kind=kreal), DIMENSION(:,:), ALLOCATABLE :: WW
    REAL   (kind=kreal), DIMENSION(:),   ALLOCATABLE :: Temp
    REAL   (kind=kreal) :: WVAL,RHO0,RHO,BETA,RHO1,C10,C1,ALPHA
    
    REAL   (kind=kreal) :: DNRM2,BNRM20,DNRM20, v
    REAL   (kind=kreal) :: STARTTIME,ENDTIME,RTIME,STARTTIME2,ENDTIME2,RTIME2
    INTEGER(kind=kint ) :: P, Q, R, Z, MAXIT, IFLAG=0, iter2

    INTEGER(kind=kint)  :: cnt,NEIBPETOT,I,J,ISU,IEU,ISL,IEL,WSIZE
    INTEGER(kind=kint)  :: INOD
    
    logical :: CONV_FL

    INTEGER(kind=kint) :: t1, t2, t_rate, t_max
    
    ALLOCATE (WW(N,3), Temp(N))

    R = 1
    Z = 2
    Q = 2
    P = 3

    MAXIT  = ITER
    REACH_CONV_FL =.false.
!C    conv_criterion_val   = RESID           

    call system_clock(t1, t_rate, t_max)
    if(SETUP_FL) then

       CALL data_creation(N, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, LEVEL_NUM)

    end if


!!$    !C arrays are allocated for ILU0
!!$    allocate(ILU_D(NP), ILU_AL(NPL), ILU_AU(NPU))
!!$    ILU_D(:) = D(:)
!!$    ILU_AL(:) = AL(:)
!!$    ILU_AU(:) = AU(:)
!!$    call ilu0_setup(NP, ILU_D, ILU_AL, INL, IAL, ILU_AU, INU, IAU)
    
    call system_clock(t2, t_rate, t_max)
    
    do j= 1, N
       WVAL= B(j) - D(j) * X(j)
       isU= INU(j-1) + 1
       ieU= INU(j  ) 

       do i= isU, ieU
          inod= IAU(i)+ZERO_ORIGIN
          WVAL= WVAL - AU(i) * X(inod)
       enddo

       isL= INL(j-1) + 1
       ieL= INL(j  ) 
       do i= isL, ieL
          inod= IAL(i)+ZERO_ORIGIN
          WVAL= WVAL - AL(i) * X(inod)
       enddo
       WW(j,R)= WVAL
    enddo

!!$    if(terminal_kind == 0) then
!!$       BNRM20 = 0.d0
!!$       do i= 1, N
!!$          BNRM20 = BNRM20 + B(i)**2
!!$       enddo
!!$       
!!$       BNRM2=BNRM20
!!$    else if(terminal_kind == 1) then
!!$       !C record infinity norm
!!$       BNRM20 = 0.d0
!!$       do i = 1, N
!!$          v = dabs(B(i))
!!$          if(v > BNRM20) BNRM20 = v
!!$       end do
!!$       
!!$       BNRM2 = BNRM20
!!$    end if
!!$
!!$    if (BNRM2.eq.0.d0) BNRM2= 1.d0
    ITER = 0


    do iter= 1, MAXIT
       do i=1,N
          WW(i,Z)=0.0
       end do


       CALL amg_cycles(N, WW(:,R), WW(:,Z), LEVEL_NUM, Temp)
!!$       call sgs(N, N, D, AL, INL, IAL, AU, INU, IAU, WW(:,R), WW(:,Z), 1, 1)
!!$       CALL ilu(N, D, ILU_D, AL, ILU_AL, INL, IAL, AU, ILU_AU, INU, IAU, WW(:,R), WW(:,Z), Temp)       

!!$       WW(:,Z) = WW(:,R)

       RHO0= 0.0

       do i= 1, N
          RHO0= RHO0 + WW(i,R)*WW(i,Z)
       enddo

       RHO=RHO0
       if ( ITER.eq.1 ) then
          do i= 1, N
             WW(i,P)= WW(i,Z)
          enddo
       else
          BETA= RHO / RHO1
          do i= 1, N
             WW(i,P)= WW(i,Z) + BETA*WW(i,P)
          enddo
       endif


       !C q = A p
       do j= 1, N
          WVAL= D(j) * WW(j,P)

          isU= INU(j-1) + 1
          ieU= INU(j  ) 

          do i= isU, ieU
             inod= IAU(i)+ZERO_ORIGIN
             WVAL= WVAL + AU(i) * WW(inod,P)
          enddo

          isL= INL(j-1) + 1
          ieL= INL(j  ) 

          do i= isL, ieL
             inod= IAL(i)+ZERO_ORIGIN
             WVAL= WVAL + AL(i) * WW(inod,P)
          enddo
          WW(j,Q)= WVAL
          
       enddo
       
       if(NULL_SPACE_CONSTANT) then
          !C---- Q = (A+ e e^T) P
          WVAL = 0
          do j = 1, N
             WVAL = WVAL + WW(j, P)
          end do
          WVAL = WVAL/N
          do j = 1, N
             WW(j, Q) = WW(j, Q)+ WVAL/N
          end do
          !C----
       end if

       C1= 0.d0
       do i= 1, N
          C1= C1 + WW(i,P)*WW(i,Q)
       enddo

       ALPHA= RHO / C1

       DNRM20= 0.d0
       
       do i= 1, N
          X(i)  = X (i)   + ALPHA * WW(i,P)
          WW(i,R)= WW(i,R) - ALPHA * WW(i,Q)
       enddo

       RHO1 = RHO                                                             

!!$       DNRM20 = 0.0
!!$       do i= 1, N
!!$          DNRM20= DNRM20 + WW(i,R)**2
!!$       enddo
!!$       
!!$       DNRM2=DNRM20
!!$       RESID= dsqrt(DNRM2/BNRM2)

       call terminal_cond_check(CONV_FL, RESID, WW(:, R), N)
       
#ifdef PRINT_REZ
       write (*, 1000) ITER, RESID
1000   format (i5, 1pe16.6)
#endif
       
!!$       if ( RESID.le.TOL   ) then
       if ( CONV_FL   ) then

          REACH_CONV_FL = .true.
!C          call single_matrix_counting(LEVEL_NUM)

#ifdef PRINT_REZ
          if(SETUP_FL) then
             write(*,*) "Setup time: ", (t2-t1)/dble(t_rate)
          end if
#endif
          exit
       end if
!!$       if(terminal_kind==0 .and.  RESID > 1.0E5 .or. isNAN(RESID)) then
       if(terminal_kind==0 .and.  RESID > 1.0E5) then
          write(*,*) "Divergence!"
          exit
       end if
    end do


    !C check convergence by real residual
!!$    if(.not.CONV_FL .or. conv_check_fl) then
    if(conv_check_fl) then
       call cal_res_vec(WW(:,R), X, B)
       call terminal_cond_check(CONV_FL, RESID, WW(:,R), N)
       if(.not.CONV_FL) then
          REACH_CONV_FL = .false.
          !C AMGCG often makes convergence unstable.
          !C these operations are comment-out.
          if(RESID > 1.0) then
             iter = -1
          end if
       else
          REACH_CONV_FL = .true.
       end if
    end if

    !C===
    if(SETUP_FL) then
       call clear_multi_lev_matrix(LEVEL_NUM)
    end if
    deallocate(WW, Temp)
  END SUBROUTINE        AMGCG


END MODULE  solver_AMGCG
