!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif
  
  MODULE solver_AMGGMRES
  CONTAINS
    
    SUBROUTINE AMGGMRES    (KRES, NP,  NPL, NPU,                         &
         &                  D,  AL, INL, IAL, AU, INU, IAU,              &
         &                  B,  X, resid,iter, SETUP_FL, LEVEL_NUM,      &
         &                  REACH_CONV_FL)     

      USE data_structure_for_AMG
      USE data_creation_stuben
      USE sgs_relaxation
      USE ilu_relaxation
      USE AMG_cycle
      USE terminal_condition, only:terminal_cond_vec_set, terminal_cond_val_set, &
           & terminal_cond_check, BNRM2, terminal_kind, conv_criterion_val,      &
           & conv_criterion_vec2
      USE residual_calculation, only: cal_res_vec
      USE amg_parameter, only: conv_check_fl

      IMPLICIT NONE
      LOGICAL, intent(in) :: SETUP_FL
      LOGICAL, intent(inout) :: REACH_CONV_FL
      INTEGER(kind=kint ), intent(inout) :: LEVEL_NUM
      INTEGER(kind=kint ), intent(in)  :: NP,NPL,NPU
      REAL   (kind=kreal)              ::  RESID, TOL
      INTEGER(kind=kint )              ::  ITER

      INTEGER(kind=kint )              :: SOLVER_COMM
      
      INTEGER(kind=kint ), DIMENSION(:), pointer ::  INU
      INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAU
      INTEGER(kind=kint ), DIMENSION(:), pointer ::  INL
      INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAL
      REAL   (kind=kreal), DIMENSION(:), pointer ::  AU
      REAL   (kind=kreal), DIMENSION(:), pointer ::  AL
      REAL   (kind=kreal), DIMENSION(:), pointer ::  D

      REAL   (kind=kreal), DIMENSION(:)  ::  B
!!$    REAL   (kind=kreal), DIMENSION(NP )  ::  X
      REAL   (kind=kreal), DIMENSION(:)  ::  X

      !C pointers for ILU
      REAL   (kind=kreal), DIMENSION(:), pointer ::  ILU_D, ILU_AL, ILU_AU


      INTEGER(kind=kint ), intent(in) :: KRES

      REAL   (kind=kreal), ALLOCATABLE :: V(:,:), H(:,:), E(:), C(:), S(:), Q(:)
      REAL   (kind=kreal), DIMENSION(:),   ALLOCATABLE :: Temp

      REAL   (kind=kreal) :: BETA,val,abc, tbs, RHO
      INTEGER(kind=kint)  :: kk,k,col,N
      REAL   (kind=kreal) :: WVAL, BNRM2_sq, abb, BNRM20
      INTEGER(kind=kint)  :: i,j,isu,ieu,isl,iel,inod
      INTEGER(kind=kint ) :: MAXIT

      INTEGER(kind=kint) :: t1, t2, t_rate, t_max
      REAL(kind=kreal), pointer :: RES_VEC(:)
      logical :: CONV_FL, CONV_FL_VEC

      !C-- INIT. 
      REACH_CONV_FL =.false.
      CONV_FL = .false.
      CONV_FL_VEC = .false.

      N = NP

      ALLOCATE(V(NP, KRES+1), H(KRES+1, KRES), E(KRES+1), C(KRES), S(KRES), Q(NP))
      ALLOCATE ( Temp(NP) )
      
      MAXIT  = ITER
      TOL   = RESID           


      call system_clock(t1, t_rate, t_max)
      if(SETUP_FL) then
         CALL data_creation(NP, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, LEVEL_NUM)
      end if

!!$    !C arrays are allocated for ILU0
!!$    allocate(ILU_D(NP), ILU_AL(NPL), ILU_AU(NPU))
!!$    ILU_D(:) = D(:)
!!$    ILU_AL(:) = AL(:)
!!$    ILU_AU(:) = AU(:)
!!$    call ilu0_setup(NP, ILU_D, ILU_AL, INL, IAL, ILU_AU, INU, IAU)

      call system_clock(t2, t_rate, t_max)


      BNRM2 = 0.d0
      do i= 1, N
         BNRM2 = BNRM2 + B(i)**2
      enddo
      
      BNRM2_sq = dsqrt(BNRM2)

      if (BNRM2.eq.0.d0) BNRM2= 1.d0

      allocate(RES_VEC(N))
      
      if(terminal_kind==1)then
         !C BNRM2 is determined here..
         !C So conv_criterion_val is also determined.
         conv_criterion_val = dsqrt(conv_criterion_vec2/BNRM2)
      end if

      iter = 0

      do kk = 1, MAXIT/KRES+1
         BETA = 0
         do j= 1, N
            WVAL= B(j) - D(j) * X(j)
            isU= INU(j-1) + 1
            ieU= INU(j  ) 
            do i= isU, ieU
               inod= IAU(i)+ZERO_ORIGIN
               WVAL= WVAL - AU(i) * X(inod)
            enddo
            isL= INL(j-1) + 1
            ieL= INL(j  ) 
            do i= isL, ieL
               inod= IAL(i)+ZERO_ORIGIN
               WVAL= WVAL - AL(i) * X(inod)
            enddo
            V(j,1)= WVAL
            BETA = BETA+WVAL*WVAL
         enddo
         BETA = DSQRT(BETA)
         do j = 1, N
            V(j,1) = V(j,1) / BETA
         end do
         E(1) = BETA
         do j = 2, KRES+1
            E(j) = 0.0
         end do
         do k = 1, KRES
            do i = 1, NP
               Q(i)=0.0
            end do
!!$          CALL v_cycle(N, V(:,k), Q(:), LEVEL_NUM, Temp)
            CALL amg_cycles(N, V(:,k), Q(:), LEVEL_NUM, Temp)
!!$          Q(:) = V(:,k)

            do i = 1, NP
               val = D(i)*Q(i)
               isl = INL(i-1)+1; iel = INL(i)
               do j = isl, iel
                  col = IAL(j)
                  val = val+AL(j)*Q(col)
               end do
               isu = INU(i-1)+1; ieu = INU(i)
               do j = isu, ieu
                  col = IAU(j)
                  val = val+AU(j)*Q(col)
               end do
               V(i,k+1) = val
            end do

            do j = 1, k
               H(j,k) = 0.0
               do i = 1, NP
                  H(j,k) = H(j,k) + V(i,k+1)*V(i,j)
               end do
               do i = 1, NP
                  V(i,k+1) = V(i,k+1) - H(j,k)*V(i,j)
               end do
            end do

            val = 0
            do i = 1, NP
               val = val + V(i,k+1)*V(i,k+1)
            end do
            val = dsqrt(val)
            H(k+1,k) = val

            do i = 1,NP
               V(i,k+1) = v(i,k+1)/val
            end do

            if(k > 1) then
               do j = 1, k-1
                  abc = C(j)*H(j,k)-S(j)*H(j+1,k)
                  tbs = S(j)*H(j,k)+C(j)*H(j+1,k)
                  H(j,k) = abc
                  H(j+1,k) = tbs
               end do
            end if
            RHO = dsqrt(H(k,k)**2+H(k+1,k)**2)
            C(k) = H(k,k)/RHO
            S(k) = -H(k+1,k)/RHO
            E(k+1) = S(k)*E(k)
            E(k)   = C(k)*E(k)
            H(k,k) = RHO
            H(k+1,k) = 0.0
            RESID = dabs(E(k+1))/BNRM2_sq

            iter = iter + 1
#ifdef PRINT_REZ
!C            write (*, 1000) (kk-1)*KRES+k, RESID
            write (*, 1000) iter, RESID
1000        format (i5, 1pe16.6)

#endif

            !C conv_criterion_val is set for two terminal conditions of val and vec
            CONV_FL = (RESID < conv_criterion_val)

!!$          if(RESID <= TOL)then
!C            iter = (kk-1)*KRES+k
            if(CONV_FL .or. iter >= MAXIT)then
               !C call direct solver for upper triangular matrix 
               call psolver(k)


               !C RESID is changed to 2-norm of relative residual.
               RESID = dsqrt(RESID)

               !C for amg preconditioner
               !C             call single_matrix_counting(LEVEL_NUM)

               exit
            end if

!!$            if( RESID >= 1.0E8 .or. isNAN(RESID)) then
            if( RESID >= 1.0E8) then
               write(*,*) "Divergence!"
!C               iter = (kk-1)*KRES+k
               exit
            end if

         end do

         !C terminal_kind==0 and reach convergence
         if(CONV_FL .and. terminal_kind==0) then
            REACH_CONV_FL = .true.
            exit
         end if
         
!!$         if(isNAN(RESID)) then
!!$            exit
!!$         end if


         !C call direct solver for upper triangular matrix 
         if(.not.CONV_FL) call psolver(KRES)

         !C for terminal_kind == 1. Here, residual vector is calculated.
         if(terminal_kind == 1) then
            call cal_res_vec(RES_VEC, X, B)             
            call terminal_cond_check(CONV_FL_VEC, RESID, RES_VEC, N)          
            
            if(CONV_FL .and. .not.CONV_FL_VEC) conv_criterion_val = conv_criterion_val/10
            
            !C terminal_kind==1 and reach convergence
            if(CONV_FL_VEC) then
               REACH_CONV_FL = .true.
               exit
            end if
         end if
         
      end do

      !C correct BNRM2 and resid when terminal_kind==1
      if(terminal_kind == 1) then
         val = BNRM2
         
         BNRM20 = 0.d0
         do i = 1, N
            abb = dabs(B(i))
            if(abb > BNRM20) BNRM20 = abb
         end do

         BNRM2 = BNRM20
         
         resid = resid*val/BNRM2
      end if

      
      !C check convergence by real residual only when terminal_kind==0
!!$      if(.not.CONV_FL .or. conv_check_fl) then
      if(terminal_kind == 0 .and. conv_check_fl) then
         call cal_res_vec(RES_VEC, X, B)
         call terminal_cond_check(CONV_FL, RESID, RES_VEC, N)
         if(.not.CONV_FL) then
            REACH_CONV_FL = .false.
            if(RESID > 1.0) then
               iter = -1
            end if
         else
            REACH_CONV_FL = .true.
         end if
      end if

      
      if(SETUP_FL) then
#ifdef PRINT_REZ
         write(*,*) "Setup time: ", (t2-t1)/dble(t_rate)
#endif
         call clear_multi_lev_matrix(LEVEL_NUM)
      end if

      DEALLOCATE (V,H,E,C,S,Q,Temp)
!C      if(terminal_kind==1) deallocate(RES_VEC)
      deallocate(RES_VEC)

    contains

      subroutine psolver(k)
        integer(kind=kint), intent(in) :: k
        integer(kind=kint) :: l,j
        do l = k, 1, -1
           do j = l+1, k
              E(l) = E(l) - H(l,j)*E(j)
           end do
           E(l) = E(l) / H(l,l)
        end do
        do i = 1, NP
           V(i,k+1) = 0.0
        end do

        do l = 1, k
           do i = 1, NP
              V(i,k+1) = V(i,k+1) + V(i,l)*E(l)
           end do
        end do

        do i = 1, NP
           Q(i)=0.0
        end do

        CALL amg_cycles(N, V(:,k+1), Q(:), LEVEL_NUM, Temp)
!!$      Q(:) = V(:,k+1)

        do i = 1, NP
           X(i)=X(i)+Q(i)
        end do
      end subroutine psolver
      

    END SUBROUTINE AMGGMRES


  END MODULE  solver_AMGGMRES
