!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif

MODULE solver_AMGRESCUT
CONTAINS
  !C SETUP_FL shows if this subroutine contains the setup phase
  !C and deletes the matrix
  SUBROUTINE AMGRESCUT    (LSIZ, N,  NPL, NPU,                        &
       &                  D,  AL, INL, IAL, AU, INU, IAU,             &
       &                  B,  X, resid, iter, SETUP_FL, LEVEL_NUM, REACH_CONV_FL)
    
    USE data_structure_for_AMG
    USE data_creation_stuben
    USE sgs_relaxation
    USE ilu_relaxation
    USE AMG_cycle
    USE terminal_condition, only:terminal_cond_vec_set, terminal_cond_val_set, &
         & terminal_cond_check, BNRM2, terminal_kind, conv_criterion_val
    USE residual_calculation
    USE amg_parameter, only: conv_check_fl
    
    IMPLICIT NONE
    LOGICAL, intent(in) :: SETUP_FL
    LOGICAL, intent(inout) :: REACH_CONV_FL
    INTEGER(kind=kint ), intent(inout) :: LEVEL_NUM
    INTEGER(kind=kint ), intent(in)    :: N, NPL, NPU
    REAL   (kind=kreal)                :: RESID, TOL
    INTEGER(kind=kint ), intent(inout) :: ITER
    INTEGER(kind=kint )                :: my_rank
    INTEGER(kind=kint ), intent(in)    :: LSIZ

    INTEGER(kind=kint ),  pointer ::  INU(:), IAU(:),INL(:),IAL(:)
    REAL   (kind=kreal),  pointer ::  D(:),AU(:), AL(:)
    REAL   (kind=kreal) ::  X(:), B(:)

    REAL   (kind=kreal),  ALLOCATABLE :: Temp(:),Residual_vec(:), psy(:,:), fai(:,:)
    REAL   (kind=kreal),  ALLOCATABLE :: alpha(:)
    REAL   (kind=kreal),  ALLOCATABLE :: AA(:,:),AA_bak(:,:), bb_bak(:)


    REAL   (kind=kreal) :: DNRM2,BNRM20,DNRM20, val
    INTEGER(kind=kint ) :: MAXIT, IFLAG=0


    INTEGER(kind=kint)  :: i,j, INOD, NOPT, zero_ind
    logical :: CONV_FL, ERR_FL


    !C-- INIT. 
    ALLOCATE (Residual_vec(N), Temp(N), alpha(0:LSIZ), psy(N, 0:LSIZ), fai(N, 0:LSIZ))
    ALLOCATE(AA(0:LSIZ, 0:LSIZ),AA_bak(0:LSIZ, 0:LSIZ) )

    MAXIT  = ITER
    
    if(SETUP_FL) then
       CALL data_creation(N, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, LEVEL_NUM)
    end if

    !C +-----------------------------------+
    !C | {Residual_vec}= ({b} - [A]{xini}) |
    !C +-----------------------------------+
    !C===
!!$    call cal_res(Residual_vec, BNRM2)
    call cal_res_vec(Residual_vec, X, B)
    
!!$    if(terminal_kind == 0) then
!!$       BNRM20 = 0.d0
!!$       do i= 1, N
!!$          BNRM20 = BNRM20 + B(i)**2
!!$       enddo
!!$       
!!$       BNRM2=BNRM20
!!$    else if(terminal_kind == 1) then
!!$       !C record infinity norm
!!$       BNRM20 = 0.d0
!!$       do i = 1, N
!!$          val = dabs(B(i))
!!$          if(val > BNRM20) BNRM20 = val
!!$       end do
!!$       
!!$       BNRM2 = BNRM20
!!$    end if
!!$    
!!$    if(BNRM2 == 0) BNRM2 = 1.0

    !C this value specify the 0 index  psy and fai vectors.
    zero_ind = 0
    !C-- BEGIN calculation
    do iter = 1, MAXIT
       NOPT = LSIZ
       if(iter <= LSIZ) NOPT = iter-1

       !C psy and fai have LSIZ+1 vectors. New one is stored at psy or fai(:,zero_ind)
       !C Third vector is at mod(zero_ind+2, LSIZ+1)
       !C psy(1:N, 0:LSIZ)
       !C +-----------------------------------+
       !C | {psy(:,0)} = M^-1 {Residual_vec}  |
       !C +-----------------------------------+
       psy(1:N, zero_ind) = 0
       call amg_cycles(N, Residual_vec, psy(:, zero_ind), LEVEL_NUM, Temp)
       
       !C +----------------------------+
       !C | {alpha(:)} is determined   |
       !C +----------------------------+
       call setalpha()


       !C +-----------------------------------------------+
       !C | {psy(:,0) = sum_{i=0,NOPT} alpha(i)*psy(:,i)} |
       !C +-----------------------------------------------+
       !C change psy(:,zero_ind)
       psy(1:N,zero_ind) = alpha(0)*psy(1:N,zero_ind)
       do j = 1, NOPT
          do i = 1, N
             psy(i,zero_ind) = psy(i,zero_ind)+alpha(j)*psy(i,mod(zero_ind+j, LSIZ+1))
          end do
       end do

       !C +------------------------+
       !C | {X} = {X} + {psy(:,0)} |
       !C +------------------------+
       X(1:N) = X(1:N) + psy(1:N,zero_ind)
       

       !C +---------------------------+
       !C | {psy(:,i+1)} = {psy(:,i)} |
       !C +---------------------------+
       !C shift psy and fai vector
       if(zero_ind == 0) then
          zero_ind = LSIZ
       else if(zero_ind >0) then
          zero_ind = zero_ind-1
       else
          stop "error zero_ind < 0: solver_AMGRESCUT"
       end if


       !C +-----------------------------+
       !C | {Residual_vec} = {B}-{A}{x} |
       !C +-----------------------------+
!!$       call cal_res(Residual_vec, DNRM2)
       call cal_res_vec(Residual_vec, X, B)
       call terminal_cond_check(CONV_FL, RESID, Residual_vec(:), N)

#ifdef PRINT_REZ
       if (terminal_kind==0) write (*, 1000) ITER, RESID
1000   format (i5, 1pe16.6)
#endif

       if ( CONV_FL   ) then
          REACH_CONV_FL = .true.
          exit
       end if
       if(terminal_kind==0 .and.  .not. RESID < 1.0E5 ) then
          write(*,*) "Divergence!"
          exit
       end if

!!$       if(isnan(RESID)) then
!!$          write(*,*) "Divergence!"
!!$          exit
!!$       end if
       
    end do

    !C check convergence only for divergence or not.
!!$    if(.not.CONV_FL ) then
    if(conv_check_fl .and. .not.CONV_FL) then
!!$       call cal_relative_res_norm(RESID, Residual_vec, X, B)
       if(RESID > 1.0) then
          iter = -1
       end if
    end if

    
    if(SETUP_FL) then
       call clear_multi_lev_matrix(LEVEL_NUM)
    end if
    
    deallocate (Residual_vec, Temp, alpha, psy, fai, AA, AA_bak)
    
  contains
    subroutine setalpha()
        use vec_sub
        integer(kind=kint), ALLOCATABLE   :: indx(:)
        integer(kind=kint) :: i,j,k
        real(kind=kreal) :: EPS 
        
        alpha(0:LSIZ) = 0.

        !C fai(:,0)  is calculated as A psy(:,0)
        call matvec_fai(zero_ind)
        !C fai(:,1)  is calculated as A psy(:,1)
        call matvec_fai(mod(zero_ind+1, LSIZ+1))

        do j = 0, NOPT
           k = mod(zero_ind+j, LSIZ+1)
           call my_dot_product(Residual_vec, fai(:,k), N, alpha(j))
        end do

        !C shift AA_bak
        do j = NOPT, 2, -1
           do i = NOPT, 2, -1
              AA_bak(i,j) = AA_bak(i-1,j-1)
           end do
        end do

        !C calculate 0th row and column of AA_bak.
        do i = 0, NOPT
           k = mod(zero_ind+i, LSIZ+1)
           call my_dot_product(fai(:,zero_ind), fai(:,k), N, AA_bak(0,i))
           AA_bak(i,0) = AA_bak(0,i)
        end do

        !C calculate 1th row and column of AA_bak.
        do i = 1, NOPT
           k = mod(zero_ind+i, LSIZ+1)
           j = mod(zero_ind+1, LSIZ+1)
           call my_dot_product(fai(:,j), fai(:,k), N, AA_bak(1,i))
           AA_bak(i,1) = AA_bak(1,i)
        end do


        !C copy AA_bak to AA
        do i = 0, NOPT
           do j = 0, NOPT
              AA(j,i) = AA_bak(j,i)
           end do
        end do

        allocate(indx(NOPT+1))
        EPS = 0.0
        call croutlu(AA(0:NOPT,0:NOPT), NOPT+1, indx, ERR_FL, EPS)
        if(.not.ERR_FL .and. NOPT >= 1) then
           call lusub(AA(0:NOPT,0:NOPT), NOPT+1, indx, alpha(0:NOPT))
        end if
        deallocate(indx)

        if(ERR_FL .or. NOPT < 1) then
!           write(*,*) "Reset alpha"
           alpha(0) = 1.0
           alpha(1:NOPT) = 0.
        end if
        
      end subroutine setalpha
      
      subroutine cal_res(res_vec, res_norm)
        REAL(kind=kreal),intent(out) :: res_vec(:), res_norm
        REAL(kind=kreal) :: WVAL
        INTEGER(kind=kint) :: isU,ieU,isL,ieL
        res_norm = 0.0
        do j= 1, N
           WVAL= B(j) - D(j) * X(j)
           isU= INU(j-1) + 1
           ieU= INU(j  ) 
           do i= isU, ieU
              inod= IAU(i)+ZERO_ORIGIN
              WVAL= WVAL - AU(i) * X(inod)
           enddo

           isL= INL(j-1) + 1
           ieL= INL(j  ) 
           do i= isL, ieL
              inod= IAL(i)+ZERO_ORIGIN
              WVAL= WVAL - AL(i) * X(inod)
           enddo
           !C res(1:N)
           res_vec(j)= WVAL
           res_norm = res_norm+WVAL*WVAL
        enddo
        res_norm = dsqrt(res_norm)
      end subroutine cal_res

      subroutine matvec_fai(k)
        integer(kind=kint), intent(in) :: k
        INTEGER(kind=kint) :: i,j,isU,ieU,isL,ieL, inod
        real(kind=kreal) :: WVAL

        do j= 1, N
           WVAL= D(j) * psy(j, k)
           isU= INU(j-1) + 1
           ieU= INU(j  ) 
           do i= isU, ieU
              inod= IAU(i)
              WVAL= WVAL + AU(i) * psy(inod, k)
           enddo

           isL= INL(j-1) + 1
           ieL= INL(j  ) 
           do i= isL, ieL
              inod= IAL(i)
              WVAL= WVAL + AL(i) * psy(inod, k)
           enddo
           fai(j, k) = WVAL
        end do
      end subroutine matvec_fai


  END SUBROUTINE AMGRESCUT
  
END MODULE  solver_AMGRESCUT
