!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

MODULE solver_GPU
CONTAINS
  
  !C SETUP_FL shows if this subroutine contains the setup phase
  !C and deletes the matrix
  SUBROUTINE solver_on_GPU (N,  NPL, NPU,                             &
       &                  D,  AL, INL, IAL, AU, INU, IAU,             &
       &                  B,  X, RESID, iter, SETUP_FL, LEVEL_NUM,    &
       &                  REACH_CONV_FL)
    
    USE data_structure_for_AMG
    USE data_creation_stuben
    USE terminal_condition, only: terminal_kind, CONV_CRITERION_VAL, CONV_CRITERION_VEC
    USE amg_parameter, only: conv_check_fl
    USE residual_calculation, only: cal_relative_res_norm
    !C CONV_CRITERION_VAL: criterion value for 2norm of relative residual with rhs vector
    !C CONV_CRITERION_VEC: criterion vector for residual
     
    IMPLICIT NONE
    LOGICAL, intent(in) :: SETUP_FL
    LOGICAL, intent(inout) :: REACH_CONV_FL
    INTEGER(kind=kint ), intent(inout) :: LEVEL_NUM
    INTEGER(kind=kint ), intent(in)    :: N, NPL, NPU
    REAL   (kind=kreal)                :: RESID, TOL
    INTEGER(kind=kint ), intent(inout) :: ITER

    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INU
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAU
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INL
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAL

    REAL   (kind=kreal), DIMENSION(: ),  pointer ::  D
    REAL   (kind=kreal), DIMENSION(: )  ::  B
    REAL   (kind=kreal), DIMENSION(: )  ::  X

    REAL   (kind=kreal), DIMENSION(:), pointer ::  AU
    REAL   (kind=kreal), DIMENSION(:), pointer ::  AL

    INTEGER(kind=kint)  :: SELECT_PARA = 1
    REAL   (kind=kreal) :: OMEGA  = 0.70

    INTEGER(kind=kint)  :: ROWSIZE

    INTEGER(kind=kint)  :: R_NPLU, P_NPLU, R_NP, P_NP

    INTEGER(kind=kint)  :: INT_DUM(1)
    REAL   (kind=kreal) :: REAL_DUM(1)

    REAL   (kind=kreal), DIMENSION(:), allocatable :: Residual_vec
    INTEGER(kind=kint) :: i    
    
    REACH_CONV_FL = .false.
    
!C    conv_criterion_val   = RESID           
    
    IF(SETUP_FL) THEN
       CALL data_creation(N, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, LEVEL_NUM)

       ROWSIZE = HIERARCHICAL_DATA(1) % N
!!$       CALL notify_level(LEVEL_NUM)

!!$       call parameter_set_on_GPU(SELECT_PARA, OMEGA)

       call mat_and_solver_set_on_GPU(LEVEL_NUM, SELECT_PARA, omega)

    END IF

!C  solverAMG_S.F90 170, 171 set call
    if(terminal_kind == 0) then
       call amgs_terminal_val_set_on_gpu(CONV_CRITERION_VAL);       
    else if(terminal_kind == 1) then
       ROWSIZE = HIERARCHICAL_DATA(1) % N
       call amgs_terminal_vec_set_on_gpu(CONV_CRITERION_VEC, ROWSIZE);
    else
       stop "No terminal condition is set in solver_GPU()"
    end if


!C  solverAMG_S.F90 39 set maxiter
    call solve_on_GPU(X, B, iter)
    
    if(conv_check_fl) then
       allocate(Residual_vec(N))
       call cal_relative_res_norm(RESID, Residual_vec, X, B)
       if(RESID > 1.0) then
          iter = -1
       end if
    end if

    if(SETUP_FL) then
       call clear_mat_on_GPU()
    end if

  END SUBROUTINE SOLVER_ON_GPU

  
  subroutine mat_and_solver_set_on_GPU(LEVEL_NUM, sol, omega)
    USE data_structure_for_AMG, only: HIERARCHICAL_DATA
    USE amg_parameter, only: iteration_number_down, iteration_number_bottom, gs_kind
    implicit none
    include 'precision.inc'

    REAL(kind=kreal), intent(in) :: omega
    !C AMG:sol == 0 AMGBICGSTAB:sol == 1
    INTEGER(kind=kint), intent(in) :: LEVEL_NUM, sol
    INTEGER(kind=kint) :: i
    INTEGER(kind=kint) :: R_N, P_N, R_NP, P_NP, R_NPLU, P_NPLU
    INTEGER(kind=kint) :: iter_down, iter_bottom, gs_at_lev1
    INTEGER(kind=kint)  :: INT_DUM(1)
    REAL   (kind=kreal) :: REAL_DUM(1)


    CALL notify_level(LEVEL_NUM)

    iter_down   = iteration_number_down
    iter_bottom = iteration_number_bottom
    if(gs_kind == 0) then
       !C GS smoother at lev1
       gs_at_lev1  = 1 
    else
       !C jacobi smoother at all levs
       gs_at_lev1  = 0
    end if

    call parameter_set_on_GPU(sol, OMEGA, iter_down, iter_bottom, gs_at_lev1)

    i = 1
    if(HIERARCHICAL_DATA(i) % N == 0) then
       !C when 0 size problem matrix comes, then dummy data is given to the sub.
       !C if 0 size problem mat at coarser level is ok, because the arrays are
       !C allocated in setup_phase.
       CALL store_mat_on_GPU(i,HIERARCHICAL_DATA(i) % N, &
            & HIERARCHICAL_DATA(i) % NP, &
            & HIERARCHICAL_DATA(i) % NPL,&
            & HIERARCHICAL_DATA(i) % NPU,&
            & INT_DUM,  &
            & INT_DUM,  &
            & INT_DUM,  &
            & REAL_DUM, &
            & INT_DUM,  &
            & REAL_DUM, &
            & REAL_DUM, &
            & INT_DUM, INT_DUM, INT_DUM, INT_DUM, INT_DUM, INT_DUM, &               
            & INT_DUM, INT_DUM, REAL_DUM, &
            & INT_DUM, INT_DUM, REAL_DUM)
    else
       CALL store_mat_on_GPU(i,HIERARCHICAL_DATA(i) % N, &
            & HIERARCHICAL_DATA(i) % NP, &
            & HIERARCHICAL_DATA(i) % NPL,&
            & HIERARCHICAL_DATA(i) % NPU,&
            & HIERARCHICAL_DATA(i) % INL,&
            & HIERARCHICAL_DATA(i) % INU,&
            & HIERARCHICAL_DATA(i) % IAL,&
            & HIERARCHICAL_DATA(i) % AL,&
            & HIERARCHICAL_DATA(i) % IAU,&
            & HIERARCHICAL_DATA(i) % AU, &
            & HIERARCHICAL_DATA(i) % D,&
            & INT_DUM, INT_DUM, INT_DUM, INT_DUM, INT_DUM, INT_DUM, &               
            & INT_DUM, INT_DUM, REAL_DUM, &
            & INT_DUM, INT_DUM, REAL_DUM)
    end if
    DO i = 2, LEVEL_NUM
       R_N       = HIERARCHICAL_DATA(i)   % N
       P_N       = HIERARCHICAL_DATA(i-1) % N
       R_NP      = HIERARCHICAL_DATA(i) % R % ROW_SIZE
       P_NP      = HIERARCHICAL_DATA(i) % P % ROW_SIZE
       R_NPLU    = HIERARCHICAL_DATA(i) % R % IN(R_NP)
       P_NPLU    = HIERARCHICAL_DATA(i) % P % IN(P_NP)

       CALL store_mat_on_GPU(i,HIERARCHICAL_DATA(i) % N, &
            & HIERARCHICAL_DATA(i) % NP, &
            & HIERARCHICAL_DATA(i) % NPL,&
            & HIERARCHICAL_DATA(i) % NPU,&
            & HIERARCHICAL_DATA(i) % INL,&
            & HIERARCHICAL_DATA(i) % INU,&
            & HIERARCHICAL_DATA(i) % IAL,&
            & HIERARCHICAL_DATA(i) % AL, &
            & HIERARCHICAL_DATA(i) % IAU,&
            & HIERARCHICAL_DATA(i) % AU, &
            & HIERARCHICAL_DATA(i) % D,  &
            & R_N, P_N, R_NP, P_NP, R_NPLU, P_NPLU, &               
            & HIERARCHICAL_DATA(i) % R % IN,&
            & HIERARCHICAL_DATA(i) % R % CN,&
            & HIERARCHICAL_DATA(i) % R % V, &
            & HIERARCHICAL_DATA(i) % P % IN,&
            & HIERARCHICAL_DATA(i) % P % CN,&
            & HIERARCHICAL_DATA(i) % P % V)
    END DO

  end subroutine mat_and_solver_set_on_GPU


END MODULE  solver_GPU
