!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif

MODULE solver_IDR_AMG
CONTAINS
  !C SETUP_FL shows if this subroutine contains the setup phase
  !C and deletes the matrix
  SUBROUTINE IDR_AMG    (N,  NPL, NPU,                                &
       &                  D,  AL, INL, IAL, AU, INU, IAU,             &
       &                  B,  X, resid,iter, SETUP_FL, LEVEL_NUM)
    
    USE data_structure_for_AMG
    USE data_creation_stuben
    USE sgs_relaxation
    USE ilu_relaxation
    USE AMG_cycle
    USE terminal_condition, only:terminal_cond_vec_set, terminal_cond_val_set, &
         & terminal_cond_check, BNRM2, terminal_kind, conv_criterion_val
    
    IMPLICIT NONE
    LOGICAL, intent(in) :: SETUP_FL
    INTEGER(kind=kint ), intent(inout) :: LEVEL_NUM
    INTEGER(kind=kint ), intent(in)    :: N, NPL, NPU
    REAL   (kind=kreal)                :: RESID, TOL
    INTEGER(kind=kint ), intent(inout) :: ITER
    INTEGER(kind=kint )                :: my_rank

    INTEGER(kind=kint )              :: NPROCS,SOLVER_COMM
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INU
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAU
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INL
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAL

    REAL   (kind=kreal), DIMENSION(: ),  pointer ::  D
    REAL   (kind=kreal), DIMENSION(: )  ::  B
!!$    REAL   (kind=kreal), DIMENSION(NP )  ::  X
    REAL   (kind=kreal), DIMENSION(: )  ::  X

    !C pointers for ILU
    REAL   (kind=kreal), DIMENSION(:), pointer ::  ILU_D, ILU_AL, ILU_AU

    REAL   (kind=kreal), DIMENSION(:), pointer ::  AU
    REAL   (kind=kreal), DIMENSION(:), pointer ::  AL
    REAL   (kind=kreal), DIMENSION(:), ALLOCATABLE :: Temp, R,dx, P
    REAL   (kind=kreal), DIMENSION(:), ALLOCATABLE :: dr,zero_vec, Residual_vec
    REAL   (kind=kreal) :: WVAL,RHO0,RHO,BETA,RHO1,C10,C1,ALPHA, gamma


    REAL   (kind=kreal) :: DNRM2,BNRM20,DNRM20, v1, v2
    REAL   (kind=kreal) :: STARTTIME,ENDTIME,RTIME,STARTTIME2,ENDTIME2,RTIME2
    INTEGER(kind=kint ) :: MAXIT, IFLAG=0


    INTEGER(kind=kint)  :: cnt,NEIBPETOT,I,J,ISU,IEU,ISL,IEL,WSIZE
    INTEGER(kind=kint)  :: INOD
    
    logical :: CONV_FL

    INTEGER(kind=kint) :: t1, t2, t_rate, t_max

    !C-- INIT. 
    NPROCS=0
    my_rank=0

    ALLOCATE (R(N), Temp(N), dx(N),dr(N), zero_vec(N), P(N), Residual_vec(N))
    zero_vec = 0
    dx = 0 
    call random_number(P)
!!$    P = 1.0

    MAXIT  = ITER
!C    conv_criterion_val   = RESID           

    call system_clock(t1, t_rate, t_max)
    if(SETUP_FL) then
       CALL data_creation(N, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, LEVEL_NUM)
    end if

!!$    !C arrays are allocated for ILU0
!!$    allocate(ILU_D(NP), ILU_AL(NPL), ILU_AU(NPU))
!!$    ILU_D(:) = D(:)
!!$    ILU_AL(:) = AL(:)
!!$    ILU_AU(:) = AU(:)
!!$    call ilu0_setup(NP, ILU_D, ILU_AL, INL, IAL, ILU_AU, INU, IAU)
    
    call system_clock(t2, t_rate, t_max)

    !C +-------------------------------+
    !C | {r0}= [M^-1]({b} - [A]{xini}) |
    !C +-------------------------------+
    !C===
    !C-- BEGIN calculation
    do j= 1, N
       WVAL= B(j) - D(j) * X(j)
       isU= INU(j-1) + 1
       ieU= INU(j  ) 

       do i= isU, ieU
          inod= IAU(i)+ZERO_ORIGIN
          WVAL= WVAL - AU(i) * X(inod)
       enddo

       isL= INL(j-1) + 1
       ieL= INL(j  ) 
       do i= isL, ieL
          inod= IAL(i)+ZERO_ORIGIN
          WVAL= WVAL - AL(i) * X(inod)
       enddo
       dr(j)= WVAL
    enddo
    R = 0
    CALL amg_cycles(N, dr, R, LEVEL_NUM, Temp)    
    

    !C norm of right hand vector is calculated
    BNRM20 = 0.d0
    do i= 1, N
       BNRM20 = BNRM20 + B(i)**2
    enddo
    BNRM2 = BNRM20


    if (BNRM2.eq.0.d0) BNRM2= 1.d0
    ITER = 0
    
    gamma = 0
    !C===
    do iter= 1, MAXIT
       do i = 1, N
          dx(i) = R(i)+gamma*(dx(i)+dr(i))
          X(i) = X(i) + dx(i)
       end do
       
       !C Residual_vec is used as temporary vector
       Residual_vec =  -R
       
       do i = 1, N
          R(i) = R(i) + gamma * dr(i)
       end do
       
       !C right hand side vector set as 0
       CALL amg_cycles(N, zero_vec, R, LEVEL_NUM, Temp)

       dr = Residual_vec + R

       !C gamma = -inner_product(p,R)/inner_product(p, dr)
       v1 = 0; v2 = 0
       do i = 1, N
          v1 = v1 + p(i) * R(i)
          v2 = v2 + p(i) * dr(i)
       end do
       gamma = -v1/v2
       
       !C -- this part calculates the real residual norm 
       DNRM20 = 0
       do j= 1, N
          WVAL= B(j) - D(j) * X(j)
          
          isU= INU(j-1) + 1
          ieU= INU(j  ) 
          do i= isU, ieU
             inod= IAU(i)+ZERO_ORIGIN
             WVAL= WVAL - AU(i) * X(inod)
          enddo

          isL= INL(j-1) + 1
          ieL= INL(j  ) 
          do i= isL, ieL
             inod= IAL(i)+ZERO_ORIGIN
             WVAL= WVAL - AL(i) * X(inod)
          enddo

          !C residual is recorded for ILU relaxation
          Residual_vec(j) = WVAL
          DNRM20 = DNRM20+WVAL*WVAL
       enddo

!!$       RESID = dsqrt(DNRM20/BNRM20)

!C --       
       DNRM2=DNRM20
       RESID= dsqrt(DNRM2/BNRM2)

       call terminal_cond_check(CONV_FL, RESID, Residual_vec(:), N)

#ifdef PRINT_REZ
       if (my_rank==0.and. terminal_kind==0) write (*, 1000) ITER, RESID
1000   format (i5, 1pe16.6)
#endif

!!$       if ( RESID.le.TOL   ) then
       if ( CONV_FL   ) then
!C          call single_matrix_counting(LEVEL_NUM)
          
#ifdef PRINT_REZ          
          write(*,*) "Setup time: ", (t2-t1)/dble(t_rate)
#endif
          exit
       end if
       if(terminal_kind==0 .and.  .not. RESID < 1.0E5 ) then
          write(*,*) "Divergence!"
          exit
       end if


    end do

    !C===
    if(SETUP_FL) then
       call clear_multi_lev_matrix(LEVEL_NUM)
    end if
    deALLOCATE (R, Temp, dx, dr, zero_vec, Residual_vec)

  END SUBROUTINE IDR_AMG


END MODULE  solver_IDR_AMG
