!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif
  
  MODULE solver_IDR_s_AMG
  CONTAINS
    !C SETUP_FL shows if this subroutine contains the setup phase
    !C and deletes the matrix
    SUBROUTINE IDR_s_AMG (s, N,  NPL, NPU,                              &
         &                  D,  AL, INL, IAL, AU, INU, IAU,             &
         &                  B,  X, resid, iter, SETUP_FL, LEVEL_NUM,REACH_CONV_FL)

      USE data_structure_for_AMG
      USE data_creation_stuben
      USE sgs_relaxation
      USE ilu_relaxation
      USE AMG_cycle
      USE vec_sub
      USE terminal_condition, only:terminal_cond_vec_set, terminal_cond_val_set, &
           & terminal_cond_check, BNRM2, terminal_kind, conv_criterion_val
      USE residual_calculation
      USE amg_parameter, only: conv_check_fl


      IMPLICIT NONE
      LOGICAL, intent(in) :: SETUP_FL
      LOGICAL, intent(inout) :: REACH_CONV_FL
      INTEGER(kind=kint ), intent(inout) :: LEVEL_NUM
      INTEGER(kind=kint ), intent(in)    :: s
      INTEGER(kind=kint ), intent(in)    :: N, NPL, NPU
      REAL   (kind=kreal)                :: RESID, TOL
      INTEGER(kind=kint ), intent(inout) :: ITER
      INTEGER(kind=kint )                :: my_rank

      INTEGER(kind=kint )              :: NPROCS,SOLVER_COMM
      INTEGER(kind=kint ), DIMENSION(:), pointer ::  INU
      INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAU
      INTEGER(kind=kint ), DIMENSION(:), pointer ::  INL
      INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAL

      REAL   (kind=kreal), DIMENSION(: ),  pointer ::  D
      REAL   (kind=kreal), DIMENSION(: )  ::  B
!!$    REAL   (kind=kreal), DIMENSION(NP )  ::  X
      REAL   (kind=kreal), DIMENSION(: )  ::  X

      !C pointers for ILU
      REAL   (kind=kreal), DIMENSION(:), pointer ::  ILU_D, ILU_AL, ILU_AU

      REAL   (kind=kreal), DIMENSION(:), pointer ::  AU
      REAL   (kind=kreal), DIMENSION(:), pointer ::  AL
      REAL   (kind=kreal), DIMENSION(:), ALLOCATABLE :: Temp, R, F, C
      REAL   (kind=kreal), DIMENSION(:), ALLOCATABLE :: zero_vec, Residual_vec
      REAL   (kind=kreal) :: WVAL,RHO0,RHO,BETA,RHO1,C10,C1,ALPHA, gamma

      REAL(kind=kreal), DIMENSION(:), pointer :: dx, dr, P
      REAL(kind=kreal), DIMENSION(:,:), ALLOCATABLE :: M
      REAL(kind=kreal), DIMENSION(:,:), ALLOCATABLE,TARGET :: Qs, Es, Ps


      REAL   (kind=kreal) :: DNRM2, v1, v2, BNRM20, val
      REAL   (kind=kreal) :: STARTTIME,ENDTIME,RTIME,STARTTIME2,ENDTIME2,RTIME2
      INTEGER(kind=kint ) :: MAXIT, IFLAG=0


      INTEGER(kind=kint)  :: cnt,NEIBPETOT,i,j,k,ISU,IEU,ISL,IEL,WSIZE
      INTEGER(kind=kint)  :: INOD

      logical :: CONV_FL 

      INTEGER(kind=kint) :: t1, t2, t_rate, t_max, sum_of_iter
      logical :: gauss_FL, reset_FL


      !C-- INIT. 
      NPROCS=0
      my_rank=0
      sum_of_iter = 0

      MAXIT  = ITER
      !C    conv_criterion_val   = RESID           
      REACH_CONV_FL =.false.


      !C pointer dx(:), dr(:), P(:)
      allocate(Qs(N,s),Es(N,s),Ps(N,s),M(s,s), F(s), C(s))
      !C dx => Qs(:,1); dr => Es(:,1)
      ALLOCATE (R(N), Temp(N), zero_vec(N), Residual_vec(N))
      zero_vec = 0

      !C set random numbers in Ps(:,:) 
      call random_seed

      DO i = 1, s
         DO j = 1, N
            call random_number(Ps(j,i))
         END DO
      END DO
      call Gram_Schmidt(Ps, s, N)
      P => Ps(:,1)
!!$    P = 1.0

      call system_clock(t1, t_rate, t_max)
      if(SETUP_FL) then
         CALL data_creation(N, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, LEVEL_NUM)
      end if
      
      call system_clock(t2, t_rate, t_max)

      !C norm of r0 is calculated
      if(terminal_kind == 0) then
         BNRM20 = 0.d0
         do i= 1, N
            BNRM20 = BNRM20 + B(i)**2
         enddo

         BNRM2=BNRM20
      else if(terminal_kind == 1) then
         !C record infinity norm
         BNRM20 = 0.d0
         do i = 1, N
            val = dabs(B(i))
            if(val > BNRM20) BNRM20 = val
         end do

         BNRM2 = BNRM20
      end if
    
      if (BNRM2.eq.0.d0) BNRM2= 1.d0

      !C Loops for reseting iteration history 
      do

         reset_FL = .false.
         !C +-------------------------------+
         !C | {r0}= [M^-1]({b} - [A]{xini}) |
         !C +-------------------------------+
         !C===
         !C Residual_vec is used as temporary vector
         !C-- BEGIN calculation
         call cal_res_vec(Residual_vec, X, B)
         R = 0
         CALL amg_cycles(N, Residual_vec, R, LEVEL_NUM, Temp) !C M^-1 Residual_vec

         dx => Qs(:,1); dr => Es(:,1)
         dx = R
         X = X+dx
         Residual_vec = -R !C Residual_vec is used as temporary vec.

         CALL amg_cycles(N, zero_vec, R, LEVEL_NUM, Temp) !C M^-1 N

         dr = Residual_vec+R
         gamma = -dot_product(P, R) / dot_product(P, dr)

         do iter= 2, s

            dx => Qs(:,iter);  dr => Es(:,iter)
            do i = 1, N
               dx(i) = R(i) + gamma * (Qs(i,iter-1)+Es(i,iter-1))
               X(i)  = X(i) + dx(i)
            end do

            !C Residual_vec is used as temporary vector
            Residual_vec =  -R

            do i = 1, N
               R(i) = R(i) + gamma * Es(i, iter-1)
            end do

            !C right hand side vector set as 0
            CALL amg_cycles(N, zero_vec, R, LEVEL_NUM, Temp)

            dr = Residual_vec + R
            gamma = -dot_product(P,R)/dot_product(P, dr)

            !C -- this part calculates the real residual norm 
            call cal_res_vec(Residual_vec, X, B)
!!$            DNRM2 = dot_product(Residual_vec, Residual_vec)
!!$
!!$            RESID = dsqrt(DNRM2/BNRM2)
            
            call terminal_cond_check(CONV_FL, RESID, Residual_vec(:), N)

#ifdef PRINT_REZ
            if (my_rank==0 ) write (*, 1000) ITER+sum_of_iter, RESID
1000        format (i5, 1pe16.6, a3)
#endif

            if ( CONV_FL   ) then
               !C          call single_matrix_counting(LEVEL_NUM)
               exit
            end if
            if(terminal_kind==0 .and.  .not. RESID < 1.0E5 ) then
               write(*,*) "Divergence!"
               exit
            end if
!!$            if(isnan(RESID)) then
!!$               exit
!!$            end if

         end do

         do i = 1, s
            do j = 1, s
               M(i,j) = dot_product(Ps(:,i),Es(:,j))
            end do
         end do

         do i = 1, s 
            f(i) = dot_product(Ps(:,i),R)
         end do

         if(.not.CONV_FL) then
            k = 1
            do iter = s+1, MAXIT

               dx => Qs(:,k); dr => Es(:,k)
               !C   solve c from MC = F

               call gauss_jordan(M,C,F,s,gauss_FL)
               if(.not.gauss_FL) then
                  reset_FL = .true.
                  exit
               end if
               
               !C   dx = (Es+Qs)c+R
               do i = 1, N
                  wval =  R(i)
                  do j = 1, s
                     wval = wval - (Es(i,j)+Qs(i,j))*C(j)
                  end do
                  dx(i) = wval
               end do

               X = X+dx

               !C check terminal condition
               !C------------
               call cal_res_vec(Residual_vec, X, B)
!!$               DNRM2 = dot_product(Residual_vec, Residual_vec)
!!$               RESID = dsqrt(DNRM2/BNRM2)
               call terminal_cond_check(CONV_FL, RESID, Residual_vec(:), N)

#ifdef PRINT_REZ
               if (my_rank==0) write(*,1000) ITER+sum_of_iter, RESID
#endif

               if ( CONV_FL ) then
                  !C          call single_matrix_counting(LEVEL_NUM)
                  !C          write(*,*) "Setup time: ", (t2-t1)/dble(t_rate)
                  exit
               end if
               if(terminal_kind==0 .and.   RESID > 1.0E5 ) then
                  write(*,*) "Divergence!"
                  exit
               end if
               !C------------

               Residual_vec = -R

               !C R = R+E c 
               do i = 1, N
                  wval = R(i)
                  do j = 1, s
                     wval = wval - Es(i,j)*C(j)
                  end do
                  R(i) = wval
               end do

               call amg_cycles(N, zero_vec, R, LEVEL_NUM, Temp)
               dr = Residual_vec + R
               do i = 1, s
                  M(i,k) = dot_product(Ps(:,i), dr)
               end do

               f = f + M(:,k)

               k = k+1
               if(k > s) k = 1

            end do
         end if

         sum_of_iter = sum_of_iter + iter
         MAXIT = MAXIT - iter
         iter = 0
         if(CONV_FL) REACH_CONV_FL = .true.
         if(reset_FL .or. MAXIT < 0 .or. CONV_FL) exit
      end do
      iter = iter + sum_of_iter
      


#ifdef PRINT_REZ
      if(SETUP_FL) then
         write(*,*) "Setup time: ", (t2-t1)/dble(t_rate)
      end if
#endif               

      !C check convergence by real residual
!!$    if(.not.CONV_FL .or. conv_check_fl) then
      if(conv_check_fl) then
!!$         call cal_relative_res_norm(RESID, R, X, B)
!!$         call terminal_cond_check(CONV_FL, v1, R, N)
         if(.not.CONV_FL) then
            if(RESID > 1.0) then
               iter = -1
            end if
         end if
      end if
      
      !C===
      if(SETUP_FL) then
         call clear_multi_lev_matrix(LEVEL_NUM)
      end if

      deallocate(Qs,Es,Ps,M, F, C, R, Temp, zero_vec, Residual_vec)
      
    END SUBROUTINE IDR_S_AMG
    
  END MODULE  solver_IDR_s_AMG
