!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

MODULE solver_IDR_s_gs
contains
  subroutine Idr_s_GS(sv,N,NPL,NPU,D,AL,INL,IAL,AU,INU,IAU,B,X,resid,iter)
    USE vec_sub
    USE terminal_condition, only:terminal_cond_vec_set, terminal_cond_val_set, &
         & terminal_cond_check, BNRM2, terminal_kind, conv_criterion_val

    implicit none
    include 'precision.inc'

    integer(kind=KINT) , intent(in) :: N,NPU,NPL
    integer(kind=KINT) , intent(in) :: sv
    real(kind=KREAL) , dimension(NPL) , intent(in) :: AL
    real(kind=KREAL) , dimension(N) , intent(in) :: D
    real(kind=KREAL) , dimension(NPU) , intent(in) :: AU

    real(kind=KREAL) , dimension(N) , intent(in) :: B
    real(kind=KREAL) , dimension(N) , intent(inout) :: X
    
    REAL   (kind=KREAL), intent(inout)  :: RESID
    INTEGER(kind=kINT), intent(inout)  :: iter
    
    integer(kind=KINT) , dimension(0:N) , intent(in) :: INL
    integer(kind=KINT) , dimension(0:N) , intent(in) :: INU

    integer(kind=KINT) , dimension(NPL) , intent(in) :: IAL
    integer(kind=KINT) , dimension(NPU) , intent(in) :: IAU

    integer(kind=KINT) i,j,k,loop

    real(kind=KREAL) , allocatable :: r(:),dr(:),S(:),dx(:),r0(:),V(:)
    real(kind=KREAL) , allocatable :: E(:,:),Q(:,:),Pt(:,:),P(:,:)
    real(kind=KREAL) , dimension(:) , pointer :: M(:,:),f(:),c(:)
    real(kind=KREAL) h,norm1,norm2,err,val,dotp
    real(kind=KREAL) nai,seki
    integer(kind=KINT) :: isU,ieU,isL,ieL,inod
    integer(kind=KINT) :: loopmax
    real(kind=KREAL)   :: er0 = 1.0E-7, DNRM2
    integer(kind=KINT) :: f2 = 10, sum_of_iter
    logical :: gauss_FL,CONV_FL,reset_FL

    loopmax = iter
    sum_of_iter = 0
    
    allocate(r(N),dr(N),S(N),dx(N),r0(N),V(N))
    allocate(E(N,sv),Q(N,sv),Pt(sv,N),P(N,sv))
    allocate(M(sv,sv),f(sv),c(sv))

    r(:)  = 0.0d0
    dr(:) = 0.0d0
    S(:)  = 0.0d0
    dx(:) = 0.0d0
    V(:)  = 0.0d0
    h     = 0.0d0
    P(:,:)  = 0.0d0
    Pt(:,:) = 0.0d0
    M(:,:)  = 0.0d0
    c(:)  = 0.0d0
    f(:)  = 0.0d0
    E(:,:)  = 0.0d0
    Q(:,:)  = 0.0d0

    !	do i=1,N
    !		call system_clock(count=ti(i))
    !	end do

    call random_seed
    DO i=1,sv
       DO j = 1,N
          call random_number(P(j,i))
       END DO
    END DO

    !C initial residual
    DO i = 1,N
       val = 0.0d0
       val = B(i) - D(i)*X(i)
       isU = INU(i-1) +1
       ieU = INU(i)
       DO j = isU,ieU
          inod = IAU(j)
          val = val - AU(j)*X(inod)
       END DO
       isL = INL(i-1) +1
       ieL = INL(i)
       DO j =isL,ieL
          inod = IAL(j)
          val = val - AL(j)*X(inod)
       END DO

       r(i) = val
    END DO

    call Gram_Schmidt(P, sv, N)

    !	DO i = 2,sv
    !		dotp = dot_product(P(1:n,i),P(1:n,i-1))
    !		write(*,*) 'dotp =',dotp
    !	END DO


    !C norm of r0 is calculated
    BNRM2 = dot_product(r0, r0)
    if (BNRM2.eq.0.d0) BNRM2= 1.d0

    DO

       reset_FL = .false.
       
       DO loop = 1,sv

          V(:) = r - (h * dr(:))

          !		DO i = 1,N
          !			write(*,*) loop, V(i),r(i)
          !		END DO

          DO i =1,N
             val = 0.0d0
             isL = INL(i-1) +1
             ieL = INL(i)
             S(i) = V(i)
             DO j =isL,ieL
                inod = IAL(j)
                val = val + AL(j)*S(inod)
             END DO
             S(i) = S(i) - val
             S(i) = S(i) / D(i)
          END DO

          dx(:) = S - (h * dx(:))

          DO i =1,N
             val = 0.0d0
             isU = INU(i-1) +1
             ieU = INU(i)
             DO j =isU,ieU
                inod = IAU(j)
                val = val + AU(j)*S(inod)
             END DO
             dr(i) = -val - r(i)
          END DO

          r(:) = r + dr(:)
          X(:) = X + dx(:)

          E(:,loop) = dr(:)
          Q(:,loop) = dx(:)
          !C-------M=Pt*E-----C!
          DO i = 1,sv
             val = 0.0d0
             DO j = 1,N
                val = val + P(j,i)*dr(j)
             END DO
             M(i,loop) = val
          END DO

          
          !C check terminal condition
          !C------------
          DNRM2 = dot_product(r, r)
          RESID = dsqrt(DNRM2/BNRM2)
          call terminal_cond_check(CONV_FL, RESID, r, N)

#ifdef PRINT_REZ
          if (terminal_kind==0) write(*,1000) loop+sum_of_iter, RESID
1000      format (i5, 1pe16.6, a3)
#endif

          if ( CONV_FL ) exit
          !C------------

          !		do i=1,N
          !			write(*,*)'loop:',loop, 'dr = ',dr(i,loop)
          !		end do
          
          nai = dot_product(P(:,1),r)
          seki = dot_product(P(:,1),dr(:))
          h = (nai / seki)
       END DO
       !C-----M=P(ž)*E(drν)-----C!
       !	DO i = 1,sv
       !		DO j = 1,sv
       !			val = 0.0d0
       !			DO k = 1,N
       !				val = val + P(k,i)*dr(k,j)
       !			END DO
       !			M(j,i) = val
       !		END DO
       !	END DO

       !C-----f= P(ž)*r-----C!
       DO i = 1,sv
          val = 0.0d0
          DO j = 1,N
             val = val + P(j,i)*r(j)
          END DO
          f(i) = val
       END DO

       if(.not.CONV_FL) then
          k = 1
          !C-----main loop-----C!
          DO loop = sv+1, loopmax

             !C-----solve c from Mc=f-----C!
             call gauss_jordan(M,c,f,sv,gauss_FL)
             if(.not.gauss_FL) then
                reset_FL = .true.
                exit
             end if

             V(:) = 0.0d0	
             DO i = 1,N
                DO j = 1,sv
                   V(i) = V(i) + E(i,j)*c(j)
                END DO
             END DO
             V(:) = r - V 
             DO i =1,N
                val = 0.0d0
                isL = INL(i-1) +1
                ieL = INL(i)
                S(i) = V(i)
                DO j =isL,ieL
                   inod = IAL(j)
                   val = val + AL(j)*S(inod)
                END DO
                S(i) = S(i) - val
                S(i) = S(i) / D(i)
             END DO

             DO i = 1,N
                val = 0.0d0
                DO j = 1,sv
                   val = val + Q(i,j)*c(j)		
                END DO
                Q(i,k) = val
             END DO

             dx(:) = S - Q(:,k)  

             DO i =1,N
                val = 0.0d0
                isU = INU(i-1) +1
                ieU = INU(i)
                DO j =isU,ieU
                   inod = IAU(j)
                   val = val + AU(j)*S(inod)
                END DO
                dr(i) = -val - r(i)
             END DO

             r(:) = r + dr(:)
             X(:) = X + dx(:)

             E(:,k) = dr(:)
             Q(:,k) = dx(:)

             DO i = 1,sv
                val = 0.0d0
                DO j = 1,N
                   val = val + P(j,i)*dr(j)
                END DO
                M(i,k) = val
             END DO

             f(:) = f + M(:,k)
             k = k + 1

             if(k>sv) k = 1

             !C check terminal condition
             !C------------
             DNRM2 = dot_product(r, r)
             RESID = dsqrt(DNRM2/BNRM2)
             call terminal_cond_check(CONV_FL, RESID, r, N)

#ifdef PRINT_REZ
             if (terminal_kind==0) write(*,1000) loop+sum_of_iter, RESID
#endif

             if ( CONV_FL ) exit
             !C------------

          END DO
       end if
       sum_of_iter = sum_of_iter + loop
       loopmax = loopmax - loop
       loop = 0
       if(.not.reset_FL .or. loopmax < 0 .or. CONV_FL) exit
    end DO
    iter = loop + sum_of_iter
    
    
    deallocate(r,dr,S,dx,r0,V)
    deallocate(E,Q,Pt,P)
    deallocate(M,f,c)

  end subroutine Idr_s_GS

END MODULE solver_IDR_s_gs
