!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University. 
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.

module terminal_condition
  include 'precision.inc'
  integer(kind=kint)   :: terminal_kind = 0
  REAL(kreal), pointer :: conv_criterion_vec(:)
  REAL(kreal)          :: conv_criterion_val=1.0E-8
  REAL(kreal)          :: BNRM2 !C BNRM2 is the sum of the squared values of rhs vector elements.
  REAL(kreal)          :: conv_criterion_vec2 !C square of 2norm of conv_criterin_vec
  
  
contains

  
  subroutine BNRM2_set(B, N)
!C    !$ USE omp_lib
    implicit none
!C    include 'mpif.h'
    
    REAL(kind=kreal)   :: B(:), BNRM20, WVAL
    INTEGER(kind=kint) :: SOLVER_COMM, ierr, i, N

    if(terminal_kind == 0) then
       BNRM20 = 0.d0
!C       !$omp parallel do reduction(+:BNRM20)
       do i= 1, N
          BNRM20 = BNRM20 + B(i)**2
       enddo
!C       !$omp end parallel do
       
!C       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
!C            &                    MPI_SUM, SOLVER_COMM, ierr)
       BNRM2 = BNRM20
    else if(terminal_kind == 1) then
       BNRM20 = 0.d0
!C       !$omp parallel do reduction(max:BNRM20)
       do i= 1, N
          WVAL = dabs(B(i))
          if(BNRM20 < WVAL) BNRM20 = WVAL
       enddo
!C       !$omp end parallel do
!C       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
!C            &                    MPI_MAX, SOLVER_COMM, ierr)
       BNRM2 = BNRM20
    else 
       stop 'unknown terminal_kind'
    end if

    if(BNRM2 == 0) BNRM2 = 1.0
    
  end subroutine BNRM2_set


  
  SUBROUTINE terminal_cond_vec_set(res_vec, N)
    implicit none
    integer(kind=kint), intent(in) :: N
    integer(kind=kint), save :: vecsize
    real(kind=kreal), intent(in) :: res_vec(:)
    integer(kind=kint) :: i
    real(kind=kreal) :: v_min

    terminal_kind=1
    if(.not.ASSOCIATED(conv_criterion_vec)) then
       allocate(conv_criterion_vec(N))
       vecsize = N
    else if(vecsize < N) then
       write(*,*) "Convergence Criterion vector is reallocated"
       deallocate(conv_criterion_vec); allocate(conv_criterion_vec(N)); 
       vecsize = N
    end if
    
    conv_criterion_vec(1:N) = res_vec(1:N)
    
    !C set conv_criterion_val for terminal detection of residual norm
    v_min = dabs(res_vec(1))
    do i = 1, N
       if(v_min > dabs(res_vec(i))) v_min = dabs(res_vec(i))
    end do
    conv_criterion_vec2 = N*(v_min*v_min)
    
  END SUBROUTINE terminal_cond_vec_set

  SUBROUTINE terminal_cond_val_set(res_val)
    implicit none
    real(kind=kreal), intent(in) ::  res_val

    terminal_kind = 0
    conv_criterion_val = res_val

  END SUBROUTINE terminal_cond_val_set

  !C terminal_kind==1: Check each element of the residual_vec, and stores the  result
  !C                   on the CONVERGENCE_FL
  !C terminal_kind==0: Calculate the residual norm from the residual_vec, and check. 
  !C                   And store the relative residual norm on RESID 
  SUBROUTINE terminal_cond_check(CONVERGENCE_FL, resid, residual_vec, N)
    implicit none
!!$     include 'mpif.h'
    logical, intent(out) :: CONVERGENCE_FL
    REAL(kind=kreal), intent(out) :: RESID
    REAL(kind=kreal) :: residual_vec(:)
    integer(kind=kint), intent(in) :: N

    real(kind=kreal)     :: RNRM, RNRM2, v
    integer(kind=kint)   :: i, ierr
    logical :: conv_fl
    
    if(terminal_kind==1) then
       conv_fl = .true.
       RNRM = 0

       do i = 1, N
          v = dabs(residual_vec(i))
          conv_fl = conv_fl .and. (v <= conv_criterion_vec(i))
          
          if(RNRM < v) RNRM = v
          
       end do
       
       !C RESID has infinity norm
       RESID = RNRM / BNRM2
       
       CONVERGENCE_FL = conv_fl

       !        call MPI_ALLREDUCE(conv_fl, CONVERGENCE_FL, 1, MPI_LOGICAL, MPI_LAND, &
       !             & SOLVER_COMM, ierr)

    else if(terminal_kind==0) then
       RNRM=0.0
       do i = 1, N
          RNRM=RNRM+residual_vec(i) ** 2
       end do
       
       RNRM2 = RNRM

       !        call MPI_ALLREDUCE(RNRM, RNRM2, 1, MPI_DOUBLE_PRECISION, MPI_SUM, &
       !             & SOLVER_COMM, ierr)
       RESID = dsqrt(RNRM2/BNRM2)
       CONVERGENCE_FL = (RESID <= conv_criterion_val)
    end if
    

  END SUBROUTINE terminal_cond_check

  subroutine terminal_cond_vec_clear()
    if(associated(conv_criterion_vec)) deallocate(conv_criterion_vec)
  end subroutine terminal_cond_vec_clear

end module terminal_condition
