!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

module vec_sub
contains
  subroutine gauss_jordan(A,X,B,N, gauss_FL)
    include 'precision.inc'
    integer(kind=kint) , intent(in) :: N
    real(kind=kreal) , dimension(:,:) , intent(in) :: A
    real(kind=kreal) , dimension(:) , intent(inout) :: X
    real(kind=kreal) , dimension(:) , intent(in) :: B
    logical, intent(out) :: gauss_FL
    real(kind=kreal) , allocatable :: A1(:,:),w(:)
    real(kind=kreal) :: ar,am,t
    integer(kind=kint) :: i,k,m

    allocate(A1(N,N),w(N))
    X(:) = 0.0d0
    A1(:,:) = A(:,:)
    X(:) = B(:)

    DO k = 1,N
       !C-----ʬpivot-----C!
       m = k
       am = abs(A1(k,k))
       DO i = k+1,N !C-----A(i,k)ͤȤʤmԤõ-----C!
          if(abs(A1(i,k)) > am) then
             am = abs(A1(i,k))
             m = i
          end if
       END DO

       if(am == 0.0d0) then
          write(*,*) "LU factorization failure" 
          gauss_FL = .false.
          return
       end if
       if(k /= m) then !C-----kԤmԤؤ-----C!
          w(k:N) = A1(k,k:N)
          A1(k,k:N) = A1(m,k:N)
          A1(m,k:N) = w(k:N)
          t = X(k)
          X(k) = X(m)
          X(m) = t
       end if
       !C-----gauss_jordann-----C!
       ar = 1.0d0 / A1(k,k)
       A1(k,k) = 1.0d0
       A1(k,k+1:N) = ar * A1(k,k+1:N)
       X(k) = ar * X(k)

       DO i = 1,N
          if (i /= k) then !C-----iԤkNǤX(i)Ф黻-----C!
             A1(i,k+1:N) = A1(i,k+1:N) - A1(i,k) * A1(k,k+1:N)
             X(i) = X(i) - A1(i,k) * X(k)
             A1(i,k) = 0.0d0
          end if
       END DO
    END DO
    gauss_FL = .true.
    deallocate(A1,w)

  end subroutine gauss_jordan

  subroutine croutlu(a, N, indx, err_fl, EPS)
    include 'precision.inc'
    real(kind=kreal), intent(inout) :: a(:,:)
    real(kind=kreal), intent(in) :: EPS
    integer(kind=kint), intent(out) :: indx(:)
    integer(kind=kint), intent(in) :: n
    logical, intent(out) :: err_fl
    
    real(kind=kreal), allocatable :: vv(:)
    real(kind=kreal) :: big, dum, sum, temp
    integer(kind=kint) :: i,j,k
    
    allocate(vv(N))

    
    do i = 1, N
       big = 0.
       do j = 1, N
          dum = dabs(a(i,j))
          if(dum > big ) big = dum
       end do
       if(big < EPS) then
          err_fl = .true.
          return
       end if
       vv(i) = 1.0/big
    end do
    
    do j = 1, N
       do i = 1, j-1
          sum = a(i,j)
          do k = 1,i-1
             sum = sum - a(i,k)*a(k,j)
          end do
          a(i,j) = sum
       end do
       big = 0.0
       do i = j, N
          sum = a(i,j)
          do k = 1, j-1
             sum = sum - a(i,k)*a(k,j)
          end do
          a(i,j) = sum
          dum = vv(i)*dabs(sum)
          if(dum >= big) then
             big = dum
             imax = i
          end if
       end do

           
       if(j /= imax ) then
          do k = 1, n
             dum = a(imax,k); a(imax, k) = a(j,k); a(j,k) = dum
          end do
          vv(imax) = vv(j)
       end if

       indx(j) = imax
       if(-EPS<a(j,j).and. a(j,j)<EPS) then
          err_fl = .true.          
          return
       end if
       if(j /= N) then
          dum = 1.0/a(j,j)
          do i = j+1,N
             a(i,j) = a(i,j)*dum
          end do
       end if
    end do
    deallocate(vv)
       
    err_fl = .false.
    
  end subroutine croutlu
  
  subroutine lusub(a, N, indx, b)
    include 'precision.inc'
    real(kind=kreal), intent(in) :: a(:,:)
    integer(kind=kint), intent(in) :: N, indx(:)
    real(kind=kreal), intent(inout) :: b(:)
    
    integer(kind=kint) :: i, ii,ip,j
    real(kind=kreal) :: sum
    ii = 0
    do i =1, N
       ip = indx(i)
       sum=b(ip);b(ip)=b(i)
       if(ii/=0) then
          do j=ii,i-1
             sum = sum - a(i,j)*b(j)
          end do
       else if(sum /= 0) then
          ii = i
       end if
       b(i) = sum
    end do
    
    do i = N, 1, -1 
       sum = b(i)
       do j =i+1,N
          sum = sum - a(i,j)*b(j)
       end do
       b(i) = sum /a(i,i)
    end do
    
  end subroutine lusub




  subroutine Gram_Schmidt(a,sv,n) 
    include 'precision.inc'
    integer(kind=kint), intent(in) :: n
    integer(kind=kint), intent(in) :: sv 
    real(kind=kreal), intent(inout) :: a(n,sv) 
    real(kind=kreal) ::  dotp
    integer(kind=kint) :: k,j

    a(:,1) = nomal_vec(a(:,1),n) 
    DO k = 2,sv
       DO j = 1,k-1
          dotp = dot_product(a(1:n,k),a(1:n,j))
          a(1:n,k) = a(1:n,k) - dotp*a(1:n,j)
       END DO
       a(1:n,k) = nomal_vec(a(1:n,k:k),n)

    END DO
  end subroutine  Gram_Schmidt

  function nomal_vec(v,n) result(nv)
    include 'precision.inc'

    integer(kind=kint), intent(in) :: n
    real(kind=kreal), intent(in) :: v(n)
    real(kind=kreal) nv(n), v1

    v1 = sqrt(dot_product(v,v))
    if(v1 == 0.0d0) then
       nv(:) = 0.0d0
    else
       nv(:) = v(:) / v1
    end if
  end function nomal_vec

  subroutine my_dot_product(vec1, vec2, n, v)
    include 'precision.inc'
    real(kind=kreal), intent(in) :: vec1(:), vec2(:)
    integer(kind=kint), intent(in) :: n
    real(kind=kreal), intent(out) :: v
    integer(kint) :: i
    v = 0
    do i = 1, n
       v = v + vec1(i)*vec2(i)
    end do
  end subroutine my_dot_product

end module vec_sub
